package org.cn3.morpho.analysis;

import java.io.File;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.data.NeuronTreeClassification;
import org.cn3.morpho.data.SWCReader;
import org.cn3.morpho.domain.Asymmetry;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.NeuronMetadataComplete;
import org.cn3.morpho.domain.NeuronProcess;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.resources.MorphoDataProperties;
import org.cn3.morpho.utils.BinaryTreeUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Reads all SWC files, generates caulescence measure for each arbor, and writes all to a file
 *
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.morpho.scripts.CaulescenceCalculator -o /home/tgillett/ClusterAnalysis/
 *
 */
public class CaulescenceCalculator {

	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_MIN_NUM_BIFS = "minBifs";
	
	private static final int DEFAULT_MIN_NUM_BIFS = 15;
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try{
			JSAPResult config = readParameters(args);
			String outputDir = config.getString(PARAM_OUTPUT_DIR);
			int minNumBifs = config.getInt(PARAM_MIN_NUM_BIFS);
			boolean debug = false;

			
			String swcDir = MorphoDataProperties.getInstance().getSwcDir("NM");
			File listFile = new File(swcDir);
			File[] directories = listFile.listFiles(), fileList;
	
			// Load metadata of neurons/files
			Map<String,NeuronMetadataComplete> metaData = MetaDataManager.getExtendedCellMetaData();
	
			Pattern namePattern = Pattern.compile("(.*?)(\\.AutoOnly)?(\\.CNG)?\\.swc");
			Matcher nameMatcher;
	
			File archiveDir, directoryOfFiles, inputFile;
			NeuronMetadataComplete currentNeuron;
			Soma<BinaryTreeNode> soma;
			NeuronProcess<BinaryTreeNode> generalDendrite;
			String currentNeuronName;
			
			double caulescence;
			NumberFormat nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(3);
			
			StringBuffer sb = new StringBuffer();
			sb.append("NeuriteName,Caulescence\n");
			
			Set<String> completedNeurons = new HashSet<String>(12000);
			
			for (int i = 0; i < directories.length; i++){
				// Makes sure we are looking at a directory
				if (directories[i].isDirectory()){
					archiveDir = directories[i];

					//System.out.println(archiveDir.getName());
					
					// Gets directory containing the swc files
					directoryOfFiles = new File(archiveDir.getAbsolutePath()+"/CNG version/");
					//System.out.println(directoryOfFiles + " " + directoryOfFiles.exists());
					if (!directoryOfFiles.exists()){
						directoryOfFiles = new File(archiveDir.getAbsolutePath()+"/CNG Version");
					}
					if (!directoryOfFiles.exists()){
						directoryOfFiles = new File(archiveDir.getAbsolutePath()+"/CNGversion");
					}
					if (!directoryOfFiles.exists()){
						directoryOfFiles = new File(archiveDir.getAbsolutePath());
					}
					if (directoryOfFiles.exists()){
						fileList = directoryOfFiles.listFiles();
						
						// Runs through all files for a given contributor
						//System.out.println("Dir "+directoryOfFiles.getAbsolutePath()+" with "+fileList.length+" files");
						for (int j = 0; j < fileList.length; j++){
							inputFile = fileList[j];
							
							if (inputFile.isFile()){
								// pull neuron metadata
								nameMatcher = namePattern.matcher(inputFile.getName());
								if (!nameMatcher.matches()){
									System.out.println("ERROR: Name pattern mismatch: "+inputFile.getName()+" (from "+archiveDir.getName()+")");
								}
								else{
									currentNeuronName = nameMatcher.group(1);
									currentNeuron = metaData.get(currentNeuronName);
									if (!completedNeurons.contains(currentNeuronName)){
										completedNeurons.add(currentNeuronName);
										
	//									if (archiveDir.getName().equals("brunjes") || currentNeuronName.equals("VN041305-0-a")){
	//										System.out.println("Running "+currentNeuronName);
	//									}
	
										if (currentNeuron != null){
											currentNeuron.setLabName(archiveDir.getName());
	
											// Convert SWC data to binary trees for axon and dendrites
											soma = SWCReader.convertSwcToBinarySoma(inputFile, NeuronTreeClassification.ARBORIZATION, true);
											// Take all lengths so as to properly determine whether both apical and basal trees exist
											// Check each arbor type and include in final output
											if (soma.hasAxonalTree() && soma.getAxonalTree().getTreeHead().getDegree()-1 >= minNumBifs){
												if (debug) System.out.println("Axon");
												caulescence = soma.getAxonalTree().getTreeHead().getAsymmetry(
														BinaryTreeUtils.METRIC_PATH_LENGTH,Asymmetry.SIZE_MODE_SUM).getCaulescence();
												sb.append(currentNeuronName+"-Axon,"+nf.format(caulescence)+"\n");
											}
	
											// Check basal and apical dendrites
											if (soma.hasApicalDendriteTree() || soma.hasBasalDendriteTree()){
												generalDendrite = null;
												if (!currentNeuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
													if (soma.hasApicalDendriteTree() && !soma.hasBasalDendriteTree()){
														generalDendrite = soma.getApicalDendriteTree();
													}
													else if (soma.hasBasalDendriteTree() && !soma.hasApicalDendriteTree()){
														generalDendrite = soma.getBasalDendriteTree();
													}
												}
	
												if (generalDendrite != null){
													if (generalDendrite.getTreeHead().getDegree()-1 >= minNumBifs){
														if (debug) System.out.println("Dendrite");
														caulescence = generalDendrite.getTreeHead().getAsymmetry(
																BinaryTreeUtils.METRIC_PATH_LENGTH,Asymmetry.SIZE_MODE_SUM).getCaulescence();
														sb.append(currentNeuronName+"-Dendrite,"+nf.format(caulescence)+"\n");
													}
												}
												else{
													if (soma.hasApicalDendriteTree() && soma.getApicalDendriteTree().getTreeHead().getDegree()-1 >= minNumBifs){
														if (debug) System.out.println("Apical");
														caulescence = soma.getApicalDendriteTree().getTreeHead().getAsymmetry(
																BinaryTreeUtils.METRIC_PATH_LENGTH,Asymmetry.SIZE_MODE_SUM).getCaulescence();
														sb.append(currentNeuronName+"-ApicalDendrite,"+nf.format(caulescence)+"\n");
													}
													if (soma.hasBasalDendriteTree() && soma.getBasalDendriteTree().getTreeHead().getDegree()-1 >= minNumBifs){
														if (debug) System.out.println("Basal");
														caulescence = soma.getBasalDendriteTree().getTreeHead().getAsymmetry(
																BinaryTreeUtils.METRIC_PATH_LENGTH,Asymmetry.SIZE_MODE_SUM).getCaulescence();
														sb.append(currentNeuronName+"-BasalDendrite,"+nf.format(caulescence)+"\n");
													}
												}
											}
	//}
										}
										else{
											System.out.println(currentNeuronName+" not in metadata");
										}
									}
								}
							}
						}
					}
				}
			}
			DateFormat df = new SimpleDateFormat("yyyy.MM.dd");
			String filename = outputDir+"/NM-Caulescence"+df.format(new Date())+".csv";
			DataWriter.writeStringToFile(filename, sb.toString());
			System.out.println("Wrote caulescence file to "+filename);

		}catch (Exception e){
			e.printStackTrace();
		}
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_OUTPUT_DIR)
            .setStringParser(JSAP.STRING_PARSER)
            .setRequired(true)
            .setShortFlag('o') 
            .setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MIN_NUM_BIFS)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+DEFAULT_MIN_NUM_BIFS)
        .setShortFlag('m') 
        .setLongFlag(PARAM_MIN_NUM_BIFS);
        jsap.registerParameter(opt);
        
        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CaulescenceCalculator.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
