package org.cn3.morpho.data;

import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.domain.SwcTreeNode;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Cleans SWC files with various errors
 * 	- Nodes are out of order (child has a lower id than parent)
 *
 */
public class SWCReorderer {

	private static Pattern swcFilenameP = Pattern.compile("(.+)\\.(SWC|swc)");
	
	private static final boolean DEFAULT_INCLUDE_COMMENTS = true;
	
	public static String correctSWCNodeOrder(String inputFilename) throws IOException{
		return correctSWCNodeOrder(inputFilename, DEFAULT_INCLUDE_COMMENTS);
	}
	public static String correctSWCNodeOrder(String inputFilename, boolean includeComments) throws IOException{
		File inputFile = new File(inputFilename);
		Matcher m = swcFilenameP.matcher(inputFile.getName());
		if (m.matches()){
			String cleanedFilename = inputFile.getParent()+"/"+m.group(1)+"-reordered.swc";
			correctSWCNodeOrder(inputFilename,cleanedFilename);
			return cleanedFilename;
		}else{
			throw new IOException("Unexpected extension for file "+inputFilename+". Expected .swc extension.");
		}
	}
	public static void correctSWCNodeOrder(String inputFilename, String cleanedFilename) throws IOException{
		correctSWCNodeOrder(inputFilename, cleanedFilename, DEFAULT_INCLUDE_COMMENTS);
	}
	public static void correctSWCNodeOrder(String inputFilename, String cleanedFilename, boolean includeComments) throws IOException{
		File inputFile = new File(inputFilename);
		
		// Read Swc structure
		List<SwcTreeNode> treeList = SWCReader.convertSwcToTreeNodesSpecial(inputFile);
		
		//System.out.println(treeList.size());
		
		// Reorder node ids
		int currentStartId = 1;
		for (SwcTreeNode root : treeList){
			currentStartId = root.reorderNodeIds(currentStartId);
		}
		
		// Write corrected Swc to file
		if (includeComments){
			// Get comments
			StringBuffer comments = new StringBuffer();
			comments.append(SWCReader.readSwcComments(inputFile));
			DateFormat df = DateFormat.getInstance();
			comments.append("# Nodes reordered on "+df.format(new Date())+" using SWCCleaner.jar (written by Todd Gillette - tgillett@gmu.edu)\n");
			//System.out.println(treeList.size()+" "+" "+treeList.get(0).getBranchSize());
			SWCWriter.writeSwcTreesToFile(new File(cleanedFilename), treeList, comments.toString());
		}else{
			SWCWriter.writeSwcTreesToFile(new File(cleanedFilename), treeList);
		}
	}

	private static final String PARAM_INPUT_DIRECTORY = "inputDir";
	private static final String PARAM_OUTPUT_DIRECTORY = "outputDir";
	//private static final String PARAM_OUTPUT_FILE = "outputFile";
/*
	public static void main(String[] args){
		try{
			correctSWCNodeOrder(
					"C:/Documents and Settings/gillette/My Documents/Downloads/ca1c-pruned.swc",
					"C:/Documents and Settings/gillette/My Documents/Downloads/Reordered/ca1c-pruned-Reordered.swc");
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	*/
	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			String inputDir = config.getString(PARAM_INPUT_DIRECTORY);
			File inputDirFile = new File(inputDir), outputDirFile;
			File[] fileList;
			
			String outputDirectory = null;
			if (config.userSpecified(PARAM_OUTPUT_DIRECTORY)){
				outputDirectory = config.getString(PARAM_OUTPUT_DIRECTORY);
			}
			
			if (inputDirFile.isFile()){
				fileList = new File[1];
				fileList[0] = inputDirFile;
				if (outputDirectory == null){
					outputDirectory = inputDirFile.getParent();
				}
			}else{
				fileList = inputDirFile.listFiles();
				if (outputDirectory == null){
					outputDirectory = inputDirFile.getAbsolutePath()+"/Reordered";
				}
			}
			outputDirFile = new File(outputDirectory);
			if (!outputDirFile.exists()){
				outputDirFile.mkdirs();
			}
			
			File inputFile;
			String outputFilename;
			for (int i = 0; i < fileList.length; i++){
				inputFile = fileList[i];
				if (inputFile.getName().endsWith(".swc") || inputFile.getName().endsWith(".SWC")){
					outputFilename = outputDirectory+"/"+inputFile.getName().replaceAll("\\.(SWC|swc)$", "-Reordered.swc");
					correctSWCNodeOrder(inputFile.getAbsolutePath(), outputFilename);
					System.out.println("Wrote corrected file to "+outputFilename);
				}
			}
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
		FlaggedOption opt;

    	opt = new FlaggedOption(PARAM_INPUT_DIRECTORY)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true) 
		.setShortFlag('i') 
		.setLongFlag(PARAM_INPUT_DIRECTORY);
        jsap.registerParameter(opt);

    	opt = new FlaggedOption(PARAM_OUTPUT_DIRECTORY)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setShortFlag('o') 
		.setLongFlag(PARAM_OUTPUT_DIRECTORY);
        jsap.registerParameter(opt);
/*
    	opt = new FlaggedOption(PARAM_INPUT_FILE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true) 
		.setShortFlag('i') 
		.setLongFlag(PARAM_INPUT_FILE);
        jsap.registerParameter(opt);

    	opt = new FlaggedOption(PARAM_OUTPUT_FILE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setShortFlag('o') 
		.setLongFlag(PARAM_OUTPUT_FILE);
        jsap.registerParameter(opt);
*/
        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SWCReorderer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        return config;
	}

}
