package org.cn3.morpho.data;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;

import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.domain.SwcDataNode;
import org.cn3.morpho.domain.SwcTreeNode;
import org.cn3.morpho.utils.BinaryTreeUtils;

public class SwcEditor extends SWCWriter {

	private File swcFile;
	private String arborType;
	private SwcTreeNode rootNode;
	private SwcTreeNode swcTree;
	private BinaryTreeNode binaryTree;
	private boolean useParentNodeForTerminalType = false;
	
	private List<SwcTreeNode> otherSwcTrees;

	public SwcEditor(String swcFilename, String arborType) throws Exception{
		this.swcFile = new File(swcFilename);
		this.arborType = arborType;
		processFile();
	}
	public SwcEditor(File swcFile, String arborType) throws Exception{
		this.swcFile = swcFile;
		this.arborType = arborType;
		processFile();
	}
	
	public File getSwcFile() {
		return swcFile;
	}
	public String getArborType() {
		return arborType;
	}
	public SwcTreeNode getSwcTree() {
		return swcTree;
	}
	public BinaryTreeNode getBinaryTree() {
		return binaryTree;
	}
	public List<SwcTreeNode> getOtherSwcTrees() {
		return otherSwcTrees;
	}
	public void setOtherSwcTrees(List<SwcTreeNode> otherSwcTrees) {
		this.otherSwcTrees = otherSwcTrees;
	}
	public boolean isUseParentNodeForTerminalType() {
		return useParentNodeForTerminalType;
	}
	public void setUseParentNodeForTerminalType(boolean useParentNodeForTerminalType) {
		this.useParentNodeForTerminalType = useParentNodeForTerminalType;
	}
	private void processFile() throws Exception{
		Soma<SwcTreeNode> soma = SWCReader.convertSwcToSoma(swcFile);
		rootNode = soma.getSomaRoot().getTreeHead();
		otherSwcTrees = new ArrayList<SwcTreeNode>();
		if(arborType.equalsIgnoreCase("dendrite")||arborType.equalsIgnoreCase("basaldendrite")){
			if (soma.getBasalDendriteTrees().size() > 1){
				// Check whether all trees have the same parent; if so, use the rootNode rather than combining
				SwcTreeNode parent = soma.getBasalDendriteTrees().get(0).getParent();
				for (SwcTreeNode treeRoot : soma.getBasalDendriteTrees()){
					if (parent != treeRoot.getParent()){
						parent = null;
					}
				}
				
				if (parent == null){
					swcTree = new SwcTreeNode();
					for (SwcTreeNode treeRoot : soma.getBasalDendriteTrees()){
						swcTree.addChildNode(treeRoot);
						swcTree.setX(swcTree.getX()+treeRoot.getX());
						swcTree.setY(swcTree.getY()+treeRoot.getY());
						swcTree.setZ(swcTree.getZ()+treeRoot.getZ());
					}
					swcTree.setX(swcTree.getX()/soma.getBasalDendriteTrees().size());
					swcTree.setY(swcTree.getY()/soma.getBasalDendriteTrees().size());
					swcTree.setZ(swcTree.getZ()/soma.getBasalDendriteTrees().size());
				}
				// If all basal dendrite trees have a parent, and the same parent, use it as the root
				else{
					swcTree = parent;
				}
			}else{
				swcTree = soma.getBasalDendriteTrees().get(0);
			}
			if (soma.hasApicalDendriteTree()) otherSwcTrees.add(soma.getApicalDendriteTree().getTreeHead());
			if (soma.hasAxonalTree()) otherSwcTrees.add(soma.getAxonalTree().getTreeHead());
		}else if(arborType.equalsIgnoreCase("apicaldendrite")){
			if (soma.getApicalDendriteTrees().size() > 1){
				swcTree = new SwcTreeNode();
				for (SwcTreeNode treeRoot : soma.getApicalDendriteTrees()){
					swcTree.addChildNode(treeRoot);
				}
			}else{
				swcTree = soma.getApicalDendriteTrees().get(0);
			}
			if (soma.hasBasalDendriteTree()) otherSwcTrees.add(soma.getBasalDendriteTree().getTreeHead());
			if (soma.hasAxonalTree()) otherSwcTrees.add(soma.getAxonalTree().getTreeHead());
		}else if(arborType.equalsIgnoreCase("axon")){
			swcTree = soma.getAxonalTree().getTreeHead();
			if (soma.hasBasalDendriteTree()) otherSwcTrees.add(soma.getBasalDendriteTree().getTreeHead());
			if (soma.hasApicalDendriteTree()) otherSwcTrees.add(soma.getApicalDendriteTree().getTreeHead());
		}else{
			throw new Exception("For 'arborType' parameters, only dendrites, apical dendrites, and axons currently supported.");
		}
		// Convert binary version of tree
		binaryTree = SWCReader.convertToBinaryTreeNode(swcTree, true);
		
	}
	
	/**
	 * Copies an input tree, providing new node ids for each node
	 * @param root
	 * @return SwcTreeNode
	 */
	public static SwcTreeNode createNodeIds(SwcTreeNode root){
		List<SwcTreeNode> trees = new ArrayList<SwcTreeNode>();
		trees.add(root);
		return createNodeIds(trees).get(0);
	}
	public static List<SwcTreeNode> createNodeIds(List<SwcTreeNode> trees){
		List<SwcTreeNode> newTrees = new ArrayList<SwcTreeNode>();
		SwcTreeNode newRoot = null, originalNode, newNode, newParent;
		Map<SwcTreeNode,SwcTreeNode> oldNewMap = new HashMap<SwcTreeNode,SwcTreeNode>();
		Stack<SwcTreeNode> nodeStack = new Stack<SwcTreeNode>();
		for (SwcTreeNode root : trees) nodeStack.push(root);
		
		int nodeId = 0;
		while (!nodeStack.isEmpty()){
			nodeId++;
			originalNode = nodeStack.pop();
			newNode = new SwcTreeNode();
			newNode.setData(originalNode.getData().copy());
			newNode.setNodeId(nodeId);
			oldNewMap.put(originalNode, newNode);
			if (originalNode.hasParent()){
				newParent = oldNewMap.get(originalNode.getParent());
				newNode.setParent(newParent);
				newNode.setParentId(newParent.getNodeId());
			}else{
				newRoot = newNode;
				newRoot.setParentId(-1);
				newTrees.add(newRoot);
			}
			
			if (originalNode.hasChildren()){
				nodeStack.addAll(originalNode.getChildren());
			}
		}
		
		return newTrees;
	}

	public void writeSwc(String outputFilename) throws Exception{
		writeSwc(new File(outputFilename));
	}
	public void writeSwc(String outputFilename, boolean includeOtherTrees) throws Exception{
		writeSwc(new File(outputFilename), includeOtherTrees);
	}
	public void writeSwc(File outputFile) throws Exception{
		writeSwc(outputFile,false);
	}
	public void writeSwc(File outputFile, boolean includeOtherTrees) throws Exception{
		// Transfer the new bifurcation types to the full tree (i.e. all compartments)
		retypeTreeByBinaryTreeTypes(swcTree, binaryTree);

		// Reset node ids for single arbor (creates node copies to avoid id corruption of original data and swc->binary identity)
		/*
		List<SwcTreeNode> trees = new ArrayList<SwcTreeNode>();
		trees.add(swcTree);
		if (includeOtherTrees){
			trees.addAll(otherSwcTrees);
		}
		trees = createNodeIds(trees);
		*/
		SwcTreeNode startPoint;
		if (includeOtherTrees){
			startPoint = rootNode;
		}else{
			startPoint = swcTree;
		}
		
		// Get all nodes and order them by id for writing to file
		Set<SwcDataNode> dataNodeSet = new TreeSet<SwcDataNode>();
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		//for (SwcTreeNode root : trees) stack.push(root);
		stack.push(startPoint);
		SwcTreeNode node;
		while(!stack.isEmpty()){
			node = stack.pop();
			if (node.getData() != null){
				dataNodeSet.add(node.getData());
			}
			if (node.hasChildren()){
				stack.addAll(node.getChildren());
			}
		}

		/**
		 * Redraw arbor with new types
		 */
		StringBuffer sb = new StringBuffer();

		// Write to string buffer in preparation for writing to file
		for (SwcDataNode dataNode : dataNodeSet){
			sb.append(dataNode.getNodeId()+" "+dataNode.getType()+" "+dataNode.getX()+" "+dataNode.getY()+" "+dataNode.getZ()+" "+dataNode.getRadius()+" "+dataNode.getParentId()+"\n");
		}
				
		// Write retyped arbor to file
		DataWriter.writeStringToFile(outputFile, sb.toString());
		System.out.println("Wrote to "+outputFile.getAbsolutePath());
	}

	private void retypeTreeByBinaryTreeTypes(SwcTreeNode tree, BinaryTreeNode binaryTree){
		// Create nodeId->type map for binary tree nodes
		Map<Integer,Integer> binaryNodeTypeMap = new HashMap<Integer,Integer>();
		List<BinaryTreeNode> binaryNodes = BinaryTreeUtils.createNodeList(binaryTree);
		for (BinaryTreeNode node : binaryNodes){
			binaryNodeTypeMap.put(node.getSwcData().getNodeId(), node.getSwcData().getType());
		}

		// Make a nodeId->type map for swc nodes associated with alignment bifurcation nodes
		// First load all terminal nodes
		Stack<SwcTreeNode> swcNodeStack = new Stack<SwcTreeNode>();
		swcNodeStack.addAll(BinaryTreeUtils.createTerminalNodeList(tree));

		/**
		 *  Traverse tree, setting node types
		 */
		
		// Work up from terminal nodes, setting type
		int tailType, headType, nodeCount, currCount, switchCount;
		SwcTreeNode childNode, traversalNode;
		boolean isRootSegment;
		Set<SwcTreeNode> completed = new HashSet<SwcTreeNode>();
		while (!swcNodeStack.isEmpty()){
			childNode = swcNodeStack.pop();
			// Determine type of current (tail) node
			if (binaryNodeTypeMap.containsKey(childNode.getNodeId())){
				tailType = binaryNodeTypeMap.get(childNode.getNodeId());
			}else{
				tailType = -1;
			}

			// Determine type of parent bifurcation node
			nodeCount = 0;
			traversalNode = childNode.getParent();
			while (traversalNode.hasParent() && traversalNode.getChildren().size() == 1){
				nodeCount++;
				traversalNode = traversalNode.getParent();
			}
			isRootSegment = !traversalNode.hasParent();
			headType = binaryNodeTypeMap.containsKey(traversalNode.getNodeId())  
					? binaryNodeTypeMap.get(traversalNode.getNodeId())
					: traversalNode.getType();
			//System.out.println(childNode.getNodeId()+" isLeaf? "+childNode.isLeaf()+"; type: "+childNode.getType()+"; headType: "+headType+"; Pos: "+childNode.getData().positionString());					
			if (tailType == -1 || (childNode.isLeaf() && useParentNodeForTerminalType)){
				tailType = headType;
			}

			// Traverse up again, this time setting the node type based on branch head and tail types
			traversalNode = childNode;
			currCount = 0;
			// If tail is leaf or head and tail are same type, all nodes are the same type, otherwise switch half-way 
			switchCount = childNode.isLeaf() || headType == tailType ? nodeCount+1 : nodeCount/2;
			do{
				if (currCount < switchCount || isRootSegment){
					//swcTypeMap.put(traversalNode.getNodeId(),tailType);
					traversalNode.setType(tailType);
				}else{
					//swcTypeMap.put(traversalNode.getNodeId(),headType);
					traversalNode.setType(headType);
				}
				currCount++;
				traversalNode = traversalNode.getParent();
			}while (traversalNode.getChildren().size() == 1 && traversalNode.hasParent());
			completed.add(childNode);
			
			// Add parent (if parent hasn't already been added)
			if (traversalNode.hasParent() && !completed.contains(traversalNode)){
				swcNodeStack.push(traversalNode);
			}
		}
	}
	
	public static void resetNodeTypes(BinaryTreeNode tree, int type){
		for (BinaryTreeNode node : BinaryTreeUtils.createNodeList(tree)){
			node.getSwcData().setType(type);
		}
	}
	public void resetNodeTypes(int type){
		for (BinaryTreeNode node : BinaryTreeUtils.createNodeList(binaryTree)){
			node.getSwcData().setType(type);
		}
	}
}
