package org.cn3.morpho.domain;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import cern.jet.stat.Probability;

/**
 * 
 * @author gillette
 *
 * Commonly used object in analysis, providing a List of numerical data
 * and parameters calculated from that data.
 *
 */
public class DistributionBean{
	protected int id;
	protected double total = 0;
	protected int count = 0;
	protected List<Double> data = null;
	protected double mean = 0;
	protected double variance = 0;
	protected double standardDeviation = 0;
	protected double standardError = 0;
	protected double skew = 0;
	protected double kurtosis = 0;
	
	// percentile and related value
	protected Map<Double,Double> percentileMap = new LinkedHashMap<Double,Double>(5);
	protected boolean meanReady = false;
	protected double interquartileRange = 0;
	
	protected double adjustedMean = 0;
	protected double adjustedStandardDeviation = 0;
	protected double adjustedVariation = 0;
	protected int adjustedCount = 0;
	protected Map<Double,Double> adjustedPercentileMap = new LinkedHashMap<Double,Double>(5);
	
	public DistributionBean(){
		data = new LinkedList<Double>();
	}
	
	public int getId() {
		return id;
	}
	public void setId(int id) {
		this.id = id;
	}
	public int getCount() {
		return count;
	}
	public void setCount(int count) {
		this.count = count;
		//meanReady = false;
	}
	public double getTotal() {
		return total;
	}
	public void setTotal(double total) {
		this.total = total;
		meanReady = false;
	}
	public double getMean(){
		if (!meanReady && count > 0){
			meanReady = true;
			mean = total/count;
		}
		return mean;
	}
	public double getAverage(){
		return getMean();
	}
	public void addData(int data){
		addData(new Integer(data));
	}
	public void addData(double data){
		addData(new Double(data));
	}
	public void addData(long data){
		addData(new Long(data));
	}
	public void addData(float data){
		addData(new Float(data));
	}
	public void addData(Number number){
		if (data == null){
			data = new LinkedList<Double>();
			total = 0;
		}
		total += number.doubleValue();
		int index = Collections.binarySearch(data, number.doubleValue());
		if (index < 0){
			data.add(-index - 1, number.doubleValue());
		}
		else{
			data.add(index, number.doubleValue());
		}
		count = data.size();
		mean = variance = standardDeviation = 0;
		meanReady = false;
	}
	public void addData(List<Double> dataList){
		for (int i = 0; i < dataList.size(); i++){
			addData((Number)dataList.get(i));
		}
	}

	public double getMedian(){
		return getPercentile(.5);
	}
	// Common percentiles
	public double get50thPercentile(){
		return getPercentile(.5);
	}
	public double get95thPercentile(){
		return getPercentile(.95);
	}
	public double get99thPercentile(){
		return getPercentile(.99);
	}

	/**
	 * @param value
	 * @return Percentile: the percent of values that value is greater than
	 */
	public double getPercentileForValue(double value){
		return getPercentileForValue(value, false);
	}
	/**
	 * @param value
	 * @return Percentile rank: the average percentile of the lowest and highest rank when the target value exists multiple 
	 *  times in the distribution
	 */
	public double getPercentileRankForValue(double value){
		return getPercentileForValue(value, true);
	}
	/**
	 * Using linear interpolation method for percentiles used in Matlab
	 * p(n) = (n-.5)/N where p = percentile, n = rank
	 * @param value
	 * @param useMiddleForDuplicates
	 * @return If useMiddleForDuplicates, the percent of values that value is greater than (percentile);
	 *  otherwise the average percentile of the lowest and highest rank when the target value exists multiple 
	 *  times in the distribution (percentile rank).
	 */
	public double getPercentileForValue(double value, boolean useMiddleForDuplicates){
		// Index is 0 to N-1, rank is 1 to N
		double aboveVal, belowVal, middleVal, abovePerc, belowPerc, percentile;
		int aboveInd = count - 1, belowInd = 0, middleInd;
	
		// If this value falls below the lowest value in the distribution
		belowVal = ((Number)data.get(belowInd)).doubleValue();
		if (value < belowVal){
			return 0;
		}
		else if (value == belowVal && !useMiddleForDuplicates){
			return 0.5/count;
		}
		// If this value falls above the highest value in the distribution
		aboveVal = ((Number)data.get(aboveInd)).doubleValue();
		if (value > aboveVal){
			return 1;
		}
		else if (value == aboveVal && !useMiddleForDuplicates){
			return (count-.5) / (count);
		}

		// Find the rank above and below the value
		while (aboveInd - belowInd > 1){
			// Find by splitting the remaining distribution in half and searching in log2 time
			middleInd = (aboveInd - belowInd)/2 + belowInd;
			middleVal = ((Number)data.get(middleInd)).doubleValue();
			if (value > middleVal){
				belowInd = middleInd;
			}
			else if (value < middleVal){
				aboveInd = middleInd;
			}
			else{
				// Check for multiple of the same value
				if (useMiddleForDuplicates){
					int maxInd = middleInd;
					while (maxInd < count && ((Number)data.get(maxInd)).doubleValue() == value){
						maxInd++;
					}
					int minInd = middleInd;
					while (minInd >= 0 && ((Number)data.get(minInd)).doubleValue() == value){
						minInd--;
					}
					maxInd--;
					minInd++;
					/**
					 *  Return the percentile rank, whereby the average of the low and high are given
					 *  when the target value exists multiple times in the distribution
					 */
					double rank = ((double)(maxInd + minInd))/2 + 1;
					return (rank - .5) / count;
				}
				else{
					int minInd = middleInd;
					while (minInd >= 0 && ((Number)data.get(minInd)).doubleValue() == value){
						minInd--;
					}
					minInd++;
					// Return the percent of values that the given value is greater than
					int rank = minInd + 1;
					return (double)(rank - .5) / count;
				}
			}
		}
		
		// Interpolate percentile
		int rank = aboveInd + 1;
		abovePerc = (double)(rank - .5) / (count);
		rank = belowInd + 1;
		belowPerc = (double)(rank - .5) / (count);
		// Old method was an alternative such that rank = percentile * (N+1)
		//belowPerc = (rank) / (count+1);
		// Other available method: rank = percentile*(N-1)+1
		
		aboveVal = ((Number)data.get(aboveInd)).doubleValue();
		belowVal = ((Number)data.get(belowInd)).doubleValue();
		
		percentile = belowPerc + 
			(abovePerc - belowPerc)*(value - belowVal)/(aboveVal - belowVal);
		
		return percentile;
	}
	
	// If percentile falls between two data points, returns an interpolation
	public double getPercentile(Double percentile){
		return getPercentile(percentile.doubleValue());
	}
	
	/**
	 * Using linear interpolation method for percentiles used in Matlab
	 * p(n) = (n-.5)/N where p = percentile, n = rank
	 * @param percentile
	 * @return value at percentile (interpolated if needed)
	 */
	public double getPercentile(double percentile){
		if (percentile >= 1 && percentile < 100){
			percentile = percentile / 100;
		}
		Double percentileObj = new Double(percentile);
		if (meanReady && percentileMap.containsKey(percentileObj)){
			return ((Double)percentileMap.get(percentileObj)).doubleValue();
		}
		if (data == null || percentile >= 1 || percentile < 0){
			return 0;
		}
		double rank = percentile * count + .5;
		double percentileIndex = rank - 1;
		double value;
		int belowInd = (int)(percentileIndex);
		int aboveInd = (int)(percentileIndex)+1;
		if (belowInd >= count - 1){
			value = ((Number)data.get(count - 1)).doubleValue();
		}
		else if (aboveInd <= 0){
			value = ((Number)data.get(0)).doubleValue();
		}
		else{
			double belowVal = ((Number)data.get(belowInd)).doubleValue();
			double aboveVal = ((Number)data.get(aboveInd)).doubleValue();
			/* only works if below and above are 1 apart, which they are in this case
			 * value = (belowVal * (percentile * count - below) 
				+ aboveVal * (above - percentile * count));*/
			value = belowVal + (aboveVal - belowVal) * (percentileIndex - belowInd)/(aboveInd - belowInd);
		}
		percentileMap.put(percentileObj, new Double(value));
		return value;
	}
	
	public double getStandardDeviation() {
		getMean();
		if (standardDeviation == 0 && data != null){
			getVariance();
			standardDeviation = Math.sqrt(variance);
		}
		return standardDeviation;
	}

	public double getStandardError() {
		getMean();
		if (standardError == 0 && data != null){
			if (variance == 0){
				getVariance();
				standardDeviation = Math.sqrt(variance);
			}
			standardError = standardDeviation/Math.sqrt(count);
		}
		return standardError;
	}

	public double getVariance() {
		getMean();
		double val;
		if (variance == 0 && data != null){
			for (int i = 0; i < data.size(); i++){
				val = ((Number)data.get(i)).doubleValue();
				variance += Math.pow((val - mean), 2);
			}
			variance = variance / (count-1);
		}
		return variance;
	}
	public List<Double> getData() {
		if (data != null){
			List<Double> copy = new LinkedList<Double>(data);
			return copy;
		}
		else{
			return null;
		}
	}
	
	public boolean equals(Object o){
		if (o != null && o.getClass().equals(this.getClass())){
			DistributionBean b = (DistributionBean)o;
			return (b.getId() == id);
		}
		return false;
	}

	public void setMean(double mean) {
		this.mean = mean;
		meanReady = true;
	}
	public void setStandardDeviation(double standardDeviation) {
		this.standardDeviation = standardDeviation;
		meanReady = true;
	}
	public void setVariance(double variance) {
		this.variance = variance;
		meanReady = true;
	}
	public void setPercentile(double percentile, double value) {
		if (percentile >= 1 && percentile < 100){
			percentile = percentile / 100;
		}
		percentileMap.put(new Double(percentile), new Double(value));
	}
	public int getAdjustedCount() {
		return adjustedCount;
	}
	public void setAdjustedCount(int adjCount) {
		this.adjustedCount = adjCount;
	}
	public double getAdjustedMean() {
		return adjustedMean;
	}
	public void setAdjustedMean(double adjMean) {
		this.adjustedMean = adjMean;
	}
	public Map<Double,Double> getAdjustedPercentileMap() {
		return adjustedPercentileMap;
	}
	public void setAdjustedPercentileMap(Map<Double,Double> adjustedPercentileMap) {
		this.adjustedPercentileMap = adjustedPercentileMap;
	}
	public void setAdjustedPercentile(double percentile, double value) {
		if (percentile >= 1 && percentile < 100){
			percentile = percentile / 100;
		}
		adjustedPercentileMap.put(new Double(percentile), new Double(value));
	}

	public double getAdjustedPercentile(Double percentile){
		return getAdjustedPercentile(percentile.doubleValue());
	}
	
	// Common percentiles
	public double getAdjusted50thPercentile(){
		return getAdjustedPercentile(.5);
	}
	public double getAdjusted95thPercentile(){
		return getAdjustedPercentile(.95);
	}
	public double getAdjusted99thPercentile(){
		return getAdjustedPercentile(.99);
	}

	public double getAdjustedPercentile(double percentile){
		Double percentileObj = new Double(percentile);
		if (adjustedPercentileMap.containsKey(percentileObj)){
			return ((Double)adjustedPercentileMap.get(percentileObj)).doubleValue();
		}
		if (percentile >= 1 && percentile < 100){
			percentileObj = new Double(percentile / 100);
			if (adjustedPercentileMap.containsKey(percentileObj)){
				return ((Double)adjustedPercentileMap.get(percentileObj)).doubleValue();
			}
		}
		return 0;
	}
	public double getAdjustedStandardDeviation() {
		return adjustedStandardDeviation;
	}
	public void setAdjustedStandardDeviation(double adjustedStandardDeviation) {
		this.adjustedStandardDeviation = adjustedStandardDeviation;
	}
	public double getAdjustedVariation() {
		return adjustedVariation;
	}
	public void setAdjustedVariation(double adjustedVariation) {
		this.adjustedVariation = adjustedVariation;
	}
	
	public double getInterquartileRange(){
		if (interquartileRange == 0){
			interquartileRange = getPercentile(.75) - getPercentile(.25);
		}
		return interquartileRange;
	}
	
	public boolean isOutlier(double value){
		return (value > getPercentile(.75) + 1.5*getInterquartileRange()
				|| value < getPercentile(.25) - 1.5*getInterquartileRange());
	}

	public double getLowerOutlierLimit(){
		return getPercentile(.25) - 1.5*getInterquartileRange();
	}

	public double getUpperOutlierLimit(){
		return getPercentile(.75) + 1.5*getInterquartileRange();
	}
	
	public double getSkewness() {
		getMean();
		double val;
		if (skew == 0 && data != null){
			getStandardDeviation();
			if (standardDeviation != 0){
				for (int i = 0; i < data.size(); i++){
					val = ((Number)data.get(i)).doubleValue();
					skew += Math.pow((val - mean), 3);
				}
				skew = (skew / count) / Math.pow(standardDeviation, 3);
			}
			else{
				skew = 0;
			}
		}
		return skew;
	}

	public double getKurtosis() {
		getMean();
		double val;
		if (kurtosis == 0 && data != null){
			getStandardDeviation();
			if (standardDeviation != 0){
				for (int i = 0; i < data.size(); i++){
					val = ((Number)data.get(i)).doubleValue();
					kurtosis += Math.pow((val - mean), 4);
				}
				kurtosis = (kurtosis / count) / Math.pow(standardDeviation, 4) - 3;
			}
			else{
				kurtosis = 0;
			}
		}
		return kurtosis;
	}

	public boolean isTtestSig(double mean2){
		return isTtestSig(mean2, .05, 0);
	}
	public boolean isTtestSig(double mean2, double alpha, int tail){
		double p = getTtestP(mean2);
		if (tail == 0){
			alpha = alpha/2;
			p = (p > .5 ? 1-p : p);
			return p < alpha;
		}
		else if (tail == -1)
			return p < alpha;
		else
			return 1-p < alpha;
	}
	public double getTtestP(double mean2){
		double p = Probability.normal(getMean(),getVariance(),mean2);
		return p;
		//return (p > .5 ? 1-p : p);
	}
	public double getTtestT(double mean2){
		return (getMean() - mean2)/(getStandardDeviation()/Math.sqrt(getCount()));
	}
	
	public boolean isTtestSig(DistributionBean dist2){
		return isTtestSig(dist2, .05, 0);
	}
	public boolean isTtestSig(DistributionBean dist2, double alpha, int tail){
		double p = getTtestP(dist2, tail);
		if (tail == 0){
			alpha = alpha/2;
			return p < alpha;
		}
		else if (tail == -1)
			return p < alpha;
		else
			return 1-p < alpha;
	}
	public double getTtestP(DistributionBean dist2){
		return getTtestP(dist2, 0);
	}
	public double getTtestP(DistributionBean dist2, int tail){
		double p = Probability.normal(getTtestT(dist2));
		switch (tail){
		case -1:
			p = 1 - p;
			break;
		case 0:
			if (p > .5){
				p = 1 - p;
			}
			break;
		case 1:
			break;
		}
		return p;
	}
	public double getTtestT(DistributionBean dist2){
		if (dist2.getCount() == getCount()){
			return (getMean() - dist2.getMean()) /
					Math.sqrt((getVariance() + dist2.getVariance())/getCount());
		}
		else{
			return (getMean() - dist2.getMean()) /
					Math.sqrt(
							((double)1/getCount()+(double)1/dist2.getCount()) *
							(getVariance()*(getCount()-1) + dist2.getVariance()*(dist2.getCount()-1)) /
							(getCount() + dist2.getCount() - 2));
		}
	}
	
	public static void main(String args[]){
		DistributionBean bean = new DistributionBean();
		Random rand = new Random();
		double val;
		/*
		for (int i = 0; i < 40; i++){
			val = rand.nextGaussian();
			//System.out.println(i+": "+val);
			bean.addData(val);
		}
		System.out.println(bean.getPercentile(.3));
		System.out.println(bean.getPercentile(.5));
		System.out.println(bean.getPercentile(.7));
		System.out.println();
		System.out.println(bean.getPercentileForValue(-1.5));
		System.out.println(bean.getPercentileForValue(-1));
		System.out.println(bean.getPercentileForValue(0));
		System.out.println(bean.getPercentileForValue(1));
		System.out.println(bean.getPercentileForValue(1.5));
		*/
//		bean.addData(1);
//
//		for (int i = 0; i < 17; i++){
//			bean.addData(2);
//		}
//		bean.addData(3);
//		bean.addData(3);
//		System.out.println(bean.getPercentileForValue(0));
//		System.out.println(bean.getPercentileForValue(0,true));
//		System.out.println(bean.getPercentileForValue(1));
//		System.out.println(bean.getPercentileForValue(1,true));
//		System.out.println(bean.getPercentileForValue(1.5,true));
//		System.out.println(bean.getPercentileForValue(2));
//		System.out.println(bean.getPercentileForValue(2, true));
//		System.out.println(bean.getPercentileForValue(3));
//		System.out.println(bean.getPercentileForValue(3, true));
		
	}
	
	public List<Double> getUniqueValues(){
		Set<Double> uniqueValues = new TreeSet<Double>(data);
		return new ArrayList<Double>(uniqueValues);
	}
	
	public void clearData(){
		data.clear();
		count = 0;
		total = 0;
	}
}