package org.cn3.morpho.domain;

/**
 * 
 * @author gillette
 *
 * Contains information regarding NMO morphometrics (derived from L-Measure).
 * Particularly important for interfacing with the NMO database and its
 * morphometrics table.
 *
 */
public class Measurements {

	public static final int N_STEMS = 1;
	public static final int N_BIFS = 2;
	public static final int N_BRANCHES = 3;
	public static final int WIDTH = 4;
	public static final int HEIGHT = 5;
	public static final int DEPTH = 6;
	public static final int DIAMETER = 7;
	public static final int LENGTH = 8;
	public static final int SURFACE = 9;
	public static final int VOLUME = 10;
	public static final int EUCLIDEAN_DIST = 11;
	public static final int PATH_DIST = 12;
	public static final int BRANCH_ORDER = 13;
	public static final int CONTRACTION = 14;
	public static final int FRAGMENTATION = 15;
	public static final int PARTITION_ASYMMETRY = 16;
	public static final int PK_CLASSIC = 17;
	public static final int BIF_AMPL_LOCAL = 18;
	public static final int BIF_AMPL_REMOTE = 19;
	public static final int FRACTAL_DIM = 20;
	
	public static final String[] MEASUREMENT_NAMES = {"Stems","Bifurcations","Branches","Width","Height","Depth","Diameter",
		"Length","Surface","Volume","EuclideanDistance","PathDistance","BranchOrder","Contraction","Fragmentation",
		"PartitionAsymmetry","PKClassic","LocalBifurcationAmpl","RemoteBifurcationAmpl","FractalDimension"};

	private int arborTypeId;
	private String neuronName;
	private double somaSurface;
	private int nStems;
	private int nBifs;
	private int nBranches;
	private double width;
	private double height;
	private double depth;
	private double diameter;
	private double length;
	private double surface;
	private double volume;
	private double euclideanDistance;
	private double pathDistance;
	private double branchOrder;
	private double contraction;
	private double fragmentation;
	private double partitionAsymmetry;
	private double pkClassic;
	private double bifAmplLocal;
	private double bifAmplRemote;
	private double fractalDim;
	
	public Measurements(){}
	
	public Measurements(int arborTypeId, String neuronName){
		this.arborTypeId = arborTypeId;
		this.neuronName = neuronName;
	}
	
	public int getArborTypeId() {
		return arborTypeId;
	}
	public void setArborTypeId(int arborTypeId) {
		this.arborTypeId = arborTypeId;
	}
	public String getArborType(){
		switch (arborTypeId){
		case NeuronMetadata.ARBOR_TYPE_APICAL:
			return "ApicalDendrite";
		case NeuronMetadata.ARBOR_TYPE_BASAL:
			return "BasalDendrite";
		case NeuronMetadata.ARBOR_TYPE_DENDRITE:
			return "Dendrite";
		case NeuronMetadata.ARBOR_TYPE_AXON:
			return "AXON";
		default:
			return null;
		}
	}
	public String getSequenceName(){
		return neuronName+"-"+getArborType();
	}
	public double getBifAmplLocal() {
		return bifAmplLocal;
	}
	public void setBifAmplLocal(double bifAmplLocal) {
		this.bifAmplLocal = bifAmplLocal;
	}
	public double getBifAmplRemote() {
		return bifAmplRemote;
	}
	public void setBifAmplRemote(double bifAmplRemote) {
		this.bifAmplRemote = bifAmplRemote;
	}
	public double getBranchOrder() {
		return branchOrder;
	}
	public void setBranchOrder(double branchOrder) {
		this.branchOrder = branchOrder;
	}
	public double getContraction() {
		return contraction;
	}
	public void setContraction(double contraction) {
		this.contraction = contraction;
	}
	public double getDepth() {
		return depth;
	}
	public void setDepth(double depth) {
		this.depth = depth;
	}
	public double getDiameter() {
		return diameter;
	}
	public void setDiameter(double diameter) {
		this.diameter = diameter;
	}
	public double getEuclideanDistance() {
		return euclideanDistance;
	}
	public void setEuclideanDistance(double euclideanDistance) {
		this.euclideanDistance = euclideanDistance;
	}
	public double getFractalDim() {
		return fractalDim;
	}
	public void setFractalDim(double fractalDim) {
		this.fractalDim = fractalDim;
	}
	public double getFragmentation() {
		return fragmentation;
	}
	public void setFragmentation(double fragmentation) {
		this.fragmentation = fragmentation;
	}
	public double getHeight() {
		return height;
	}
	public void setHeight(double height) {
		this.height = height;
	}
	public double getLength() {
		return length;
	}
	public void setLength(double length) {
		this.length = length;
	}
	public int getNBifs() {
		return nBifs;
	}
	public void setNBifs(int bifs) {
		nBifs = bifs;
	}
	public int getNBranches() {
		return nBranches;
	}
	public void setNBranches(int branches) {
		nBranches = branches;
	}
	public String getNeuronName() {
		return neuronName;
	}
	public void setNeuronName(String neuronName) {
		this.neuronName = neuronName;
	}
	public int getNStems() {
		return nStems;
	}
	public void setNStems(int stems) {
		nStems = stems;
	}
	public double getPartitionAsymmetry() {
		return partitionAsymmetry;
	}
	public void setPartitionAsymmetry(double partitionAsymmetry) {
		this.partitionAsymmetry = partitionAsymmetry;
	}
	public double getPathDistance() {
		return pathDistance;
	}
	public void setPathDistance(double pathDistance) {
		this.pathDistance = pathDistance;
	}
	public double getPkClassic() {
		return pkClassic;
	}
	public void setPkClassic(double pkClassic) {
		this.pkClassic = pkClassic;
	}
	public double getSomaSurface() {
		return somaSurface;
	}
	public void setSomaSurface(double somaSurface) {
		this.somaSurface = somaSurface;
	}
	public double getSurface() {
		return surface;
	}
	public void setSurface(double surface) {
		this.surface = surface;
	}
	public double getVolume() {
		return volume;
	}
	public void setVolume(double volume) {
		this.volume = volume;
	}
	public double getWidth() {
		return width;
	}
	public void setWidth(double width) {
		this.width = width;
	}
	public double getMetricValue(int metricId){
		switch (metricId){
		case N_STEMS:
			return nStems;
		case N_BIFS:
			return nBifs;
		case N_BRANCHES:
			return nBranches;
		case WIDTH:
			return width;
		case HEIGHT:
			return height;
		case DEPTH:
			return depth;
		case DIAMETER:
			return diameter;
		case LENGTH:
			return length;
		case SURFACE:
			return surface;
		case VOLUME:
			return volume;
		case EUCLIDEAN_DIST:
			return euclideanDistance;
		case PATH_DIST:
			return pathDistance;
		case BRANCH_ORDER:
			return branchOrder;
		case CONTRACTION:
			return contraction;
		case FRAGMENTATION:
			return fragmentation;
		case PARTITION_ASYMMETRY:
			return partitionAsymmetry;
		case PK_CLASSIC:
			return pkClassic;
		case BIF_AMPL_LOCAL:
			return bifAmplLocal;
		case BIF_AMPL_REMOTE:
			return bifAmplRemote;
		case FRACTAL_DIM:
			return fractalDim;
		default: 
			return 0;
		}
	}

	public double setMetricValue(int metricId, double value){
		switch (metricId){
		case N_STEMS:
			nStems = (int)value;
		case N_BIFS:
			nBifs = (int)value;
		case N_BRANCHES:
			nBranches = (int)value;
		case WIDTH:
			width = value;
		case HEIGHT:
			height = value;
		case DEPTH:
			depth = value;
		case DIAMETER:
			diameter = value;
		case LENGTH:
			length = value;
		case SURFACE:
			surface = value;
		case VOLUME:
			volume = value;
		case EUCLIDEAN_DIST:
			euclideanDistance = value;
		case PATH_DIST:
			pathDistance = value;
		case BRANCH_ORDER:
			branchOrder = (int)value;
		case CONTRACTION:
			contraction = value;
		case FRAGMENTATION:
			fragmentation = value;
		case PARTITION_ASYMMETRY:
			partitionAsymmetry = value;
		case PK_CLASSIC:
			pkClassic = value;
		case BIF_AMPL_LOCAL:
			bifAmplLocal = value;
		case BIF_AMPL_REMOTE:
			bifAmplRemote = value;
		case FRACTAL_DIM:
			fractalDim = value;
		default: 
			return 0;
		}
	}

}
