package org.cn3.morpho.generation;

import java.math.BigInteger;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Random;
import java.util.Stack;
import java.util.TreeSet;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.Soma.ArborType;
import org.cn3.morpho.utils.BinaryTreeUtils;
import org.cn3.morpho.utils.SystemUtils;

/**
 * 
 * @author gillette
 *
 * Contains multiple methods of generating binary tree structures.
 *
 */
public class BinaryTreeGenerator {
	
	protected static final int MAX_TREE_SIZE = 4095;
	protected static final int MAX_SEQ_STORAGE_DEGREE = 17;
	protected static final int MAX_GLOBAL_SHAPE_STORAGE_DEGREE = 16;
	protected static final int MAX_GLOBAL_CONSTRAINED_SHAPE_STORAGE_DEGREE = 12;
	protected static final int MAX_GLOBAL_PARTITION_STORAGE_DEGREE = 16;

	protected static final boolean RETURN_COPIES = true;

	public static enum TreeDistribution{
		NeuroMorpho("0","NeuroMorpho","NMO","NeuroMorpho.Org"), 
		SequenceGrowth("1","SequenceGrowth"),
		TerminalGrowth("2","TerminalGrowth"),
		Uniform("3","Uniform","TreeShapes"),
		NMOLengthMatchedRandom("4","NMOLengthMatchedRandom"),
		NodeTypeAssembled("5","NodeTypeAssembled","MonomerConstrained"),
		NMOLengthMatchedRandom2("6","NMOLengthMatchedRandom2"),
		NodeTypeAssembled2("7","NodeTypeAssembled2","DimerConstrained"),
		NodeTypeAssembled3("8","NodeTypeAssembled3","TrimerConstrained"),
		SegmentalGrowth("9","SegmentalGrowth");
		
		private final List<String> names;
		TreeDistribution(String ...names){
			this.names = Arrays.asList(names);
		}
		public List<String> getNames() {
	        return names;
	    }
		
		public static TreeDistribution find(String name) {
		    for (TreeDistribution treeDistribution : TreeDistribution.values()) {
		        if (treeDistribution.getNames().contains(name)) {
		            return treeDistribution;
		        }
		    }
		    return null;
		}
	}
	
	public static BinaryTreeNode generateRandomBinaryTree(){
		return generateRandomBinaryTree(.5, .2);
	}

	public static BinaryTreeNode generateRandomBinaryTree(
			double initialProbability, double probabilityReduction){
		BinaryTreeNode head = new BinaryTreeNode();
		head.setData("A");
		BinaryTreeNode node;
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(head);
		double currentProbability;
		while (!stack.isEmpty()){
			node = (BinaryTreeNode)stack.pop();
			currentProbability = initialProbability * 
				Math.pow(probabilityReduction, node.getBranchOrder());
			if (Math.random() < currentProbability){
				node.setData("C");
				node.setLeft(new BinaryTreeNode());
				node.setRight(new BinaryTreeNode());
				node.getLeft().setData("A");
				node.getRight().setData("A");
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return head;
	}

	private static BinaryTreeNode generateTreeStart(int size){
		BinaryTreeNode head = new BinaryTreeNode();
		if (size > 1){
			head.setLeft(new BinaryTreeNode());
			head.setRight(new BinaryTreeNode());
			if (size > 3){
				head.getLeft().setLeft(new BinaryTreeNode());
				head.getLeft().setRight(new BinaryTreeNode());
			}
		}
		return head;
	}
	
	public static BinaryTreeNode growTreeTerminally(
			BinaryTreeNode root, int endSize){
		BinaryTreeNode node;
		List<BinaryTreeNode> tailList = BinaryTreeUtils.createTerminalNodeList(root);
		int pos;
		
		for (int currSize = root.getTreeSize(); currSize < endSize; currSize+=2){			
			pos = (int)(tailList.size() * Math.random());
			node = tailList.remove(pos);
			node.setLeft(new BinaryTreeNode());
			node.setRight(new BinaryTreeNode());
			tailList.add(node.getLeft());
			tailList.add(node.getRight());
		}
		return root;
	}

	public static BinaryTreeNode generateTerminalGrowthBinaryTreeByDegree(
			int degree){
		return generateTerminalGrowthBinaryTreeBySize(2*degree-1);
	}
	public static BinaryTreeNode generateTerminalGrowthBinaryTreeBySize(
			int size){
		BinaryTreeNode node, root = generateTreeStart(size);
		
		if (size > 5){
			List<BinaryTreeNode> tailList = BinaryTreeUtils.createTerminalNodeList(root);
			
			int pos;
			for (int currSize = 5; currSize < size; currSize+=2){
				pos = (int)(tailList.size() * Math.random());
				node = tailList.remove(pos);
				node.setLeft(new BinaryTreeNode());
				node.setRight(new BinaryTreeNode());
				tailList.add(node.getLeft());
				tailList.add(node.getRight());
			}
		}
		return root;
	}
	
	public static BinaryTreeNode growTreeSegmentally(BinaryTreeNode root, int endSize){
		BinaryTreeNode node, newNode;
		List<BinaryTreeNode> nodeList = BinaryTreeUtils.createNodeList(root);
		int pos;
		
		for (int currSize = root.getTreeSize(); currSize < endSize; currSize+=2){
			pos = (int)(nodeList.size() * Math.random());
			node = nodeList.get(pos);
			if (node.isLeaf()){
				node.setLeft(new BinaryTreeNode());
				node.setRight(new BinaryTreeNode());
				nodeList.add(node.getLeft());
				nodeList.add(node.getRight());
			}
			else{
				newNode = new BinaryTreeNode();
				newNode.setRight(new BinaryTreeNode());
				nodeList.add(newNode);
				nodeList.add(newNode.getRight());
				if (node.isRoot()){
					root = newNode;
				}
				else{
					// Need to place newNode between node's parent and itself
					if (node.isLeft()){
						node.getParent().setLeft(newNode);
					}
					else{
						node.getParent().setRight(newNode);
					}
				}
				newNode.setLeft(node);
			}
		}
		return root;
	}
	
	public static BinaryTreeNode generateSegmentalGrowthBinaryTreeByDegree(
			int degree){
		return generateSegmentalGrowthBinaryTreeBySize(2*degree-1);
	}
	public static BinaryTreeNode generateSegmentalGrowthBinaryTreeBySize(
			int size){

		BinaryTreeNode node, newNode, root = generateTreeStart(size);
		
		if (size > 5){
			List<BinaryTreeNode> nodeList = BinaryTreeUtils.createNodeList(root);

			int pos;
			for (int currSize = 5; currSize < size; currSize+=2){
				pos = (int)(nodeList.size() * Math.random());
				node = nodeList.get(pos);
				if (node.isLeaf()){
					node.setLeft(new BinaryTreeNode());
					node.setRight(new BinaryTreeNode());
					nodeList.add(node.getLeft());
					nodeList.add(node.getRight());
				}
				else{
					newNode = new BinaryTreeNode();
					newNode.setRight(new BinaryTreeNode());
					nodeList.add(newNode);
					nodeList.add(newNode.getRight());
					if (node.isRoot()){
						root = newNode;
					}
					else{
						// Need to place newNode between node's parent and itself
						if (node.isLeft()){
							node.getParent().setLeft(newNode,false);
						}
						else{
							node.getParent().setRight(newNode,false);
						}
					}
					newNode.setLeft(node);
				}
			}
		}
		return root;
	}
	
	public static BinaryTreeNode generateIntermediateGrowthBinaryTreeBySize(int size){
		BinaryTreeNode node, newNode, head = generateTreeStart(size);
		
		if (size > 5){
			List<BinaryTreeNode> nodeList = BinaryTreeUtils.createBifurcationList(head);
			
			int pos;
			for (int currSize = 5; currSize < size; currSize+=2){
				pos = (int)(nodeList.size() * Math.random());
				node = nodeList.get(pos);
				newNode = new BinaryTreeNode();
				newNode.setRight(new BinaryTreeNode());
				nodeList.add(newNode);
				if (node.isRoot()){
					head = newNode;
				}
				else{
					// Need to place newNode between node's parent and itself
					if (node.isLeft()){
						node.getParent().setLeft(newNode);
					}
					else{
						node.getParent().setRight(newNode);
					}
				}
				newNode.setLeft(node);
			}
		}
		return head;
	}
	
	public static BinaryTreeNode generateQSBinaryTreeBySize(
			int size, double q, double s){
		
		if (size == 1){
			return new BinaryTreeNode();
		}
		
		// List of current nodes
		List<BinaryTreeNode> nodeList = new LinkedList<BinaryTreeNode>();
		NavigableSet<Double> cumProbSet = new TreeSet<Double>();
		Map<Double,BinaryTreeNode> cumProbNodeMap = new HashMap<Double,BinaryTreeNode>();
		
		// Map between node and relative probability before multiplying by C
		Map<BinaryTreeNode,Double> nodeRelProbMap = new HashMap<BinaryTreeNode,Double>(size);
		BinaryTreeNode node, currNode, newNode, left, right;
		
		// Create head and initial nodes
		BinaryTreeNode head = new BinaryTreeNode();
		head.setRight(new BinaryTreeNode());
		head.setLeft(new BinaryTreeNode());
		
		nodeRelProbMap.put(head, 1.0);
		
		// Initialize probability variables
		double p, pSum, C = 1,// C = n/pSum
		 qInterMult, qInterMinusTerm = 0; // For dealing with case where Q=1, making Q/(1-Q) = Inf
		if (q == 1){ // Intermediate nodes only
			qInterMult = 1;
			nodeList.add(head);
			nodeRelProbMap.put(head, 1.0);
			pSum = 1;
		}else if (q == 0){ // Terminal nodes only
			qInterMult = 0;
			nodeList.add(head.getLeft());
			nodeList.add(head.getRight());
			p = Math.pow(2, -s);
			nodeRelProbMap.put(head.getLeft(), p);
			nodeRelProbMap.put(head.getRight(), p);
			pSum = 2*p;
		}else{ // Mixture of terminal and intermediate
			qInterMult = q/(1-q);
			
			nodeList.add(head);
			pSum = p = qInterMult*1;
			nodeRelProbMap.put(head, p);
			
			nodeList.add(head.getLeft());
			nodeList.add(head.getRight());
			p = Math.pow(2, -s);
			nodeRelProbMap.put(head.getLeft(), p);
			nodeRelProbMap.put(head.getRight(), p);
			pSum += p;
		}
		qInterMinusTerm = qInterMult - 1; // Reduces calculations when updating nodes and probabilities
		
		// Run for "size" steps
		for (int currSize = 3; currSize < size; currSize+=2){
			// Set up cumulative probability to randomly choose segment
			cumProbSet.clear();
			cumProbNodeMap.clear();
			p = 0;
			C = 1.0/pSum;
			cumProbSet.add(0.0);
			for (int i = 0; i < nodeList.size(); i++){
				currNode = nodeList.get(i);
				cumProbNodeMap.put(p,currNode);
				p += C*nodeRelProbMap.get(currNode);
				cumProbSet.add(p);
			}
			// Select node randomly given probabilities
			node = cumProbNodeMap.get(cumProbSet.floor(Math.random()));
			
			if (node.isLeaf()){
				if (q == 0){
					nodeList.remove(node);
					pSum -= nodeRelProbMap.remove(node);
				}else{ // q cannot be 1 since this node is a leaf
					p = nodeRelProbMap.get(node);
					pSum += p*qInterMinusTerm; // Update pSum with new value now that this node is an intermediate
					nodeRelProbMap.put(node, qInterMult*p); // Ignoring division by qTermMult since it equals 1
				}
				
				// Create left node
				left = new BinaryTreeNode();
				node.setLeft(left);
				// Add it to node list
				nodeList.add(left);
				// Calculate relative probability
				p = Math.pow(2, -s*left.getBranchOrder());
				nodeRelProbMap.put(left, p);

				right = new BinaryTreeNode();
				node.setRight(right);
				nodeList.add(right);
				nodeRelProbMap.put(right, p);
				
				pSum += 2*p;
			}
			else{ // Intermediate node
				// Create new node with right subtree as terminal node and left subtree rooted by chosen node
				newNode = new BinaryTreeNode();
				if (node.isRoot()){
					// Reset root
					head = newNode;
				}
				else{
					// Need to place newNode between node's parent and itself
					if (node.isLeft()){
						node.getParent().setLeft(newNode);
					}
					else{
						node.getParent().setRight(newNode);
					}
				}
				newNode.setLeft(node);

				// Create right terminal node
				right = new BinaryTreeNode();
				newNode.setRight(right);

				// Add relevant nodes to nodeList
				nodeList.add(newNode);
				p = nodeRelProbMap.get(node);
				pSum += p;
				nodeRelProbMap.put(newNode, p);
				if (q < 1){
					nodeList.add(right);
					p = Math.pow(2, -s*right.getBranchOrder());
					pSum += p;
					nodeRelProbMap.put(right, p);
				}
				
				// Propagate branch order and base bifurcation probability
				pSum += propagateBranchOrderAndProbability(nodeRelProbMap, node, s, qInterMult);
			}
		}
		return head;
	}
	
	private static double propagateBranchOrderAndProbability(
			Map<BinaryTreeNode,Double> nodeRelProbMap, BinaryTreeNode subtreeRoot, double s, double qInterMult){
		double pSumChange = 0, oldP, newP;

		BinaryTreeNode node;
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(subtreeRoot);
		while (!stack.isEmpty()){
			node = stack.pop();
			if (nodeRelProbMap.containsKey(node)){
				oldP = nodeRelProbMap.get(node);
				newP = Math.pow(2,-s*node.getBranchOrder());
				if (!node.isLeaf()){
					newP *= qInterMult;
				}
				nodeRelProbMap.put(node, newP);
				pSumChange += newP-oldP;
			}
		}
		
		return pSumChange;
	}
	
	public static String generateRandomBinaryTreeString(){
		return generateRandomBinaryTreeStringByLengths(0,0);
	}
	
	public static String generateRandomBinaryTreeString(int length){
		return generateRandomBinaryTreeStringByLengths(length,length);
	}
	
	/**
	 *  All trees have an odd number of nodes, so if 'length' is even, the resulting length will be length + 1
	 */
	public static String generateRandomBinaryTreeStringByLengths(
			int minLength, int maxLength){
		StringBuffer sb;
		int branches, terminations, total, iterations = 0;
		
		if (maxLength > MAX_TREE_SIZE){
			System.out.println("WARNING: maxLength "+maxLength+" can be no larger than "
					+MAX_TREE_SIZE+". Reducing maxLength to limit.");
			maxLength = MAX_TREE_SIZE;
		}
		// make length odd
		//System.out.println("min: "+minLength+"; max: "+maxLength); 
		if (maxLength > 0 && maxLength % 2 == 0){
			maxLength++;
		}
		else if (maxLength == 0){
			maxLength = MAX_TREE_SIZE;
		}
		do{
			sb = new StringBuffer();
			branches = total = 1;
			terminations = 0;
			sb.append("C");
			while ((total < maxLength || maxLength == 0)
					&& (branches >= terminations)){
				if (Math.random() >= .5){
					sb.append("C");
					total++;
					branches++;
				}
				else{
					sb.append("A");
					total++;
					terminations++;
				}
			}
			if (branches == terminations){
				sb.append("A");
				total++;
				terminations++;
			}
			iterations++;
			if (iterations > 10){
				System.gc();
				iterations = 0;
			}
		} while (branches + 1 != terminations
				|| total < minLength 
					|| (maxLength > 0 && total > maxLength));
		return sb.toString();
	}

	/*
	public static int debugLevel = 1;
	private static void debug(String string, int level){
		if (level <= debugLevel){
			System.out.println(string);
		}
	}*/

	public static BinaryTreeNode generateRandomHardingTree(long size)
	throws Exception{
		long degree = (size - 1)/2 + 1;
		return generateRandomHardingTreeByDegree(degree);
	}

	public static BinaryTreeNode generateRandomHardingTreeByDegree(long degree)
	throws Exception{
		BigInteger possibleShapes = BinaryTreeUtils.numShapes(degree);
		BigInteger shapeNum = randomBigInteger(possibleShapes);
		
		return generateHardingTree(degree,shapeNum);
	}

	public static BinaryTreeNode generateRandomHardingTreeByDegree(long degree, boolean returnCopies)
	throws Exception{
		BigInteger possibleShapes = BinaryTreeUtils.numShapes(degree);
		BigInteger shapeNum = randomBigInteger(possibleShapes);
		
		return generateHardingTree(degree,shapeNum,returnCopies);
	}

	private static Map<ShapeKey,BinaryTreeNode> shapeMap = new HashMap<ShapeKey,BinaryTreeNode>();
	private static class ShapeKey{
		private long degree;
		private BigInteger shapeNum;
		public ShapeKey(long degree, BigInteger shapeNum){
			this.degree = degree;
			this.shapeNum = shapeNum;
		}
		public String toString(){
			return degree+" - "+shapeNum;
		}
		public int hashCode(){
			return (int)(shapeNum.hashCode()+(degree<<4));
		}
		public boolean equals(Object obj){
			if (obj.getClass().equals(this.getClass())){
				ShapeKey shapeKey = (ShapeKey)obj;
				return (degree==shapeKey.getDegree()
						&& shapeNum.equals(shapeKey.getShapeNum()));
			}
			return false;
		}
		public long getDegree() {
			return degree;
		}
		public BigInteger getShapeNum() {
			return shapeNum;
		}
	}
	/**
	 * Produces right side heavy tree
	 * leftDegree != rightDegree
	 *  shape = partialSum_degree(degree-rightDegree-1) + (rightShape-1)*shapes(leftDegree) + leftShape
	 * leftDegree == rightDegree
	 *  shape = partialSum_degree(degree-rightDegree-1) + 1/2 * rightShape*(rightShape-1) + leftShape
	 */
	public static BinaryTreeNode generateHardingTree(
			long degree, BigInteger shapeNum) throws Exception{
		return generateHardingTree(degree, shapeNum, RETURN_COPIES);
	}
	public static BinaryTreeNode generateHardingTree(
			long degree, BigInteger shapeNum, boolean returnCopies) throws Exception{
		BinaryTreeNode head = new BinaryTreeNode();
		head.setShapeNum(shapeNum);
		head.setData(new Long(degree));
		if (degree == 1){
			return head;
		}
		else if (degree == 2){
			head.setLeft(new BinaryTreeNode());
			head.setRight(new BinaryTreeNode());
			return head;
		}
		else{
			// If this shape has been generated, take a short-cut and copy rather than recalculating
			ShapeKey shapeKey = new ShapeKey(degree, shapeNum);
			if (degree <= MAX_GLOBAL_SHAPE_STORAGE_DEGREE && shapeMap.containsKey(shapeKey)){
				if (returnCopies){
					return BinaryTreeUtils.copyTree(shapeMap.get(shapeKey));
				}
				else{
					return shapeMap.get(shapeKey);
				}
			}
			
			//System.out.println("Degree: "+degree+"; shape: "+shapeNum);
			BigInteger possibleShapes = BinaryTreeUtils.numShapes(degree);
			if (shapeNum.compareTo(possibleShapes) > 0){
				throw new Exception("Invalid Shape number: "+shapeNum
						+"\nMax shape is "+possibleShapes+" for degree "+degree);
			}
			long leftDegree = 0;
			BigInteger partialSum;
			do{
				leftDegree++;
				partialSum = BinaryTreeUtils.partialSumOfNumShapes(degree,leftDegree);
				//System.out.println("PartialSum("+degree+","+leftDegree+") = "+partialSum);
			} while(partialSum.compareTo(shapeNum) < 0);
			long rightDegree = degree - leftDegree;
			
			BigInteger remainder, numShapesLeft, rightShape, leftShape;
			remainder =
				shapeNum.subtract(BinaryTreeUtils.partialSumOfNumShapes(degree,leftDegree-1));

			//System.out.println("Left Degree: "+leftDegree+"; Right Degree: "+rightDegree+"; Remainder: "+remainder);

			if (rightDegree == leftDegree){
				numShapesLeft = BinaryTreeUtils.numShapes(leftDegree);
				rightShape = numShapesLeft;
				leftShape = rightShape.add(BigInteger.ONE);
				BigInteger two = BigInteger.ONE.add(BigInteger.ONE);
				// Find the shape out of n*(n+1)/2
				rightShape = findRightShape(remainder,BigInteger.ONE,numShapesLeft);
				leftShape = remainder.subtract(
						rightShape.multiply(rightShape.subtract(BigInteger.ONE))
							.divide(two));
			}
			else{
				numShapesLeft = BinaryTreeUtils.numShapes(leftDegree);
				rightShape = remainder.subtract(BigInteger.ONE).divide(numShapesLeft).add(BigInteger.ONE);
				leftShape = remainder.subtract(numShapesLeft.multiply(rightShape.subtract(BigInteger.ONE)));
			}
			//System.out.println("RightShape: "+rightShape+"; LeftShape: "+leftShape);
			
			head.setRight(generateHardingTree(rightDegree,rightShape,returnCopies));
			head.setLeft(generateHardingTree(leftDegree,leftShape,returnCopies));
			
			if (degree <= MAX_GLOBAL_SHAPE_STORAGE_DEGREE){
				if (returnCopies){
					shapeMap.put(shapeKey, BinaryTreeUtils.copyTree(head));
				}
				else{
					shapeMap.put(shapeKey, head);
				}
			}
			return head;
		}
	}
	
	public static BigInteger findRightShape(
			BigInteger target, BigInteger low, BigInteger high){
		BigInteger two = BigInteger.ONE.add(BigInteger.ONE);
		BigInteger middle = high.subtract(low).divide(two).add(low);
		BigInteger middleValLow = middle.multiply(middle.add(BigInteger.ONE)).divide(two);
		BigInteger middleValHigh = middleValLow.add(middle.add(BigInteger.ONE));
		
		int compareLow = target.compareTo(middleValLow);
		if (compareLow == 0){
			return middle;
		}
		else if (compareLow < 0){
			return findRightShape(target, low, middle);
		}
		else{
			int compareHigh = target.compareTo(middleValHigh);
			if (compareHigh <= 0){
				return middle.add(BigInteger.ONE);
			}
			else {
				return findRightShape(target, middle, high);
			}
		}
	}
	
	// Produces value from 1 to range
	public static BigInteger randomBigInteger(BigInteger range){
		BigInteger random;
		Random rand = SystemUtils.getRandom();
		int count = 0;
		do{
			random = new BigInteger(range.bitLength(),rand).add(BigInteger.ONE);
			count++;
		} while (range.compareTo(random) < 0 || count == 1);
		return random;
	}

	public static void bifsToAchieveE() throws Exception{
		int degree, numBifsStart = 8, numBifs, numComparisons;
		double pVal, eVal, numCompsDouble;
		BigInteger numShapes;
		NumberFormat nf = new DecimalFormat("0.0E0");
		List<Integer> numComparisonsList = new ArrayList<Integer>(),
			minBifs05 = new ArrayList<Integer>(), 
			minBifs01 = new ArrayList<Integer>(), 
			minBifs001 = new ArrayList<Integer>();
		//for (int numComparisons = 1; numComparisons < 100000; numComparisons*=10){
		for (numCompsDouble = 1; numCompsDouble < 100001; numCompsDouble*=3.162278){
			numComparisons = (int)Math.round(numCompsDouble);
			numComparisonsList.add(numComparisons);
			numBifs = numBifsStart;
			while (minBifs001.size() < numComparisonsList.size()){
				degree = numBifs+1;
				numShapes = BinaryTreeUtils.numShapes(degree);
				pVal = BigInteger.ONE.doubleValue()/numShapes.doubleValue();
				eVal = -numComparisons*Math.log(1-pVal);
				if (minBifs05.size() < numComparisonsList.size() && eVal <= .05){
					minBifs05.add(numBifs);
					numBifsStart = numBifs;
				}
				if (minBifs01.size() < numComparisonsList.size() && eVal <= .01){
					minBifs01.add(numBifs);
				}
				if (minBifs001.size() < numComparisonsList.size() && eVal <= .001){
					minBifs001.add(numBifs);
				}
				System.out.println(numBifs+"\t"+numShapes+"\t"+nf.format(pVal)+"\t"+nf.format(eVal));
				numBifs++;
			}
		}
		StringBuffer sb = new StringBuffer();
		sb.append("NumComparisons,0.05,0.01,0.001\n");
		for (int i = 0; i < numComparisonsList.size(); i++){
			sb.append(numComparisonsList.get(i)+","+minBifs05.get(i)+","+minBifs01.get(i)+","+minBifs001.get(i)+"\n");	
		}
		DataWriter.writeStringToFile("C:/Documents and Settings/gillette/My Documents/research/BifsToAchieveE.csv", sb);
	}
	
	private static final double LOG2 = Math.log(2.0);
	private static final double LOG10 = Math.log(10.0);
	public static double log10BigInteger(BigInteger val) {
		return logBigInteger(val)/LOG10;
	}
	public static double logBigInteger(BigInteger val) {
	    int blex = val.bitLength() - 1022; // any value in 60..1023 works
	    if (blex > 0)
	        val = val.shiftRight(blex);
	    double res = Math.log(val.doubleValue());
	    return blex > 0 ? res + blex * LOG2 : res;
	}
	public static void numShapesChart(int minNumBifurcations, int maxNumBifurcation) throws Exception{
		int degree;
		BigInteger numShapes;
		for (int numBifs = minNumBifurcations; numBifs <= maxNumBifurcation; numBifs++){
			degree = numBifs+1;
			numShapes = BinaryTreeUtils.numShapes(degree);
			System.out.println(numBifs+","+numShapes+","+log10BigInteger(numShapes));
		}
		
	}
	
	public static void main(String args[]){
		try{
			numShapesChart(1,200);
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
}
