package org.cn3.morpho.utils;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;

import org.cn3.morpho.domain.Asymmetry;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.morpho.domain.SwcTreeNode;


/**
 * @author gillette
 *
 */
public class BinaryTreeUtils {

	public static final int METRIC_NODES = 1;
	public static final int METRIC_EUCLIDEAN_DISTANCE = 2;
	public static final int METRIC_PATH_LENGTH = 3;
	public static final int METRIC_SURFACE_AREA = 4;
	public static final int METRIC_VOLUME = 5;
	public static final int METRIC_DIAMETER = 6;
	public static final int METRIC_TORTUOSITY = 7;
	public static final int METRIC_ANGLE_OF_BIFURCATION = 8;
	public static final int METRIC_ANGLE_OF_ELEVATION = 9;

	private static Map<Long,BigInteger> numShapesMap = new HashMap<Long,BigInteger>();
	public static BigInteger numShapes(long degree) throws Exception{
		return numShapes(degree, numShapesMap);
	}
	public static BigInteger numShapes(long degree, Map<Long,BigInteger> numShapesMap) throws Exception{
		Long key = new Long(degree);
		synchronized(numShapesMap){
			if (numShapesMap.containsKey(key)){
				return (BigInteger)numShapesMap.get(key);
			}
		}
		if (degree < 1){
			throw new Exception("Degree must be greater than zero. Found "+degree);
		}
		if (degree == 1 || degree == 2){
			return BigInteger.ONE;
		}
		BigInteger numShapes = BigInteger.ZERO, two = BigInteger.ONE.add(BigInteger.ONE);
		BigInteger tmp, tmp1, tmp2;
		long m;
        if (degree%2 == 0){
            // Half Sum when degree is even: numShapes + (1/2) * numShapes(degree/2) * (numShapes(degree/2)+1);
        	m = degree / 2;
            tmp = numShapes(m,numShapesMap);
            numShapes = numShapes.add(
            		tmp.multiply(tmp.add(BigInteger.ONE))
            			.divide(two)
            	);
        }
        else{
            m = (degree+1)/2;
        }

        for (int i = 1; i < m; i++){
        	// numShapes = SUM(numShapes(i) * numShapes(degree-i))
        	tmp1 = numShapes(i,numShapesMap);
            tmp2 = numShapes(degree-i,numShapesMap);
            if (tmp1.compareTo(tmp2) == 0){
            	numShapes = numShapes.add(
            		tmp1.multiply(tmp1.add(BigInteger.ONE)).divide(two));
            }
            else{
            	numShapes = numShapes.add(tmp1.multiply(tmp2));
            }
        }
		synchronized(numShapesMap){
			numShapesMap.put(key, numShapes);
		}
		return numShapes;
	}

	// REFERENCE Harding 1971: The probabilities of rooted tree-shapes generated by random bifurcation
	public static BigInteger partialSumOfNumShapes(
			long degree, long parts) 
	throws Exception{
		if (parts == 0){
			return BigInteger.ZERO;
		}
		if (parts > (degree/2)){
			throw new Exception("Parts should be no larger than degree/2. "+
					"Degree: "+degree+"; Parts: "+parts);
		}
		if (parts == degree - parts){
			return numShapes(degree);
		}
		BigInteger s = BigInteger.ZERO;
		for (int i = 1; i <= parts; i++){
			// S_degree(parts) = ... + S_degree(i)*S_degree(degree-i) ...
			s = s.add(BinaryTreeUtils.numShapes(i).multiply(BinaryTreeUtils.numShapes(degree-i)));
		}
		return s;
	}
	private static Map<Long,BigInteger> numTerminalGrowthTreesMap = new HashMap<Long,BigInteger>();
	public static BigInteger numTerminalGrowthTrees(long degree){
		return numTerminalGrowthTrees(degree, numTerminalGrowthTreesMap);
	}
	public static BigInteger numTerminalGrowthTrees(long degree, Map<Long,BigInteger> numTerminalGrowthTreesMap){
		Long key = new Long(degree);
		if (numTerminalGrowthTreesMap.containsKey(key)){
			return (BigInteger)numTerminalGrowthTreesMap.get(key);
		}
		if (degree == 1 || degree == 2){
			return BigInteger.ONE;
		}
		BigInteger numTrees = BigInteger.ZERO, two = BigInteger.ONE.add(BigInteger.ONE);
		BigInteger tmp, tmp1, tmp2;
		long m;
        if (degree%2 == 0){
            // Half Sum when degree is even: numTrees + (1/2) * numTrees(degree/2) * (numTrees(degree/2)+1);
        	m = degree / 2;
            tmp = numTerminalGrowthTrees(m,numTerminalGrowthTreesMap);
            numTrees = numTrees.add(
            		tmp.multiply(tmp.add(BigInteger.ONE))
            			.divide(two)
            	);
        }
        else{
            m = (degree+1)/2;
        }

        for (int i = 1; i < m; i++){
        	// numTrees = SUM(numTrees(i) * numTrees(degree-i))
        	tmp1 = numTerminalGrowthTrees(i,numTerminalGrowthTreesMap);
            tmp2 = numTerminalGrowthTrees(degree-i,numTerminalGrowthTreesMap);
            if (tmp1.compareTo(tmp2) == 0){
            	numTrees = numTrees.add(
            		tmp1.multiply(tmp1.add(BigInteger.ONE)).divide(two));
            }
            else{
            	numTrees = numTrees.add(tmp1.multiply(tmp2));
            }
        }
        numTerminalGrowthTreesMap.put(key, numTrees);
		return numTrees;
	}
	
	public static Stack<BinaryTreeNode> getLeavesToRootStack(BinaryTreeNode tree){
		return getLeavesToRootStack(tree, false);
	}
	public static Stack<BinaryTreeNode> getLeavesToRootStack(BinaryTreeNode tree, boolean mustHaveChildren){
		Stack<BinaryTreeNode> initStack = new Stack<BinaryTreeNode>(), 
			stack = new Stack<BinaryTreeNode>();
		BinaryTreeNode node;
		initStack.push(tree);
		while (!initStack.isEmpty()){
			node = initStack.pop();
			if (!mustHaveChildren || node.hasChildren()){
				stack.push(node);
			}
			if (node.hasChildren()){
				initStack.push(node.getLeft());
				initStack.push(node.getRight());
			}
		}
		return stack;
	}

	public static Set<BinaryTreeNode> getAllNodesSet(BinaryTreeNode tree){
		Stack<BinaryTreeNode> initStack = new Stack<BinaryTreeNode>();
		Set<BinaryTreeNode> allNodes = new HashSet<BinaryTreeNode>();
		BinaryTreeNode node;
		initStack.push(tree);
		while (!initStack.isEmpty()){
			node = (BinaryTreeNode)initStack.pop();
			allNodes.add(node);
			if (node.hasChildren()){
				initStack.push(node.getLeft());
				initStack.push(node.getRight());
			}
		}
		return allNodes;
	}

	public static BinaryTreeNode copyTree(BinaryTreeNode tree){
		return copyTree(tree,false);
	}
	public static BinaryTreeNode copyTree(
			BinaryTreeNode tree, boolean keepData){
		BinaryTreeNode treeCopy = tree.copy(keepData), node, copyNode;
		Queue<BinaryTreeNode> queue = new LinkedList<BinaryTreeNode>();
		Queue<BinaryTreeNode> copyQueue = new LinkedList<BinaryTreeNode>();
		queue.offer(tree);
		copyQueue.offer(treeCopy);
		while (!queue.isEmpty()){
			node = queue.remove();
			copyNode = copyQueue.remove();
			if (node.hasChildren()){
				copyNode.setRight(node.getRight().copy(), false);
				copyNode.getRight().setData(node.getRight().getData());
				copyNode.setLeft(node.getLeft().copy(), false);
				copyNode.getLeft().setData(node.getLeft().getData());
				queue.offer(node.getLeft());
				queue.offer(node.getRight());
				copyQueue.offer(copyNode.getLeft());
				copyQueue.offer(copyNode.getRight());
			}
		}
		BinaryTreeNode.updateStatisticsDown(treeCopy);
		return treeCopy;
	}
	
	public static double getExcessAsymmetry(BinaryTreeNode tree, int metric)
	throws Exception{
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(tree);
		BinaryTreeNode node;
		double totalAsym = 0;
		int nodes = 0;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.getRight().hasChildren() && node.getLeft().hasChildren()){
				nodes++;
				totalAsym += getExcessPartitionAsymmetry(node, metric);
			}
			if (node.getLeft().hasChildren()){
				stack.push(node.getLeft());
			}
			if (node.getRight().hasChildren()){
				stack.push(node.getRight());
			}
		}
		if (nodes == 0){
			return -1;
		}
		return totalAsym/nodes;
	}

	public static double getExcessPartitionAsymmetry(BinaryTreeNode node, int metric)
	throws Exception{
		double a = 0, b = 0, c = 0, d = 0;
		double excessPartitionAsymmetry = 0;
		Asymmetry asymLeft = node.getLeft().getAsymmetry(metric);
		Asymmetry asymRight = node.getRight().getAsymmetry(metric);
		if (node.getLeft().hasChildren()){
			a = asymLeft.getLargerSize();
			b = asymLeft.getSmallerSize();
		}
		else{
			a = 0;
			b = 0;
		}
		if (node.getRight().hasChildren()){
			c = asymRight.getLargerSize();
			d = asymRight.getSmallerSize();
		}
		else{
			c = 0;
			d = 0;
		}
		// Ap = |a+b-c-d|/(a+b+c+d); Ap' = |a-b+c-d|/(a+b+c+d);  Ap'' = |a-b-c+d|/(a+b+c+d);
		double denom = a+b+c+d;
		// For nodal all are divided by (a+b+c+d-2)
		if (metric == METRIC_NODES){
			denom -= 2;
		}
		double Ap = Math.abs(a+b-c-d)/denom;
		double Ap2 = Math.abs(a-b+c-d)/denom;
		double Ap3 = Math.abs(a-b-c+d)/denom;
		// Ep = Ap - <Ap|alpha> = (2Ap - A'p - A''p) / 3
		excessPartitionAsymmetry = 
			(2 * Ap - Ap2 - Ap3) / 3;
		return excessPartitionAsymmetry;
	}

	public static DistributionBean getBranchPathLengthDistribution(BinaryTreeNode tree){
		return getBranchMetricDistribution(tree, METRIC_PATH_LENGTH);
	}
	
	public static void appendBifurcationsToTerminations(BinaryTreeNode tree){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		BinaryTreeNode node;
		stack.push(tree);
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.isLeaf()){
				node.setLeft(new BinaryTreeNode());
				node.setRight(new BinaryTreeNode());
			}
			else{
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
	}

	/**
	 * 
	 * @param node
	 * @param metricId
	 * @return The metric value for the selected node. This function acts as a multiplexer, selecting the appropriate metric based on the metricId.
	 */
	public static double getNodeMeasure(BinaryTreeNode node, int metricId){
		switch (metricId){
		case METRIC_PATH_LENGTH:
			return node.getSwcData().getSecondaryData().getPathLength();
		case METRIC_ANGLE_OF_BIFURCATION:
			return node.getSwcData().getSecondaryData().getLocalAngleOfBifurcation();
		case METRIC_ANGLE_OF_ELEVATION:
			return node.getSwcData().getSecondaryData().getLocalAngleOfElevation();
		case METRIC_DIAMETER:
			return node.getSwcData().getDiameter();
		case METRIC_EUCLIDEAN_DISTANCE:
			return node.getSwcData().getSecondaryData().getEuclideanDistance();
		case METRIC_NODES:
			return 1;
		case METRIC_SURFACE_AREA:
			return node.getSwcData().getSecondaryData().getSurfaceArea();
		case METRIC_TORTUOSITY:
			return node.getSwcData().getSecondaryData().getTortuosity();
		case METRIC_VOLUME:
			return node.getSwcData().getSecondaryData().getVolume();
		default:
			return 0;
		}
	}

	/**
	 * 
	 * @param root
	 * @param metricId
	 * @return The sum over all branches within the tree of the given metric
	 */
	public static double getTreeMetricSummed(BinaryTreeNode root, int metricId){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		double sum = 0;
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			sum += getNodeMeasure(node, metricId);
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return sum;
	}

	/**
	 * 
	 * @param tree
	 * @param metricId
	 * @param weightMetric
	 * @param subtreeWeight
	 * @return Weighted average of the given metric over branches in the tree. Weighting is down by the weight metric either on
	 * the branch's metric value or on the subtree's metric value (summed metric over all branches within tree).
	 */
	public static double getGlobalBranchMetric(BinaryTreeNode tree, int metricId, int weightMetric, boolean subtreeWeight){
		double sum = 0, weight, totalWeight = 0;
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(tree);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (subtreeWeight){
				weight = getTreeMetricSummed(node, weightMetric);
			}
			else{
				weight = getNodeMeasure(node, weightMetric);
			}
			totalWeight += weight;
			sum += weight * getNodeMeasure(node, metricId);
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return sum/totalWeight;
	}

	/**
	 * 
	 * @param tree
	 * @param metricId
	 * @return The distribution of metric values across the branches within the tree
	 */
	public static DistributionBean getBranchMetricDistribution(BinaryTreeNode tree, int metricId){
		DistributionBean distribution = new DistributionBean();
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(tree);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			switch (metricId){
			case METRIC_PATH_LENGTH:
				distribution.addData(node.getSwcData().getSecondaryData().getPathLength());
				break;
			case METRIC_ANGLE_OF_BIFURCATION:
				distribution.addData(node.getSwcData().getSecondaryData().getLocalAngleOfBifurcation());
				break;
			case METRIC_ANGLE_OF_ELEVATION:
				distribution.addData(node.getSwcData().getSecondaryData().getLocalAngleOfElevation());
				break;
			case METRIC_DIAMETER:
				distribution.addData(node.getSwcData().getDiameter());
				break;
			case METRIC_EUCLIDEAN_DISTANCE:
				distribution.addData(node.getSwcData().getSecondaryData().getEuclideanDistance());
				break;
			case METRIC_NODES:
				distribution.addData(1);
				break;
			case METRIC_SURFACE_AREA:
				distribution.addData(node.getSwcData().getSecondaryData().getSurfaceArea());
				break;
			case METRIC_TORTUOSITY:
				distribution.addData(node.getSwcData().getSecondaryData().getTortuosity());
				break;
			case METRIC_VOLUME:
				distribution.addData(node.getSwcData().getSecondaryData().getVolume());
				break;
			}
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return distribution;
	}

	public static int numTwoDegreeNodes(BinaryTreeNode tree){
		if (tree.getDegree() == 2){
			return 1;
		}
		int count = 0;
		if (tree.getLeft().hasChildren()){
			count += numTwoDegreeNodes(tree.getLeft());
		}
		if (tree.getRight().hasChildren()){
			count += numTwoDegreeNodes(tree.getRight());
		}
		return count;
	}
	
	/**
	 * @param root
	 * @return
	 * 	Set of nodes under and including root
	 */
	public static Set<BinaryTreeNode> createNodeSet(BinaryTreeNode root){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		Set<BinaryTreeNode> set = new HashSet<BinaryTreeNode>();
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			set.add(node);
			if (node.hasChildren()){
				stack.push(node.getRight());
				stack.push(node.getLeft());
			}
		}
		return set;
	}

	public static List<BinaryTreeNode> createNodeList(BinaryTreeNode root){
		return createNodeList(root, false);
	}

	public static List<BinaryTreeNode> createBifurcationList(BinaryTreeNode root){
		return createNodeList(root, true);
	}

	public static List<BinaryTreeNode> createTerminalBifurcationList(BinaryTreeNode root){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		List<BinaryTreeNode> list = new ArrayList<BinaryTreeNode>();
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.getDegree() == 2){
				list.add(node);
			}
			else if (node.getDegree() > 2){ // Just in case the root in a terminal (no bifurcations)
				stack.push(node.getRight());
				stack.push(node.getLeft());
			}
		}
		return list;
	}

	public static List<BinaryTreeNode> createTerminalNodeList(BinaryTreeNode root){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		List<BinaryTreeNode> list = new ArrayList<BinaryTreeNode>();
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (!node.hasChildren()){
				list.add(node);
			}
			else {
				stack.push(node.getRight());
				stack.push(node.getLeft());
			}
		}
		return list;
	}

	/**
	 * @param root
	 * @param onlyBifurcations
	 * @return
	 * 	List of nodes under and including root
	 */
	public static List<BinaryTreeNode> createNodeList(BinaryTreeNode root, boolean onlyBifurcations){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		List<BinaryTreeNode> list = new ArrayList<BinaryTreeNode>();
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (!onlyBifurcations || node.hasChildren()){
				list.add(node);
			}
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return list;
	}
	public static List<BinaryTreeNode> createNodeListRightFirst(BinaryTreeNode root, boolean onlyBifurcations){
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		List<BinaryTreeNode> list = new ArrayList<BinaryTreeNode>();
		stack.push(root);
		BinaryTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (!onlyBifurcations || node.hasChildren()){
				list.add(node);
			}
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
		}
		return list;
	}

	public static void main(String argv[]){
		try{
			for (int bifs = 1; bifs <= 150; bifs++){
				BigInteger numShapes = BinaryTreeUtils.numShapes(bifs+1);
				System.out.println(bifs+","+numShapes);
			}
			//System.out.println(numShapes.toString().length());
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	public static List<SwcTreeNode> createTerminalNodeList(SwcTreeNode root){
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		List<SwcTreeNode> list = new ArrayList<SwcTreeNode>();
		stack.push(root);
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (!node.hasChildren()){
				list.add(node);
			}
			else {
				stack.addAll(node.getChildren());
			}
		}
		return list;
	}
	
	public static Map<Integer,SwcTreeNode> createNodeIdMap(SwcTreeNode root){
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		Map<Integer,SwcTreeNode> bifurcationIdMap = new HashMap<Integer,SwcTreeNode>();
		stack.push(root);
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.hasChildren()){
				if (node.getChildren().size() > 1){
					bifurcationIdMap.put(node.getNodeId(), node);
				}
				stack.addAll(node.getChildren());
			}
			else{
				bifurcationIdMap.put(node.getNodeId(), node);
			}
		}
		return bifurcationIdMap;
	}

}
