package org.cn3.ns.analysis;

import java.io.File;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.ns.data.DataConverter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.nodeencoding.NodeEncoder;
import org.cn3.ns.nodeencoding.Topological1stOrderEncoder;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * The purpose of this class is to determine how similar or different subtrees
 * of similar size are within the dataset. Subtrees of similar size but distinct
 * topology are more likely to result in suboptimal alignment.
 * 
 * Produced data referenced in discussion section of Gillette, Hosseini, and Ascoli, 2015
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.analysis.SubtreeSizesAndTopologyComparer -D NM_SpurLen2D -T 6 -o /home/tgillett/results/
 * 
 */
public class SubtreeSizesAndTopologyComparer {

	private static final String PARAM_FASTA_FILE = "fasta";
	private static final String PARAM_THRESHOLD = "threshold";
	private static final String PARAM_OUTPUT_DIR = "output-dir";

	private static final int DEFAULT_THRESHOLD = 6;
	private static final int MIN_NUM_BIFURCATIONS = 10;
	
	private static float threshold = 0;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			String fastaFilename = config.getString(PARAM_FASTA_FILE);
			String dataset = config.getString(NSConstants.PARAM_DATASET);
			if (fastaFilename.equals("")){
				fastaFilename = DataProperties.getInstance().getFastaArborDir(dataset)+"/Mapping2/Size/"+dataset+"ArborSizeSTLMap2.fasta";
			}else if (!config.userSpecified(NSConstants.PARAM_DATASET)){
				File fastaFile = new File(fastaFilename);
				dataset = fastaFile.getName().replaceAll(".fasta", "");
			}
			
			threshold = config.getInt(PARAM_THRESHOLD);
			
			// Detect sufficiently large trees with sufficient symmetry 
			//   in absolute terms of differences in number of branches.
			FastaReader fr = new FastaReader(fastaFilename);
			BinaryTreeNode tree, node, smaller, larger;
			Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
			Set<String> foundTrees = new HashSet<String>();
			StringBuffer dataSb = new StringBuffer();
			dataSb.append("Neurite,SubtreeSize,LargerSideSize,SmallerSideSize,SizeDifference,AsymmetryLS,AsymmetrySS,AsymmetryDiff,CaulescenceLS,CaulescenceSS,CaulescenceDiff\n");
			
			StringBuffer fastaSbA = new StringBuffer(), fastaSbB = new StringBuffer();
			int subtreeCount, rightNumBifs, leftNumBifs, largerNumBifs, smallerNumBifs;
			NumberFormat nf = NumberFormat.getNumberInstance();
			nf.setMaximumFractionDigits(3);
			nf.setGroupingUsed(false);
			
			NodeEncoder encoder = new Topological1stOrderEncoder();
			Map<String,Long> totalCounts = new HashMap<String,Long>();
			totalCounts.put("ApicalDendrite", 0L);
			totalCounts.put("Axon", 0L);
			totalCounts.put("Dendrite", 0L);
			totalCounts.put("BasalDendrite", 0L);
			long totalAxonCount = 0, totalDendriteCount = 0, totalApicalCount = 0;
			
			while (fr.next()){
				subtreeCount = 1;
				// Convert sequence into simple binary tree
				tree = DataConverter.fastaACTStringToTree(fr.getSequence());
				stack.push(tree);
				while (!stack.isEmpty()){
					node = stack.pop();
					rightNumBifs = node.getRight().getNumberOfBifurcations();
					leftNumBifs = node.getLeft().getNumberOfBifurcations();
					if (node.getRight().hasChildren() && rightNumBifs >= MIN_NUM_BIFURCATIONS){
						stack.push(node.getRight());
					}
					if (node.getLeft().hasChildren() && leftNumBifs >= MIN_NUM_BIFURCATIONS){
						stack.push(node.getLeft());
					}
					if (rightNumBifs > leftNumBifs){
						larger = node.getRight();
						smaller = node.getLeft();
					}else{
						larger = node.getLeft();
						smaller = node.getRight();
					}
					largerNumBifs = larger.getNumberOfBifurcations();
					smallerNumBifs = smaller.getNumberOfBifurcations();
					/*
					if (Math.abs(rightNumBifs - leftNumBifs) <= threshold){
						foundTrees.add(fr.getSequenceName());
						// Add data line
						dataSb.append(fr.getSequenceName()+","+node.getNumberOfBifurcations()+","+
								largerNumBifs+","+smallerNumBifs+","+Math.abs(leftNumBifs-rightNumBifs)+","+
								// Asymmetry
								larger.getTreeAsymmetry()+","+smaller.getTreeAsymmetry()+","+Math.abs(larger.getTreeAsymmetry()-smaller.getTreeAsymmetry())+","+
								// Caulescence
								larger.getAsymmetry().getCaulescence()+","+smaller.getAsymmetry().getCaulescence()+
								","+Math.abs(larger.getAsymmetry().getCaulescence()-smaller.getAsymmetry().getCaulescence())+"\n");
						
						fastaSbA.append("> "+fr.getSequenceName()+"_"+subtreeCount+"A\n");
						fastaSbA.append(encoder.encodeTree(larger, "STL")+"\n");
						fastaSbB.append("> "+fr.getSequenceName()+"_"+subtreeCount+"B\n");
						fastaSbB.append(encoder.encodeTree(smaller, "STL")+"\n");
						subtreeCount++;
						//System.out.println(fr.getSequenceName()+"; Total Degree: "+tree.getDegree()+"; Equal degree: "+node.getRight().getDegree());
					}*/
					totalCounts.put(fr.getArborType(), 1+totalCounts.get(fr.getArborType()));
				}
				while (!stack.isEmpty()){
					stack.pop();
				}
			}
			
			System.out.println("Axon subtrees: "+totalCounts.get("Axon"));
			System.out.println("Apical subtrees: "+totalCounts.get("ApicalDendrite"));
			System.out.println("Basal subtrees: "+totalCounts.get("BasalDendrite"));
			System.out.println("Dendrite subtrees: "+totalCounts.get("Dendrite"));
			
			String outputDir = config.getString(PARAM_OUTPUT_DIR);
			//DataWriter.writeStringToFile(outputDir+"SubtreeSizeData_"+dataset+"_Threshold"+threshold+".csv", dataSb.toString());
			//DataWriter.writeStringToFile(outputDir+"SubtreeSizeFastaA_"+dataset+"_Threshold"+threshold+".fasta", fastaSbA.toString());
			//DataWriter.writeStringToFile(outputDir+"SubtreeSizeFastaB_"+dataset+"_Threshold"+threshold+".fasta", fastaSbB.toString());
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(PARAM_FASTA_FILE)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("")
        .setShortFlag('f') 
        .setLongFlag(PARAM_FASTA_FILE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DEFAULT_DATASET)
        .setShortFlag('D') 
        .setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_THRESHOLD)
            .setStringParser(JSAP.INTEGER_PARSER)
            .setDefault(""+DEFAULT_THRESHOLD)
            .setRequired(false)
            .setShortFlag('T') 
            .setLongFlag(PARAM_THRESHOLD);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT_DIR)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag('o') 
        .setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SubtreeSizesAndTopologyComparer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	
}
