package org.cn3.ns.analysis;

import java.text.NumberFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.utils.ClassificationUtils;
import org.cn3.ns.data.DataConverter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 *	Takes average over all nodes of the distance of one data structure from the node of interest
 *  divided by the distance in the other data structure. For each node, samples from nearby
 *  nodes of distance 2 and averages. Can go from tree to sequence and vice versa.
 *  
 *  Main function calculates over all arborizations and produces distribution data.
 *
 */
public class TreeSequenceLocalityMeasurer {

	private static final String ORDER_STL = "STL";
	private static final String ORDER_LTS = "LTS";
	
	private static final String PARAM_METRIC = "metric";
	private static final String PARAM_RADIUS = "radius";
	private final static String PARAM_SPECIFIC_CELL_TYPE = "specific-cell-type";
	private final static String PARAM_RANDOMS = "randoms";
	
	private static final int METRIC_FROM_SEQUENCE = 1;
	private static final int METRIC_FROM_TREE = 2;

	private static final int DEFAULT_RADIUS = 2;
	
	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			
			if (config.getBoolean(PARAM_RANDOMS)){
				findRandomTreeLocalities(config);
			}
			else{
				findArborizationLocalities(config);
			}
			
		} catch (Exception e){
			e.printStackTrace();
		}
		
	}
	
	private static void findArborizationLocalities(JSAPResult config) throws Exception{
		DataProperties props = DataProperties.getInstance();

		String sequenceName;
		String traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
		String order = config.getString(NSConstants.PARAM_ORDER);
		int mapping = config.getInt(NSConstants.PARAM_MAPPING_MODE);
		int radius = config.getInt(PARAM_RADIUS);
		int metric = config.getInt(PARAM_METRIC);
		boolean specificCellClass = config.getBoolean(PARAM_SPECIFIC_CELL_TYPE);
		
		List<String> seqList = NSDataReader.loadSequenceList(NSConstants.MAPPING_MODE_TOPO_CA);
		Map<String,Integer> seqLenMap = NSDataReader.loadLengthDatabase(NSConstants.MAPPING_MODE_TOPO_CA);

		Map<String,NeuronMetadata> cellMetaData = MetaDataManager.getCellMetaData();
		NeuronMetadata neuron;
		String neuronName, fastaFilename, 
			baseDir = props.getFastaDir()+"Arborization/Mapping"+mapping+"/"+traversal+"/";
		FastaReader fr;

		Map<Object,DistributionBean> arbTypeLocMap = new HashMap<Object,DistributionBean>();
		Map<Object,DistributionBean> cellTypeLocMap = new HashMap<Object,DistributionBean>();
		Map<Object,DistributionBean> labLocMap = new HashMap<Object,DistributionBean>();
		Map<Object,DistributionBean> cellTypeArbTypeLocMap = new HashMap<Object,DistributionBean>();
		Map<Object,DistributionBean> groupLocMap = new HashMap<Object,DistributionBean>();
		
		Map<Object,DistributionBean> seqLenDistMap = new HashMap<Object,DistributionBean>();
		
		DistributionBean distBean = new DistributionBean();
		String arborizationType, cellTypeArbType, group;
		BinaryTreeNode tree;
		StringBuffer fullResults = new StringBuffer();
		DistributionBean distanceDist;
		fullResults.append("Arborization,Length,Locality,Std Dev (across nodes)\r\n");

		for (int i = 0; i < seqList.size(); i++){
			// Load sequenceName
			sequenceName = (String)seqList.get(i);
			// Get Neuron data
			neuronName = NeuriteSequenceStringUtils.getNeuronNameFromSequence(sequenceName);
			neuron = (NeuronMetadata)cellMetaData.get(neuronName);
			
			arborizationType = NeuriteSequenceStringUtils.getArborizationType(sequenceName);
			if (!specificCellClass){
				neuron.setCellType(ClassificationUtils.reclassifyCellTypes(
					neuron.getCellType()));
				arborizationType = ClassificationUtils
					.reclassifyArborizationTypes(arborizationType);
			}
			cellTypeArbType = neuron.getCellType()+"."+arborizationType;
			group = neuron.getCellType()+"."+arborizationType+"."+neuron.getLabName();
			
			// Load fasta file / actual sequence
			fastaFilename = baseDir+neuron.getLabName()+"/"+sequenceName+"."+order.toLowerCase()+".fasta";
			fr = new FastaReader(fastaFilename);
			fr.next();
			// Get distance for sequence / tree analysis
			if (metric == METRIC_FROM_SEQUENCE){
				distanceDist = distancePerBit(fr.getSequence(),traversal,order,radius);
			}
			else{
				tree = DataConverter.fastaStringToTree(fr.getSequence());
				distanceDist = distancePerBit(tree,traversal,order,radius);
			}
			fullResults.append(sequenceName+","+fr.getSequenceLength()+","
					+distanceDist.getMean()+","+distanceDist.getStandardDeviation()+"\r\n");
			distBean.addData(distanceDist.getMean());
			
			addToClassTypeDist(seqLenDistMap,seqLenMap.get(sequenceName),distanceDist.getMean());
			
			addToClassTypeDist(arbTypeLocMap,arborizationType,distanceDist.getMean());
			addToClassTypeDist(cellTypeLocMap,neuron.getCellType(),distanceDist.getMean());
			addToClassTypeDist(labLocMap,neuron.getLabName(),distanceDist.getMean());
			addToClassTypeDist(cellTypeArbTypeLocMap,cellTypeArbType,distanceDist.getMean());
			addToClassTypeDist(groupLocMap,group,distanceDist.getMean());
		}
		if (metric == METRIC_FROM_SEQUENCE){
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromSeq-rad"+radius+"-"+traversal+order+".csv",
					fullResults.toString());
		}
		else{
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromTree-rad"+radius+"-"+traversal+order+".csv",
					fullResults.toString());
		}
		
		StringBuffer sb = new StringBuffer();
		sb.append("Stats for "
				+config.getString(NSConstants.PARAM_TRAVERSAL)
				+config.getString(NSConstants.PARAM_ORDER)
				+": \r\n");
		sb.append("Mean: "+distBean.getMean()+"\r\n");
		sb.append("Std Dev: "+distBean.getStandardDeviation()+"\r\n");
		sb.append("1st%: "+distBean.getPercentile(.01)+"\r\n");
		sb.append("5th%: "+distBean.getPercentile(.05)+"\r\n");
		sb.append("50th%: "+distBean.get50thPercentile()+"\r\n");
		sb.append("95th%: "+distBean.get95thPercentile()+"\r\n");
		sb.append("99th%: "+distBean.get99thPercentile()+"\r\n");
		if (metric == METRIC_FROM_SEQUENCE){
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromSeqSummary-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		else{
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromTreeSummary-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		
		sb = new StringBuffer();
		writeClassTypeDist(sb,arbTypeLocMap,"Arborization Type");
		writeClassTypeDist(sb,cellTypeLocMap,"Cell Type");
		writeClassTypeDist(sb,labLocMap,"Lab");
		writeClassTypeDist(sb,cellTypeArbTypeLocMap,"CellType.ArbType");
		writeClassTypeDist(sb,groupLocMap,"CellType.ArbType.Lab");
		if (metric == METRIC_FROM_SEQUENCE){
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromSeqSummary-perClass-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		else{
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/localityFromTreeSummary-perClass-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		/*
		sb = new StringBuffer();
		sb.append("Length,Mean Locality,Std Dev\r\n");
		List lenList = new LinkedList(seqLenDistMap.keySet());
		Collections.sort(lenList);
		for (int i = 0; i < lenList.size(); i+=2){
			distBean = (DistributionBean)seqLenDistMap.get(lenList.get(i));
			sb.append(lenList.get(i)+","+distBean.getMean()+","+distBean.getStandardDeviation()+"\r\n");
		}
		if (metric == METRIC_FROM_SEQUENCE){
			DataWriter.writeStringToFile(
					PropertyUtils.getResultsDir()+"locality/localityFromSeq-perLength-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		else{
			DataWriter.writeStringToFile(
					PropertyUtils.getResultsDir()+"locality/localityFromTree-perLength-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}*/
	}
	
	private static void addToClassTypeDist(Map<Object,DistributionBean> locMap, Object key, double value){
		if (!locMap.containsKey(key)){
			locMap.put(key, new DistributionBean());
		}
		locMap.get(key).addData(value);
	}

	private static void writeClassTypeDist(StringBuffer sb, Map<Object,DistributionBean> locMap, String classType){
		DistributionBean distBean;
		sb.append(classType+",N,MeanLocality,Std Dev,Std Err\r\n");
		for (Object className : locMap.keySet()){
			distBean = locMap.get(className);
			sb.append(className+","+distBean.getCount()+","+distBean.getMean()+","
					+distBean.getStandardDeviation()+","+distBean.getStandardError()+"\r\n");
		}
		sb.append("\r\n");
	}

	private static void findRandomTreeLocalities(JSAPResult config) throws Exception{
		
		String traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
		String order = config.getString(NSConstants.PARAM_ORDER);
		int mapping = config.getInt(NSConstants.PARAM_MAPPING_MODE);
		int radius = config.getInt(PARAM_RADIUS);
		int metric = config.getInt(PARAM_METRIC);
		DataProperties props = DataProperties.getInstance();
		String baseDir = props.getFastaDir()+"Random/Method3/Mapping"+mapping+"/"+traversal+order+"/";
		FastaReader fr;
	
		DistributionBean distBean;
		String sequenceName, fastaFilename;
		BinaryTreeNode tree;
		DistributionBean distanceDist;
		
		List<Integer> lengthList = NSDataReader.loadUniqueSequenceLengths(NSConstants.MAPPING_MODE_TOPO_CA);
		int length;
		NumberFormat nf = NumberFormat.getInstance();
		
		StringBuffer sb = new StringBuffer();
		sb.append("Length,Mean Locality,Std Dev\r\n");
		for (int i = 0; i < lengthList.size(); i+=2){
			distBean = new DistributionBean();
			length = lengthList.get(i);
			for (int j = 1; j <= 100; j++){
				// Load sequenceName
				sequenceName = "LEN"+length+"NUM"+j+order+".fasta";
				
				// Load fasta file / actual sequence
				fastaFilename = baseDir+sequenceName;
				fr = new FastaReader(fastaFilename);
				fr.next();
				// Get distance for sequence / tree analysis
				if (metric == METRIC_FROM_SEQUENCE){
					distanceDist = distancePerBit(fr.getSequence(),traversal,order,radius);
				}
				else{
					tree = DataConverter.fastaStringToTree(fr.getSequence());
					distanceDist = distancePerBit(tree,traversal,order,radius);
				}
				distBean.addData(distanceDist.getMean());				
			}
			sb.append(length+",");
			sb.append(nf.format(distBean.getMean())+",");
			sb.append(nf.format(distBean.getStandardDeviation())+"\r\n");
		}
		if (metric == METRIC_FROM_SEQUENCE){
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/RandomLocalityFromSeq-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		else{
			DataWriter.writeStringToFile(
					props.getResultsDir()+"locality/RandomLocalityFromTree-rad"+radius+"-"+traversal+order+".csv",
					sb.toString());
		}
		
	}
	
	/**
	 * @param sequence
	 * @param traversal
	 * @param order
	 * @return
	 * @throws Exception
	 * 
	 * Returns average across sequence of average ratio of differences of nearby nodes to
	 * their counterparts in the tree representation
	 */
	public static DistributionBean distancePerBit(
			String sequence, String traversal, String order)
	throws Exception {
		return distancePerBit(sequence,traversal,order,DEFAULT_RADIUS);
	}
	public static DistributionBean distancePerBit(
			String sequence, String traversal, String order, int radius)
	throws Exception {
		DistributionBean distancePerBit = new DistributionBean();
		BinaryTreeNode tree = DataConverter.fastaStringToTree(sequence);
		Map<Integer,BinaryTreeNode> map = null;
		if (traversal.equals("Height")){
			map = setNodeIdsBySize(tree, order);
		}
		else if (traversal.equals("Size")){
			map = setNodeIdsByHeight(tree, order);
		}
		double nodeDistance;
		int nodesCounted;
		for (int i = 0; i < sequence.length(); i++){
			nodesCounted = 0;
			nodeDistance = 0;
			for (int j = -radius; j <= radius; j++){
				if (j == 0) j++;
				if (i+j >= 0 && i+j < sequence.length()){
					nodeDistance += distanceOf(
							map.get(i), 
							map.get(i+j))/Math.abs(j);
					nodesCounted++;
				}
			}
			distancePerBit.addData(nodeDistance/nodesCounted);
		}
		return distancePerBit;
	}
	
	public static DistributionBean distancePerBit(
			BinaryTreeNode tree, String traversal, String order)
	throws Exception {
		return distancePerBit(tree,traversal,order,DEFAULT_RADIUS);
	}
	public static DistributionBean distancePerBit(
			BinaryTreeNode tree, String traversal, String order, int radius)
	throws Exception {
		DistributionBean distancePerBit = new DistributionBean();
		BinaryTreeNode currentNode;
		if (traversal.equals("Height")){
			setNodeIdsByHeight(tree, order);			
		}
		else if (traversal.equals("Size")){
			setNodeIdsBySize(tree, order);
		}
		
		Stack<BinaryTreeNode> mainStack = new Stack<BinaryTreeNode>(), subStack = new Stack<BinaryTreeNode>();
		mainStack.push(tree);
		double summedDistance = 0;
		int currNodeDistance = 0, nodeId;
		// Holds distance from currentNode for subStack loop
		Map<Object,Integer> distanceMap = new HashMap<Object,Integer>();
		// Holds which nodes have already been checked
		Set<Object> dataSet = new TreeSet<Object>();
		// Main loop, go through every node in the tree
		while (!mainStack.isEmpty()){
			currentNode = (BinaryTreeNode)mainStack.pop();
			nodeId = ((Integer)currentNode.getData()).intValue();
			if (!currentNode.isLeaf()){
				mainStack.push(currentNode.getLeft());
				mainStack.push(currentNode.getRight());
			}
			// Subloop, go through every node within 2 edges
			subStack.push(currentNode);
			distanceMap.put(currentNode.getData(), new Integer(0));
			int nodeCount = 0;
			summedDistance = 0;
			while (!subStack.isEmpty()){
				currentNode = subStack.pop();
				currNodeDistance = distanceMap.get(currentNode.getData());
				nodeCount++;

				if (currNodeDistance > 0){
					summedDistance += Math.abs(
						((Integer)currentNode.getData()).intValue() - nodeId) / currNodeDistance;
				}
				
				if (currNodeDistance < radius){
					if (currentNode.hasLeft() && !dataSet.contains(currentNode.getLeft().getData())){
						dataSet.add(currentNode.getLeft().getData());
						subStack.push(currentNode.getLeft());
						distanceMap.put(currentNode.getLeft().getData(), new Integer(currNodeDistance+1));
					}
					if (currentNode.hasRight() && !dataSet.contains(currentNode.getRight().getData())){
						dataSet.add(currentNode.getRight().getData());
						subStack.push(currentNode.getRight());
						distanceMap.put(currentNode.getRight().getData(), new Integer(currNodeDistance+1));
					}
					if (currentNode.getParent() != null && !dataSet.contains(currentNode.getParent().getData())){
						dataSet.add(currentNode.getParent().getData());
						subStack.push(currentNode.getParent());
						distanceMap.put(currentNode.getParent().getData(), new Integer(currNodeDistance+1));
					}
				}
			}
			distancePerBit.addData(summedDistance / nodeCount);
			dataSet.clear();
			distanceMap.clear();
		}
		return distancePerBit;
	}
	
	public static int distanceOf(BinaryTreeNode node1, BinaryTreeNode node2){
		BinaryTreeNode currentNode;
		Set<Object> dataSet = new TreeSet<Object>();
		Map<Object,Integer> distanceMap = new HashMap<Object,Integer>();
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
		stack.push(node1);
		boolean found = false;
		int distance = 0;
		distanceMap.put(node1.getData(), new Integer(distance));
		dataSet.add(node1.getData());
		
		int currNodeId, node2Id = ((Integer)node2.getData()).intValue();
			
		
		while (!stack.isEmpty() && !found){
			currentNode = stack.pop();
			distance = distanceMap.get(currentNode.getData());
			if (currentNode.getData().equals(node2.getData())){
				found = true;
			}
			else{
				currNodeId = ((Integer)currentNode.getData()).intValue();
				if (currNodeId < node2Id && node2Id <= (currNodeId+currentNode.getTreeSize()-1)){
					if (currentNode.hasLeft() && !dataSet.contains(currentNode.getLeft().getData())){
						dataSet.add(currentNode.getLeft().getData());
						stack.push(currentNode.getLeft());
						distanceMap.put(currentNode.getLeft().getData(), new Integer(distance+1));
					}
					if (currentNode.hasRight() && !dataSet.contains(currentNode.getRight().getData())){
						dataSet.add(currentNode.getRight().getData());
						stack.push(currentNode.getRight());
						distanceMap.put(currentNode.getRight().getData(), new Integer(distance+1));
					}
				}
				else{
					if (currentNode.getParent() != null && !dataSet.contains(currentNode.getParent().getData())){
						dataSet.add(currentNode.getParent().getData());
						stack.push(currentNode.getParent());
						distanceMap.put(currentNode.getParent().getData(), new Integer(distance+1));
					}
				}
			}
		}
		return distance;
	}

	
    public static Map<Integer,BinaryTreeNode> setNodeIdsBySize(BinaryTreeNode tree, String order) throws Exception{
    	Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
    	BinaryTreeNode current, right, left;
    	Map<Integer,BinaryTreeNode> map = new HashMap<Integer,BinaryTreeNode>();
    	stack.push(tree);
    	int count = 0;
    	while (!stack.isEmpty()){
    		current = stack.pop();
    		current.setData(new Integer(count));
    		map.put(count, current);
    		count++;
    		
    		right = current.getRight();
    		left = current.getLeft();
        	if (right != null){
        		if (left != null){
    		    	if (right.getTreeSize() > left.getTreeSize()){
    		    		addToStack(stack, right, left, order);
    		    	}
    		    	else if (left.getTreeSize() > right.getTreeSize()){
    		    		addToStack(stack, left, right, order);
    		    	}
    		    	else if (right.getHeight() > left.getHeight()){
    		    		addToStack(stack, right, left, order);
        	    	}
        	    	else if (left.getHeight() > right.getHeight()){
    		    		addToStack(stack, left, right, order);
        	    	}
        	    	else if (right.getAverageHeight() > left.getAverageHeight()){
    		    		addToStack(stack, right, left, order);
            	    }
        	    	else if (left.getAverageHeight() > right.getAverageHeight()){
    		    		addToStack(stack, left, right, order);
        	    	}
        	    	else if (right.getTreeAsymmetry() > left.getTreeAsymmetry()){
    		    		addToStack(stack, right, left, order);
            	    }
        	    	else {
    		    		addToStack(stack, left, right, order);
        	    	}
        		}
        		else{
        			stack.push(right);
        		}
        	}
        	else if (left != null){
        		stack.push(left);
        	}
    	}

    	return map;
    }

    public static Map<Integer,BinaryTreeNode> setNodeIdsByHeight(BinaryTreeNode tree, String order) throws Exception{
    	Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
    	BinaryTreeNode current, right, left;
    	Map<Integer,BinaryTreeNode> map = new HashMap<Integer,BinaryTreeNode>();
    	int count = 0;
    	stack.push(tree);
    	while (!stack.isEmpty()){
    		current = stack.pop();
			current.setData(new Integer(count));
			map.put(count, current);
    		count++;

    		right = current.getRight();
    		left = current.getLeft();
        	if (right != null){
        		if (left != null){
    		    	if (right.getHeight() > left.getHeight()){
    		    		addToStack(stack, right, left, order);
        	    	}
        	    	else if (left.getHeight() > right.getHeight()){
    		    		addToStack(stack, left, right, order);
        	    	}
        	    	else if (right.getAverageHeight() > left.getAverageHeight()){
    		    		addToStack(stack, right, left, order);
            	    }
        	    	else if (left.getAverageHeight() > right.getAverageHeight()){
    		    		addToStack(stack, left, right, order);
        	    	}
        	    	else if (right.getTreeSize() > left.getTreeSize()){
    		    		addToStack(stack, right, left, order);
    		    	}
    		    	else if (left.getTreeSize() > right.getTreeSize()){
    		    		addToStack(stack, left, right, order);
    		    	}
        	    	else if (right.getTreeAsymmetry() > left.getTreeAsymmetry()){
    		    		addToStack(stack, right, left, order);
            	    }
        	    	else {
    		    		addToStack(stack, left, right, order);
        	    	}
        		}
        		else{
        			stack.push(right);
        		}
        	}
        	else if (left != null){
        		stack.push(left);
        	}
    	}
    	return map;
    }

    private static void addToStack(
    		Stack<BinaryTreeNode> stack, 
    		BinaryTreeNode larger, 
    		BinaryTreeNode smaller,
    		String direction) throws Exception{
    	if (direction.equals(ORDER_LTS)){
    		stack.push(smaller);
    		stack.push(larger);
    	}
    	else if (direction.equals(ORDER_STL)){
    		stack.push(larger);
    		stack.push(smaller);
    	}
    	else{
    		throw new Exception("Improper value '"+direction+"' for parameter 'direction'."
    				+"Value must be '"+ORDER_STL
    				+"' or '"+ORDER_LTS+"'.");
    	}
    }
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(NSConstants.PARAM_TRAVERSAL)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(true) 
        	.setShortFlag('T') 
        	.setLongFlag(NSConstants.PARAM_TRAVERSAL);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(true) 
        	.setShortFlag('O') 
        	.setLongFlag(NSConstants.PARAM_ORDER);
    	jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
    		.setStringParser(JSAP.INTEGER_PARSER)
    		.setRequired(true) 
    		.setShortFlag('M') 
    		.setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_METRIC)
    		.setStringParser(JSAP.INTEGER_PARSER)
    		.setRequired(false)
    		.setDefault(""+METRIC_FROM_SEQUENCE)
    		.setShortFlag('m') 
    		.setLongFlag(PARAM_METRIC);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_RADIUS)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_RADIUS)
			.setShortFlag('r') 
			.setLongFlag(PARAM_RADIUS);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SPECIFIC_CELL_TYPE)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setDefault("false")
			.setRequired(false) 
			.setShortFlag('c') 
			.setLongFlag(PARAM_SPECIFIC_CELL_TYPE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_RANDOMS)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setDefault("false")
			.setRequired(false) 
			.setShortFlag('R') 
			.setLongFlag(PARAM_RANDOMS);
        jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + TreeSequenceLocalityMeasurer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
