package org.cn3.ns.data;

import java.io.File;
import java.text.DateFormat;
import java.util.Date;
import java.util.List;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataUpdater;
import org.cn3.ns.nodeencoding.EncodingMode;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Run when there is new data in the NeuroMorpho.Org database. This program
 * will pull the new meta-data (assuming the swc files have already been downloaded)
 * and update local data files.
 * The following procedures are run:
 *  - data.MetaDataUpdater
 *  - data.SwcToFastaConverter (creates FASTA files)
 *  - generation.CreateLengthDatabase (for all mapping types, sequence lengths file is generated)
 *  - generation.MakeMasterSequence
 *
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.data.NeuroMorphoUpdater --minLength 20 --maxLength 0
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.data.NeuroMorphoUpdater --minLength 18 --maxLength 1200 --spurSize 5
 *
 */
public class NeuroMorphoUpdater {

	private static final String order = "";
	
	private static final String PARAM_MIN_LENGTH = "minLength";
	private static final String PARAM_MAX_LENGTH = "maxLength";
	private static final String PARAM_SPUR_SIZE = "spurSize";
	
	private static final int DEFAULT_MIN_LENGTH = 18;
	private static final int DEFAULT_MAX_LENGTH = 2000;
	private static final float DEFAULT_SPUR_SIZE = 5;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			int minLength = config.getInt(PARAM_MIN_LENGTH);
			int maxLength = config.getInt(PARAM_MAX_LENGTH);
			float spurSize = config.getFloat(PARAM_SPUR_SIZE);
			updateForMapping(NSConstants.MAPPING_MODE_TOPO_ACT, minLength, maxLength, spurSize);
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static void updateForMapping(int mapping, int minLength, int maxLength, float spurSize) throws Exception{
		// Update Metadata file
		System.out.println("Updating meta data...");
		MetaDataUpdater.update();
		System.out.println();
		
		// Convert all swc files to fasta for each mapping (also creates BLAST database)
		DataProperties dataProps = DataProperties.getInstance();
		SwcToFastaConverter swcToFastaConverter = new SwcToFastaConverter(dataProps.getGlobalFastaDir());
		swcToFastaConverter.setEncoding(EncodingMode.find(mapping));
		swcToFastaConverter.setOrder(order);
		swcToFastaConverter.setMinLength(minLength);
		swcToFastaConverter.setMaxLength(maxLength);
		swcToFastaConverter.setSpurLength(spurSize);
			
		System.out.println("Converting SWC files to FASTA files...");
		String swcDir = dataProps.getSwcDir("NM");
		swcToFastaConverter.convertSwcArchivesToFastaDatabase(swcDir);
		System.out.println();

		/*

		// Create sequence lengths file
		Set<Integer> typesCovered = new TreeSet<Integer>();
		Integer currentType;
		StringBuffer sb;
		currentType = new Integer(NSConstants.mappingModeSequenceType[mapping]);
		
		String dataset = "NM";
		if (spurSize > 0){
			NumberFormat nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(2);
			dataset += "_SpurLen"+nf.format(spurSize);
		}

		CreateLengthDatabase createLD = new CreateLengthDatabase(dataset,NSConstants.MAPPING_MODE_TOPO_ACT);
		typesCovered.add(currentType);
		createLD.setMapping(mapping);
		System.out.println("Creating length database...");
		createLD.run();
		System.out.println();
		List<String>[] newSequenceLists = new List[NSConstants.SEQUENCE_TYPES.length+1];
		newSequenceLists[currentType.intValue()] = createLD.getNewSequences();

		// Update databaseLengths with number of sequences
		Properties prop = NSDataReader.loadProperties("databaseLengths");
		sb = new StringBuffer();
		String key, currentKey = "NMArbor.Map"+mapping+".length";
		for (Iterator<Object> it = prop.keySet().iterator(); it.hasNext();){
			key = (String)it.next();
			if (key.equals(currentKey)){
				sb.append(currentKey+"="+createLD.getNumTotalSequences());
			}
			else{
				sb.append(key+"="+prop.getProperty(key));
			}
		}
		DataWriter.writeStringToFile("resources/databaseLengths.properties", sb.toString());

		if (StringUtils.isEmpty(order)){
			updateForOrder(mapping, "LTS", newSequenceLists[currentType]);
			updateForOrder(mapping, "STL", newSequenceLists[currentType]);
		}
		else{
			updateForOrder(mapping, order, newSequenceLists[currentType]);
		}
		*/
	}

	private static void updateForOrder(int mapping, String order, List<String> newSequences) throws Exception{
		File prevFile, newFile;
		String dateStr = DateFormat.getDateInstance().format(new Date());

		// Create separate FASTA file for new sequences
		StringBuffer sb = new StringBuffer();
		FastaReader fr = new FastaReader(
				DataProperties.getInstance().getFastaFilename(mapping));
		while (fr.next()){
			if (newSequences.contains(fr.getSequenceName())){
				sb.append(fr.getHeader()+"\n");
				sb.append(NeuriteSequenceStringUtils.insertBreaksAtLength(fr.getSequence(),80)+"\n");
			}
		}
		newFile = new File(DataProperties.getInstance().getNewSequencesFastaFilename(
				NSConstants.DEFAULT_DATASET, mapping, order));
		if (newFile.exists()){
			prevFile = new File(DataProperties.getInstance().getFastaArborDir()+
					"/Mapping"+mapping+"/Size/NMArborSize"+order+"Map"+mapping+"-New."+dateStr+".fasta");
			newFile.renameTo(prevFile);
		}
		DataWriter.writeStringToFile(newFile, sb.toString());
		
		// Generate BLAST database for the new sequences file
		/*
		BlastUtils.generateDatabase(
				"NMArborSize"+order+"Map"+mapping+"-New", newFile.getAbsolutePath(), NSConstants.mappingModeBlastP[mapping]);

		// Make the master sequence for each mapping
		MakeMasterSequence.makeFullDataSequence(
				NSConstants.DEFAULT_DATASET, mapping, order);
		*/
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_MIN_LENGTH)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setDefault(""+DEFAULT_MIN_LENGTH)
        .setRequired(false)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_MIN_LENGTH);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MAX_LENGTH)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setDefault(""+DEFAULT_MAX_LENGTH)
        .setRequired(false)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_MAX_LENGTH);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SPUR_SIZE)
        .setStringParser(JSAP.FLOAT_PARSER)
        .setDefault(""+DEFAULT_SPUR_SIZE)
        .setRequired(false)
        .setShortFlag('s') 
        .setLongFlag(PARAM_SPUR_SIZE);
	    jsap.registerParameter(opt);

	    JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + NeuroMorphoUpdater.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}
	
}
