package org.cn3.ns.data;

import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.data.NeuronTreeClassification;
import org.cn3.morpho.data.SWCReader;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.Neuron;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.domain.NeuronMetadataComplete;
import org.cn3.morpho.domain.NeuronProcess;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.utils.StringUtils;
import org.cn3.ns.domain.FastaBean;
import org.cn3.ns.domain.TraversalOrder;
import org.cn3.ns.nodeencoding.EncodingMode;
import org.cn3.ns.nodeencoding.NodeEncoder;
import org.cn3.ns.nodeencoding.NodeEncoderFactory;
import org.cn3.ns.resources.NSConstants;

import org.cn3.morpho.utils.StringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;
import org.w3c.dom.Node;

/**
 * 
 * @author gillette
 * Converts .swc files info .fasta files. Main function calls readSWC(), which appends 
 * all fasta sequences together into one database.
 * Conversion based on method of division (full cell, arborizations, trees), primary 
 * method of traversal (height, size), and order of traversal (small then large, large
 * then small).
 * 
 * Initialize with destination directory (fastaDir), then call basic method with swcFile
 * parameter to create a fasta file from swc.
 * 
 * Function convertSwcArchivesToFastaDatabase() takes swcDir, converting all swc files
 * and producing archives of swc files based on other input parameters. 
 * 
 * java org.cn3.ns.data.SwcToFastaConverter --swc-dir C:/swcData/ --fasta-dir C:/fastaFromSwc/ --write-celltypes true
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.data.SwcToFastaConverter --swc-dir /home/tgillett/swcFiles/ --fasta-dir /home/tgillett/fastaFromSwc/
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.data.SwcToFastaConverter --swc-dir /home/tgillett/swcFiles/ --fasta-dir /home/tgillett/fastaFromSwc/ --spurLength 2 --removeDendriteSpurs
 * 
 */
public class SwcToFastaConverter {

	private final static String CLASSIFICATION_ARBOR = "Arborization";
	private final static String CLASSIFICATION_ARBOR_SUB = "ArborizationSub";
	private final static String CLASSIFICATION_TREE = "Tree";
	
	private enum TreeType { CELL, AXON, BASAL_DENDRITE, APICAL_DENDRITE, GENERAL_DENDRITE, BDTREE, ADTREE };
/*	
	private static final int TREE_TYPE_CELL = 1;
	private static final int TREE_TYPE_AXON = 2;
	private static final int TREE_TYPE_BASAL_DENDRITE = 3;
	private static final int TREE_TYPE_APICAL_DENDRITE = 4;
	private static final int TREE_TYPE_GENERAL_DENDRITE = 7;
	private static final int TREE_TYPE_BDTREE = 5;
	private static final int TREE_TYPE_ADTREE = 6;
*/
	// For swc->TreeNodes
	public static final int INDEX_AXON = 0;
	public static final int INDEX_APICAL = 1;
	public static final int INDEX_DENDRITE = 2;
	
	private static final String PARAM_SWC_DIR = "swc-dir";
	private static final String PARAM_FASTA_DIR = "fasta-dir";
	private static final String PARAM_MIN_LENGTH = "min-length";
	private static final String PARAM_WRITE_ARCHIVES = "write-archives";
	private static final String PARAM_WRITE_CELLTYPES = "write-celltypes";
	private static final String PARAM_GENERATE_DATABASE = "genDB";
	private static final String PARAM_SPUR_LENGTH = "spurLength";
	private static final String PARAM_REMOVE_DENDRITE_SPURS = "removeDendriteSpurs";
	private static final String PARAM_REMOVE_APICAL_SPURS = "removeApicalSpurs";
	private static final String PARAM_REMOVE_AXON_SPURS = "removeAxonSpurs";

	private static final String PARAM_REMOVE_NON_CONTROL = "removeNonControl";
	private static final String PARAM_REMOVE_ACUTE_CULTURE = "removeAcuteCulture";
	private static final String PARAM_REMOVE_FLAGGED_NEURITES = "removeFlaggedNeurites";

	private static final int DEFAULT_MIN_LENGTH = 20;
	private static final int DEFAULT_MAX_LENGTH = 0;
	private static final boolean DEFAULT_WRITE_ARCHIVES = false;
	private static final boolean DEFAULT_WRITE_CELLTYPES = false;
	private static final boolean DEFAULT_GENERATE_DATABASE = false;

	public SwcToFastaConverter(String fastaDir) throws Exception{
		setDefaults();
		this.fastaDir = fastaDir;
	}

	public SwcToFastaConverter(String fastaDir, String order) throws Exception{
		setDefaults();
		this.fastaDir = fastaDir;
		this.order = order;
	}

	public void setDefaults() {
		classification = NSConstants.DEFAULT_CLASSIFICATION;
		minLength = DEFAULT_MIN_LENGTH;
		maxLength = DEFAULT_MAX_LENGTH;
		traversal = NSConstants.DEFAULT_TRAVERSAL;
		encoding = NSConstants.DEFAULT_MAPPING;
		try {
			nodeEncoder = NodeEncoderFactory.createNewNodeEncoder(encoding);
		} catch (Exception e){
			System.out.println("Unknown node encoding given a known node encoding?!?");
		}
		spurLength = 0;
		debug = false;
	}

	private static boolean debug = false;
	
	private String fastaDir;
	private String swcDir;
	private String dataset;
	private NeuronTreeClassification classification;
	private String traversal;
	private String order;
	private NodeEncoder nodeEncoder;
	private EncodingMode encoding;
	private int minLength;
	private int maxLength;
	private int lineLength = 0;
	private boolean writeArchives;
	private boolean writeCellTypes;
	private float spurLength; // microns
	private boolean removeDendriteSpurs;
	private boolean removeApicalSpurs;
	private boolean removeAxonSpurs;

	private boolean removeFlaggedNeurites;
	private boolean removeNonControl;
	private boolean removeAcuteCulture;
	
	private Set<String> flaggedNeurites = null;

	public boolean isWriteArchives() {
		return writeArchives;
	}

	public void setWriteArchives(boolean writeArchives) {
		this.writeArchives = writeArchives;
	}

	public boolean isWriteCellTypes() {
		return writeCellTypes;
	}

	public void setWriteCellTypes(boolean writeCellTypes) {
		this.writeCellTypes = writeCellTypes;
	}

	public int getLineLength() {
		return lineLength;
	}

	public void setLineLength(int lineLength) {
		this.lineLength = lineLength;
	}

	public int getMinLength() {
		return minLength;
	}

	public void setMinLength(int minLength) {
		this.minLength = minLength;
	}

	public int getMaxLength() {
		return maxLength;
	}

	public void setMaxLength(int maxLength) {
		this.maxLength = maxLength;
	}

	public float getSpurLength() {
		return spurLength;
	}

	public void setSpurLength(float spurLength) {
		this.spurLength = spurLength;
	}

	public boolean isRemoveDendriteSpurs(){
		return removeDendriteSpurs;
	}

	public void setRemoveDendriteSpurs(boolean removeDendriteSpurs){
		this.removeDendriteSpurs = removeDendriteSpurs;
	}

	public boolean isRemoveApicalSpurs(){
		return removeApicalSpurs;
	}

	public void setRemoveApicalSpurs(boolean removeApicalSpurs){
		this.removeApicalSpurs = removeApicalSpurs;
	}

	public boolean isRemoveAxonSpurs(){
		return removeAxonSpurs;
	}

	public void setRemoveAxonSpurs(boolean removeAxonSpurs){
		this.removeAxonSpurs = removeAxonSpurs;
	}

	public EncodingMode getEncoding() {
		return encoding;
	}

	public void setEncoding(EncodingMode encoding) throws Exception {
		this.encoding = encoding;
		this.nodeEncoder = NodeEncoderFactory.createNewNodeEncoder(encoding);
	}

	public String getTraversal() {
		return traversal;
	}

	public void setTraversal(String traversal) {
		this.traversal = traversal;
	}

	public String getOrder() {
		return order;
	}

	public void setOrder(String order) {
		this.order = order;
	}

	public NeuronTreeClassification getClassification() {
		return classification;
	}

	public void setClassification(NeuronTreeClassification classification) {
		this.classification = classification;
	}

	public void setClassification(String classification) {
		this.classification = NeuronTreeClassification.find(classification);
	}

	public String getDataset() {
		return dataset;
	}

	public void setDataset(String dataset) {
		this.dataset = dataset;
	}

	public String getFastaDir() {
		return fastaDir;
	}

	public void setFastaDir(String fastaDir) {
		this.fastaDir = fastaDir;
	}

	public String getSwcDir() {
		return swcDir;
	}

	public void setSwcDir(String swcDir) {
		this.swcDir = swcDir;
	}

	public boolean isRemoveSpursDendrite() {
		return removeDendriteSpurs;
	}

	public void setRemoveSpursDendrite(boolean removeSpursDendrite) {
		this.removeDendriteSpurs = removeSpursDendrite;
	}

	public boolean isRemoveSpursApical() {
		return removeApicalSpurs;
	}

	public void setRemoveSpursApical(boolean removeSpursApical) {
		this.removeApicalSpurs = removeSpursApical;
	}

	public boolean isRemoveSpursAxon() {
		return removeAxonSpurs;
	}

	public void setRemoveSpursAxon(boolean removeSpursAxon) {
		this.removeAxonSpurs = removeSpursAxon;
	}
	
	public boolean isRemoveFlaggedNeurites() {
		return removeFlaggedNeurites;
	}

	public void setRemoveFlaggedNeurites(boolean removeFlaggedNeurites) {
		this.removeFlaggedNeurites = removeFlaggedNeurites;
	}

	public boolean isRemoveNonControl() {
		return removeNonControl;
	}

	public void setRemoveNonControl(boolean removeNonControl) {
		this.removeNonControl = removeNonControl;
	}

	public boolean isRemoveAcuteCulture() {
		return removeAcuteCulture;
	}

	public void setRemoveAcuteCulture(boolean removeAcuteCulture) {
		this.removeAcuteCulture = removeAcuteCulture;
	}

	
	public String getAllFilename(String dataset){
		String allFilename = dataset;
		switch(classification){
		case ARBORIZATION:
			allFilename = allFilename + "Arbor" + traversal;
			break;
		case SUB_ARBORIZATION:
			allFilename = allFilename + "SubArbor" + traversal;
			break;
		default:
			allFilename = allFilename + classification + traversal;
		}
		return allFilename;
	}
	
	public static void main(String args[]){
		try{
			System.out.println("Running");
			
			JSAPResult config = readParameters(args);
			String swcDir = config.getString(PARAM_SWC_DIR);
			String dataset = config.getString(NSConstants.PARAM_DATASET);
			String fastaDir = config.getString(PARAM_FASTA_DIR);
			String classification = config.getString(NSConstants.PARAM_CLASSIFICATION);
			String traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
			String order = config.getString(NSConstants.PARAM_ORDER);
			//int mapping = config.getInt(NSConstants.PARAM_MAPPING_MODE);
			EncodingMode mapping = EncodingMode.find(config.getString(NSConstants.PARAM_MAPPING_MODE));
			
			debug = config.getBoolean(NSConstants.PARAM_DEBUG);
			
			SwcToFastaConverter swcToFastaConverter = new SwcToFastaConverter(fastaDir);
			// Classification is "Cell", "Arborization", or "Tree"
			swcToFastaConverter.setClassification(classification);
			// Traversal is "Size", "Height", or "AvgHeight"
			swcToFastaConverter.setTraversal(traversal);
			swcToFastaConverter.setOrder(order);
			swcToFastaConverter.setEncoding(mapping);
			swcToFastaConverter.setMinLength(config.getInt(PARAM_MIN_LENGTH));

			swcToFastaConverter.setRemoveFlaggedNeurites(config.getBoolean(PARAM_REMOVE_FLAGGED_NEURITES));
			swcToFastaConverter.setRemoveNonControl(config.getBoolean(PARAM_REMOVE_NON_CONTROL));
			swcToFastaConverter.setRemoveAcuteCulture(config.getBoolean(PARAM_REMOVE_ACUTE_CULTURE));

			swcToFastaConverter.setSpurLength(config.getFloat(PARAM_SPUR_LENGTH));
			swcToFastaConverter.setRemoveDendriteSpurs(config.getBoolean(PARAM_REMOVE_DENDRITE_SPURS));
			swcToFastaConverter.setRemoveApicalSpurs(config.getBoolean(PARAM_REMOVE_APICAL_SPURS));
			swcToFastaConverter.setRemoveAxonSpurs(config.getBoolean(PARAM_REMOVE_AXON_SPURS));

			boolean writeArchives = config.getBoolean(PARAM_WRITE_ARCHIVES);
			boolean writeCellTypes = config.getBoolean(PARAM_WRITE_CELLTYPES);
			
			swcToFastaConverter.convertSwcArchivesToFastaDatabase(swcDir+"/"+dataset, dataset, writeArchives, writeCellTypes);

		} catch (Exception e){
			e.printStackTrace();
		}
	}

	public List<File> writeFastaBeansToFiles(List<FastaBean> fastaBeans) throws IOException{
		return writeFastaBeansToFiles(fastaBeans, fastaDir);
	}
	public static List<File> writeFastaBeansToFiles(List<FastaBean> fastaBeans, String fastaDir) throws IOException{
		FastaBean fastaBean;
		String fastaFilename;
		File fastaFile;
		List<File> fastaFiles = new ArrayList<File>();
		for (int i = 0; i < fastaBeans.size(); i++){
			fastaBean = fastaBeans.get(i);
			fastaFilename = fastaBean.getSequenceName()+"."+fastaBean.getTraversalOrder()+".fasta";
			fastaFile = new File(fastaDir+"/"+fastaFilename);
			fastaFiles.add(fastaFile);
			writeFastaBeanToFile(fastaBean, fastaFile);
		}
		return fastaFiles;
	}
	public static void writeFastaBeanToFile(FastaBean fastaBean, File fastaFile) throws IOException{
		DataWriter.writeStringToFile(
				fastaFile, 
				//">"+fastaBean.getHeader()+"\n"+StringUtils.insertBreaksAtLength(fastaBean.getSequence()+"\n",80)
				">"+fastaBean.getHeader()+"\n"+fastaBean.getSequence()+"\n"
		);
	}
	
	public List<File> convertSwcToFastaFiles(File swcFile) throws Exception{
		List<FastaBean> fastaBeans = convertSwcToFasta(swcFile);
		List<File> fastaFiles = writeFastaBeansToFiles(fastaBeans);
		return fastaFiles;
	}
	
	public List<FastaBean> convertSwcToFasta(File swcFile) throws Exception{
		List<FastaBean> fastaBeans = new ArrayList<FastaBean>();

		String neuronName = swcFile.getName().replace(".CNG", "").replace(".SWC", "").replace(".swc", "");
		Soma<BinaryTreeNode> soma = SWCReader.convertSwcToBinarySoma(swcFile, classification, encoding.getSizeFromLength(minLength));
		
		if (isSTL(order)){
			fastaBeans.addAll(convertSomaToFastaBeans(neuronName, soma, "STL"));
		}
		if (isLTS(order)){
			fastaBeans.addAll(convertSomaToFastaBeans(neuronName, soma, "LTS"));
		}
		return fastaBeans;
	}

	private List<FastaBean> convertSomaToFastaBeans(String neuronName, Soma<BinaryTreeNode> soma, String order)
	throws Exception{
		List<FastaBean> fastaBeans = new ArrayList<FastaBean>();
		BinaryTreeNode tree;

		int minSize = encoding.getSizeFromLength(minLength);
		switch(classification){
		case ARBORIZATION:
			if (soma.hasAxonalTree() && 
					soma.getAxonalTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-Axon",
						nodeEncoder.encodeTree(soma.getAxonalTree().getTreeHead(), order),
						"Axon", order));
			}
			if (soma.hasBasalDendriteTree() && 
					soma.getBasalDendriteTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-Dendrite",
						nodeEncoder.encodeTree(soma.getBasalDendriteTree().getTreeHead(), order),
						"Dendrite", order));
			}
			break;
		case TREE:
			if (soma.hasAxonalTree() && 
					soma.getAxonalTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-Axon",
						nodeEncoder.encodeTree(soma.getAxonalTree().getTreeHead(), order),
						"Axon", order));
			}
			if (soma.hasBasalDendriteTrees()){
				for (int i = 1; i <= soma.getBasalDendriteTrees().size(); i++){
					tree = soma.getBasalDendriteTrees().get(i-1);
					if (tree.getTreeSize() >= minSize){
						fastaBeans.add(new FastaBean(
								neuronName+"-Dendrite"+i,
								nodeEncoder.encodeTree(tree, order),
								"Dendrite", order));
					}
				}
			}
			if (soma.hasApicalDendriteTrees()){
				for (int i = 1; i <= soma.getApicalDendriteTrees().size(); i++){
					tree = soma.getApicalDendriteTrees().get(i-1);
					if (tree.getTreeSize() >= minSize){
						fastaBeans.add(new FastaBean(
								neuronName+"-ApicalDendrite"+i,
								nodeEncoder.encodeTree(tree, order),
								"ApicalDendrite", order));
					}
				}
			}
			break;
		case NEURON:
			break;
		case SUB_ARBORIZATION:
		default: // Default to SUB_ARBORIZATION if not selected
			if (soma.hasAxonalTree() && 
					soma.getAxonalTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-Axon",
						nodeEncoder.encodeTree(soma.getAxonalTree().getTreeHead(), order),
						"Axon", order));
			}
			if (soma.hasBasalDendriteTree() && 
					soma.getBasalDendriteTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-Dendrite",
						nodeEncoder.encodeTree(soma.getBasalDendriteTree().getTreeHead(), order),
						"Dendrite", order));
			}
			if (soma.hasApicalDendriteTree() && 
					soma.getApicalDendriteTree().getTreeHead().getTreeSize() >= minSize){
				fastaBeans.add(new FastaBean(
						neuronName+"-ApicalDendrite",
						nodeEncoder.encodeTree(soma.getApicalDendriteTree().getTreeHead(), order),
						"ApicalDendrite", order));
			}
		}
		return fastaBeans;
	}
	
	private List<String> neuriteTypes = null;
	private Map<String,StringBuilder[]> neuriteGroupContent;
	/**
	 * 
	 * @throws Exception
	 */
	
	public void convertSwcArchivesToFastaDatabase(String swcDir) throws Exception{
		dataset = NSConstants.DEFAULT_DATASET;
		writeArchives = DEFAULT_WRITE_ARCHIVES;
		writeCellTypes = DEFAULT_WRITE_CELLTYPES;
		convertSwcArchivesToFastaDatabase(swcDir, dataset, writeArchives, writeCellTypes);
	}
	public void convertSwcArchivesToFastaDatabase(String swcDir, String dataset, boolean writeArchives, boolean writeCellTypes)
	throws Exception{
		// Validate Arguments
		if (!classification.equals(CLASSIFICATION_ARBOR)
				&& !classification.equals(CLASSIFICATION_ARBOR_SUB)
				&& !classification.equals(CLASSIFICATION_TREE)){
			throw new Exception("Classification "+classification+" unknown; must be one of {"+
					CLASSIFICATION_ARBOR+", "+CLASSIFICATION_ARBOR_SUB+", "+CLASSIFICATION_TREE+"}");
		}
		if (!traversal.equals("Size")
				&& !traversal.equals("Height")
				&& !traversal.equals("AvgHeight")){
			throw new Exception("Traversal "+traversal+" unknown; must be one of {Sise, Height, AvgHeight}");
		}

		if (flaggedNeurites == null){
			flaggedNeurites = new TreeSet<String>(NSDataReader.getFlaggedSequences());
		}
		
		boolean isSTL = isSTL(order);
		boolean isLTS = isLTS(order);
		neuriteTypes = NSDataReader.getRawWellRepresentedNeuriteClassesFromFile();
		neuriteGroupContent = new HashMap<String,StringBuilder[]>();
		
		File inputFile, fileTest;
		String baseDestDir;
		baseDestDir = fastaDir;
		fileTest = new File(baseDestDir);
		if (!fileTest.exists()){
			fileTest.mkdir();
		}
		else if (!fileTest.isDirectory()){
			throw new Exception("parameter --fasta-dir/-f is not a directory!");
		}
		// Load metadata of neurons/files
		Map<String,NeuronMetadataComplete> metaData = MetaDataManager.getExtendedCellMetaData();
		//System.out.println("Metadata size: "+metaData.size());
		NeuronMetadataComplete neuronMeta = null;
		Neuron<BinaryTreeNode> currentNeuron;
		Soma<BinaryTreeNode> soma;
		String currentNeuronName;
		Pattern namePattern = Pattern.compile("(.*?)(\\.AutoOnly)?(\\.CNG)?\\.swc");
		Matcher nameMatcher;
		
		// Directory of SWC files to load
		File listFile = new File(swcDir);
		if (!listFile.exists() || !listFile.isDirectory()){
			throw new Exception("swc directory parameter (--swc-dir/-s) does not exist or is not a directory! Value: "+listFile.getAbsolutePath());
		}
		File archiveDir, directoryOfFiles;
		File[] directories = listFile.listFiles();
		File[] fileList;
		
		// Output vars
		File outputFileBunch, outputDir;
		File outputFileAll[] = new File[2];
	    String[] bunchFileData = new String[2];
	    StringBuilder[] allFileData = {new StringBuilder(),new StringBuilder()};
	    int fileCount = 0;

		// Set spur-removal string
		String spurString = "";
		if (spurLength > 0 && (removeDendriteSpurs || removeAxonSpurs || removeApicalSpurs)){
			NumberFormat nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(2);
			spurString = "_SpurLen"+nf.format(spurLength);
			if (!removeDendriteSpurs || !removeAxonSpurs || !removeApicalSpurs){
				if (removeDendriteSpurs){
					spurString += "D";					
				}
				if (removeApicalSpurs){
					spurString += "Ap";					
				}
				if (removeAxonSpurs){
					spurString += "Ax";					
				}
			}
		}
		
		// Set base output directory based on classification and traversal parameters
		baseDestDir = baseDestDir + "/" + dataset+spurString + "/" + classification + "/Mapping" + encoding.getId() + "/" + traversal;
		outputDir = new File(baseDestDir);
		outputDir.mkdirs();

		// Initialize 'All' file
		String allFilename = getAllFilename(dataset+spurString);
		if (isSTL){
			outputFileAll[0] = new File(baseDestDir+"/"+allFilename+"STLMap"+encoding.getId()+".fasta");
			DataWriter.writeStringToFile(outputFileAll[0], "");
		}
		if (isLTS){
			outputFileAll[1] = new File(baseDestDir+"/"+allFilename+"LTSMap"+encoding.getId()+".fasta");
			DataWriter.writeStringToFile(outputFileAll[1], "");
		}

		int count = 0;
	    // Loops through directories of swc file contributors
		System.out.println("Directories: "+directories.length);
		for (int i = 0; i < directories.length; i++){
			// Makes sure we are looking at a directory
			if (directories[i].isDirectory()){
				archiveDir = directories[i];

				// Gets directory containing the swc files
				directoryOfFiles = new File(archiveDir.getAbsolutePath()+"/CNG version/");
				//System.out.println(directoryOfFiles + " " + directoryOfFiles.exists());
				if (!directoryOfFiles.exists()){
					directoryOfFiles = new File(archiveDir.getAbsolutePath()+"/CNGversion");
				}
				if (!directoryOfFiles.exists()){
					directoryOfFiles = new File(archiveDir.getAbsolutePath());
				}
				if (directoryOfFiles.exists()){
					fileList = directoryOfFiles.listFiles();
					// initializes data files that will be fasta formatted output for each of the contributor's data
					bunchFileData[0] = ""; bunchFileData[1] = "";
					
					outputDir = new File(baseDestDir+"/"+archiveDir.getName());
					if (!outputDir.exists()){
						outputDir.mkdir();
					}
					
					// Runs through all files for a given contributor
					System.out.println("Dir "+archiveDir.getName()+" with "+fileList.length+" files");
					for (int j = 0; j < fileList.length; j++){
						fileCount++;
						inputFile = fileList[j];
						if (inputFile.isFile()){
							if (debug) System.out.println("Num "+(++count)+": "+inputFile.getName());
							
							// pull neuron metadata
							nameMatcher = namePattern.matcher(inputFile.getName());
							if (!nameMatcher.matches()){
								System.out.println("ERROR: Name pattern mismatch: "+inputFile.getName()+" (from "+archiveDir.getName()+")");
							}
							else{
								currentNeuronName = nameMatcher.group(1);
								neuronMeta = metaData.get(currentNeuronName);
								//System.out.println(currentNeuronName);
								if (neuronMeta != null && 
										(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(currentNeuronName)) &&
										(!isRemoveNonControl() || neuronMeta.getExperimentCondition().equals("Control")) && 
										(!isRemoveAcuteCulture() || !neuronMeta.getProtocol().equals("Culture") || !neuronMeta.getArchive().equalsIgnoreCase("ascoli"))){
									neuronMeta.setLabName(archiveDir.getName());
									// Convert SWC data to binary trees for axon and dendrites
									soma = SWCReader.convertSwcToBinarySoma(inputFile, classification, encoding.getSizeFromLength(minLength), spurLength>0);

									currentNeuron = new Neuron<BinaryTreeNode>(neuronMeta,soma);

									if (spurLength > 0){
										//System.out.println("SpurLength > 0");
										// Remove spurs (segments as short as or shorter than spurLength
										if (removeApicalSpurs && soma.hasApicalDendriteTree()){
											soma.getApicalDendriteTree().getTreeHead().removeSpurs(spurLength);
										}else if (removeApicalSpurs && currentNeuron.getSoma().hasApicalDendriteTrees()){
											for (BinaryTreeNode root : soma.getApicalDendriteTrees()){
												root.removeSpurs(spurLength);
											}
										}
										if (removeAxonSpurs && soma.hasAxonalTree()){
											soma.getAxonalTree().getTreeHead().removeSpurs(spurLength);
										}
										if (removeDendriteSpurs && soma.hasBasalDendriteTree()){
											soma.getBasalDendriteTree().getTreeHead().removeSpurs(spurLength);
										}else if (removeDendriteSpurs && soma.hasBasalDendriteTrees()){
											for (BinaryTreeNode root : soma.getBasalDendriteTrees()){
												root.removeSpurs(spurLength);
											}
										}
									}
									
									writeNeuronToFastaFiles(
											baseDestDir+"/"+archiveDir.getName(),
											currentNeuron, 
											classification, 
											traversal, order,
											encoding,
											bunchFileData,
											minLength, maxLength, spurString);
									
									// Clear memory
									currentNeuron.setSoma(null);
									System.gc();
									
								}
								else{
									System.out.println("Neuron "+currentNeuronName+" not found in metadata file");
								}
							}
						}
					}
					// Write complete FASTA file(s)
					if (isSTL) allFileData[0] = allFileData[0].append(bunchFileData[0]);
					if (isLTS) allFileData[1] = allFileData[1].append(bunchFileData[1]);

					if (writeArchives){
						if (isSTL){
							outputFileBunch = new File(baseDestDir+"/"+archiveDir.getName()+spurString+".stl.fasta");
							DataWriter.writeStringToFile(outputFileBunch, bunchFileData[0]);
						}
						if (isLTS){
							outputFileBunch = new File(baseDestDir+"/"+archiveDir.getName()+spurString+".lts.fasta");
							DataWriter.writeStringToFile(outputFileBunch, bunchFileData[1]);
						}
					}
				}
			}
		}

		// Write complete file(s)
		if (isSTL){
			DataWriter.appendStringToFile(outputFileAll[0], allFileData[0].toString());
			allFileData[0] = new StringBuilder();
			System.out.println("Wrote to "+outputFileAll[0]);
		}
		if (isLTS){
			DataWriter.appendStringToFile(outputFileAll[1], allFileData[1].toString());
			allFileData[1] = new StringBuilder();
			System.out.println("Wrote to "+outputFileAll[1]);
		}
		
		// Write specific cell type FASTA files
		if (writeCellTypes){
			String className;
			File file;
			StringBuilder[] sb;
			for (Iterator<String> it = neuriteGroupContent.keySet().iterator(); it.hasNext();){
				className = it.next();
				sb = neuriteGroupContent.get(className);
				if (isSTL){
					file = new File(baseDestDir+"/cellTypes/"+className.replace("/", "_")+"-STLMap"+encoding.getId()+".fasta");
					DataWriter.appendStringToFile(file, sb[0].toString());
				}
				if (isLTS){
					file = new File(baseDestDir+"/cellTypes/"+className.replace("/", "_")+"-LTSMap"+encoding.getId()+".fasta");
					DataWriter.appendStringToFile(file, sb[1].toString());
				}
			}
		}
	}

	public void convertSwcFilesToFasta(String swcDir, String outputDirectory)
			throws Exception{
		File fileTest = new File(outputDirectory);
		if (!fileTest.exists()){
			fileTest.mkdir();
		}
		else if (!fileTest.isDirectory()){
			throw new Exception("parameter --fasta-dir/-f is not a directory!");
		}

		StringBuilder filenameBuilder = new StringBuilder();
		filenameBuilder.append(outputDirectory+"/");

		File swcDirectory = new File(swcDir);
		filenameBuilder.append(swcDirectory.getName().trim().replaceAll(" ","_"));

		// Set spur-removal string
		String spurString = "";
		if (spurLength > 0 && (removeDendriteSpurs || removeAxonSpurs || removeApicalSpurs)){
			NumberFormat nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(2);
			spurString = "_SpurLen"+nf.format(spurLength);
			if (!removeDendriteSpurs || !removeAxonSpurs || !removeApicalSpurs){
				if (removeDendriteSpurs){
					spurString += "D";
				}
				if (removeApicalSpurs){
					spurString += "Ap";
				}
				if (removeAxonSpurs){
					spurString += "Ax";
				}
			}
		}
		filenameBuilder.append(spurString);
		filenameBuilder.append("_Encoding"+encoding.getId());
		filenameBuilder.append(".fasta");

		File fastaFile = new File(filenameBuilder.toString());
		convertSwcFilesToFasta(swcDir, fastaFile);
	}

	public void convertSwcFilesToFasta(String swcDir, File fastaFile)
			throws Exception {
		// Validate Arguments
		if (classification == null) {
			throw new Exception("Classification " + classification + " unknown; must be one of {"+NeuronTreeClassification.allNames()+"}");
		}
		if (!traversal.equals("Size")
				&& !traversal.equals("Height")
				&& !traversal.equals("AvgHeight")) {
			throw new Exception("Traversal " + traversal + " unknown; must be one of {Sise, Height, AvgHeight}");
		}

		neuriteGroupContent = new HashMap<String, StringBuilder[]>();

		File inputFile;
		TraversalOrder traversalOrder = TraversalOrder.find(this.order);

		Neuron<BinaryTreeNode> currentNeuron;
		Soma<BinaryTreeNode> soma;
		Pattern namePattern = Pattern.compile("(.*?)(\\.AutoOnly)?(\\.CNG)?\\.swc");
		Matcher nameMatcher;

		// Directory of SWC files to load
		File listFile = new File(swcDir);
		if (!listFile.exists() || !listFile.isDirectory()) {
			throw new Exception("swc directory parameter (--swc-dir/-s) does not exist or is not a directory! Value: " + listFile.getAbsolutePath());
		}
		File[] fileList;

		// Output vars
		StringBuilder allFileData = new StringBuilder();

		int count = 0;
		// Loops through directories of swc file contributors
		File swcDirF = new File(swcDir);
		fileList = swcDirF.listFiles();

		for (int j = 0; j < fileList.length; j++) {
			inputFile = fileList[j];
			if (inputFile.isFile()) {
				if (debug) System.out.println("Num " + (++count) + ": " + inputFile.getName());

				// pull neuron metadata
				nameMatcher = namePattern.matcher(inputFile.getName());
				if (!nameMatcher.matches()) {
					System.out.println("ERROR: Name pattern mismatch: " + inputFile.getName());
				} else {
					// Convert SWC data to binary trees for axon and dendrites
					soma = SWCReader.convertSwcToBinarySoma(inputFile, classification, encoding.getSizeFromLength(minLength), spurLength > 0);

					currentNeuron = new Neuron<BinaryTreeNode>(soma);
					currentNeuron.setName(nameMatcher.group(1));

					if (spurLength > 0) {
						// Remove spurs (segments as short as or shorter than spurLength
						if (removeApicalSpurs && soma.hasApicalDendriteTree()) {
							soma.getApicalDendriteTree().getTreeHead().removeSpurs(spurLength);
						} else if (removeApicalSpurs && currentNeuron.getSoma().hasApicalDendriteTrees()) {
							for (BinaryTreeNode root : soma.getApicalDendriteTrees()) {
								root.removeSpurs(spurLength);
							}
						}
						if (removeAxonSpurs && soma.hasAxonalTree()) {
							soma.getAxonalTree().getTreeHead().removeSpurs(spurLength);
						}
						if (removeDendriteSpurs && soma.hasBasalDendriteTree()) {
							soma.getBasalDendriteTree().getTreeHead().removeSpurs(spurLength);
						} else if (removeDendriteSpurs && soma.hasBasalDendriteTrees()) {
							for (BinaryTreeNode root : soma.getBasalDendriteTrees()) {
								root.removeSpurs(spurLength);
							}
						}
					}
					allFileData.append(
							convertNeuronToFasta(
								currentNeuron,
								classification,
								traversal, traversalOrder,
								encoding,
								minLength, maxLength));

					// Clear memory
					currentNeuron.setSoma(null);
					System.gc();

				}
			}
		}
		// Write complete file(s)
		DataWriter.appendStringToFile(fastaFile, allFileData.toString());
		System.out.println("Wrote to " + fastaFile.getAbsolutePath());
	}

	private void addToGroupString(String key, String[] outputs){
		if (neuriteTypes.contains(key)){
			StringBuilder[] sb;
			if (neuriteGroupContent.containsKey(key)){
				sb = neuriteGroupContent.get(key);
			}
			else{
				sb = new StringBuilder[2];
				sb[0] = new StringBuilder();
				sb[1] = new StringBuilder();
				neuriteGroupContent.put(key, sb);
			}
			sb[0].append(outputs[0]);
			sb[1].append(outputs[1]);
		}
	}
	
	public void writeNeuronToFastaFiles(String baseDirectory,
			Neuron<BinaryTreeNode> neuron, NeuronTreeClassification classification, String traversal, String order, 
			EncodingMode encoding, String[] bunchFileData, int minLength, int maxLength, String spurString)
	throws IOException, Exception{

		String[] outputs = new String[2];
		Soma<BinaryTreeNode> soma = neuron.getSoma();
		NeuronMetadataComplete metadata = (NeuronMetadataComplete)neuron.getMetadata();
		String neuronName = neuron.getName();
		
		//System.out.println(baseDirectory +" " +neuron.getName());
		
		StringBuilder header;
		String class2 = metadata.getClass1()+"/"+metadata.getClass2(), 
			brainRegion = metadata.getClass1()+"/"+metadata.getBrainRegion(),
			class2Region = class2+"/"+metadata.getBrainRegion();


		switch(classification){
		case NEURON:
			if (soma.getSomaRoot() != null && soma.getSomaRoot().getTreeHead() != null
				&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(metadata,TreeType.CELL));
				header.append("\n");

				outputs = writeTree(baseDirectory+"/"+neuron.getName(),
						soma.getSomaRoot().getTreeHead(),
						header.toString(),
						traversal, order,
						encoding,
						bunchFileData,
						minLength, maxLength, spurString, lineLength);

			}
			break;
		// Putting these three together to process axon, then separating out the classifications with a sub-switch
		case TREE:
		case SUB_ARBORIZATION:
		case ARBORIZATION:
			if (soma.getAxonalTree() != null && soma.getAxonalTree().getTreeHead() != null
			&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Axon"))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(metadata,TreeType.AXON));
				header.append("\n");

				outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Axon",
						soma.getAxonalTree().getTreeHead(),
						header.toString(),
						traversal, order,
						encoding,
						bunchFileData,
						minLength, maxLength, spurString, lineLength);
		
				addToGroupString("Axon", outputs);
				addToGroupString("Axon/"+metadata.getClass1(), outputs);
				addToGroupString("Axon/"+class2, outputs);
				addToGroupString("Axon/"+brainRegion, outputs);
				addToGroupString("Axon/"+class2Region, outputs);
			}/*else{
				System.out.println("File "+neuron.getName()+" has no axon tree information");
			}*/

			// Sub-switch for Arborization, Sub-arborization, and Tree
			switch(classification){
			case TREE:
				// Create file for each basal dendrite tree that comes from the soma
				if (soma.getBasalDendriteTrees() != null && soma.getBasalDendriteTrees().size() > 0 && 
						(!isRemoveFlaggedNeurites() || 
								(!flaggedNeurites.contains(neuronName+"-BasalDendrite") && !flaggedNeurites.contains(neuronName+"-Dendrite")))){
					for (int i = 0; i < soma.getBasalDendriteTrees().size(); i++){
						// create header for fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TreeType.BDTREE,i+1));
						header.append("\n");

						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-BasalDendrite"+(i+1),
								(BinaryTreeNode)soma.getBasalDendriteTrees().get(i),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);

					}
				}

				// Create file for each apical dendrite tree that comes from the soma
				if (soma.getApicalDendriteTrees() != null && soma.getApicalDendriteTrees().size() > 0 && 
						(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-ApicalDendrite"))){
					for (int i = 0; i < soma.getApicalDendriteTrees().size(); i++){
						// create header for fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TreeType.ADTREE,i+1));
						header.append("\n");

						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-ApicalDendrite"+(i+1),
								(BinaryTreeNode)soma.getApicalDendriteTrees().get(i),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);

					}
				}	
				break;
				
			case SUB_ARBORIZATION:
				NeuronProcess<BinaryTreeNode> generalDendrite = null;
				if (soma.hasApicalDendriteTree() || soma.hasBasalDendriteTree()){
					//if (!(metadata.getCellType().equalsIgnoreCase("Pyramidal cell") || metadata.getCellType().equalsIgnoreCase("Pyramidal-like"))){
					if (!(metadata.getCellType().equalsIgnoreCase("Pyramidal cell"))){
						if (soma.hasApicalDendriteTree() && !soma.hasBasalDendriteTree()){
							generalDendrite = soma.getApicalDendriteTree();
						}
						else if (soma.hasBasalDendriteTree() && !soma.hasApicalDendriteTree()){
							generalDendrite = soma.getBasalDendriteTree();
						}
					}
				}
				
				if (generalDendrite != null){
					if (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Dendrite")){
						header = new StringBuilder();
						header.append(createHeader(metadata,TreeType.GENERAL_DENDRITE));
						header.append("\n");
		
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Dendrite",
								generalDendrite.getTreeHead(),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);

//						if (neuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
//							System.out.println(header.toString()+" Pyramidal cell w/o Apical/Basal trees");
//						}
						addToGroupString("Dendrite", outputs);
						addToGroupString("Dendrite/"+metadata.getClass1(), outputs);
						addToGroupString("Dendrite/"+class2, outputs);
						addToGroupString("Dendrite/"+brainRegion, outputs);
						addToGroupString("Dendrite/"+class2Region, outputs);
					}
				}
				else{
					// write basal dendrite file
					if (soma.hasBasalDendriteTree() && 
							(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-BasalDendrite"))){
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TreeType.BASAL_DENDRITE));
						header.append("\n");
			
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-BasalDendrite",
								soma.getBasalDendriteTree().getTreeHead(),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);
						if (outputs[0] != null || outputs[1] != null){
							addToGroupString("BasalDendrite", outputs);
							addToGroupString("BasalDendrite/"+metadata.getClass1(), outputs);
							addToGroupString("BasalDendrite/"+class2, outputs);
							addToGroupString("BasalDendrite/"+brainRegion, outputs);
							addToGroupString("BasalDendrite/"+class2Region, outputs);
						}
					}
					
					if (soma.hasApicalDendriteTree() && 
							(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-ApicalDendrite"))){
						// write apical dendrite file
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TreeType.APICAL_DENDRITE));
						header.append("\n");
		
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-ApicalDendrite",
							soma.getApicalDendriteTree().getTreeHead(),
							header.toString(),
							traversal, order,
							encoding,
							bunchFileData,
							minLength, maxLength, spurString, lineLength);
					
//						if (!neuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
//							System.out.println(header.toString()+" non-Pyramidal cell with Apical/Basal trees");
//						}
						if (outputs[0] != null || outputs[1] != null){
							addToGroupString("ApicalDendrite", outputs);
							addToGroupString("ApicalDendrite/"+metadata.getClass1(), outputs);
							addToGroupString("ApicalDendrite/"+class2, outputs);
							addToGroupString("ApicalDendrite/"+brainRegion, outputs);
							addToGroupString("ApicalDendrite/"+class2Region, outputs);
						}
					}
				}
				break;
				
			case ARBORIZATION:
				if (soma.getBasalDendriteTree() != null && soma.getBasalDendriteTree().getTreeHead() != null
				&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Dendrite"))){
					header = new StringBuilder();
					header.append(createHeader(metadata,TreeType.GENERAL_DENDRITE));
					header.append("\n");
	
					outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Dendrite",
							soma.getBasalDendriteTree().getTreeHead(),
							header.toString(),
							traversal, order,
							encoding,
							bunchFileData,
							minLength, maxLength, spurString, lineLength);

//					if (neuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
//						System.out.println(header.toString()+" Pyramidal cell w/o Apical/Basal trees");
//					}
					addToGroupString("Dendrite", outputs);
					addToGroupString("Dendrite/"+metadata.getClass1(), outputs);
					addToGroupString("Dendrite/"+class2, outputs);
					addToGroupString("Dendrite/"+brainRegion, outputs);
					addToGroupString("Dendrite/"+class2Region, outputs);
				}
				break;
			default:
				// Can't get here	
			}
			
			break;
		}
		/*
		if (classification.toLowerCase().equals("arborization")){
			// write axon file
			if (soma.getAxonalTree() != null && soma.getAxonalTree().getTreeHead() != null
					&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Axon"))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(metadata,TREE_TYPE_AXON));
				header.append("\n");

				outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Axon",
						soma.getAxonalTree().getTreeHead(),
						header.toString(),
						traversal, order,
						encoding,
						bunchFileData,
						minLength, maxLength, spurString, lineLength);
				
				addToGroupString("Axon", outputs);
				addToGroupString("Axon/"+metadata.getClass1(), outputs);
				addToGroupString("Axon/"+class2, outputs);
				addToGroupString("Axon/"+brainRegion, outputs);
				addToGroupString("Axon/"+class2Region, outputs);
			}
			else{
				//System.out.println("File "+neuron.getName()+" has no axon tree information");
			}
			
			// Check if cell has basal/apical differentiation
			NeuronProcess<BinaryTreeNode> generalDendrite = null;
			if (!soma.hasApicalDendriteTree() && !soma.hasBasalDendriteTree()){
				//System.out.println(neuron.getName()+" has no dendrite information");
			}
			else{
				//if (!(metadata.getCellType().equalsIgnoreCase("Pyramidal cell") || metadata.getCellType().equalsIgnoreCase("Pyramidal-like"))){
				if (!(metadata.getCellType().equalsIgnoreCase("Pyramidal cell"))){
					if (soma.hasApicalDendriteTree() && !soma.hasBasalDendriteTree()){
						generalDendrite = soma.getApicalDendriteTree();
					}
					else if (soma.hasBasalDendriteTree() && !soma.hasApicalDendriteTree()){
						generalDendrite = soma.getBasalDendriteTree();
					}
				}
				
				if (generalDendrite != null){
					if (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Dendrite")){
						header = new StringBuilder();
						header.append(createHeader(metadata,TREE_TYPE_GENERAL_DENDRITE));
						header.append("\n");
		
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Dendrite",
								generalDendrite.getTreeHead(),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);

//						if (neuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
//							System.out.println(header.toString()+" Pyramidal cell w/o Apical/Basal trees");
//						}
						addToGroupString("Dendrite", outputs);
						addToGroupString("Dendrite/"+metadata.getClass1(), outputs);
						addToGroupString("Dendrite/"+class2, outputs);
						addToGroupString("Dendrite/"+brainRegion, outputs);
						addToGroupString("Dendrite/"+class2Region, outputs);
					}
				}
				else{
					// write basal dendrite file
					if (soma.hasBasalDendriteTree() && 
							(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-BasalDendrite"))){
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TREE_TYPE_BASAL_DENDRITE));
						header.append("\n");
			
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-BasalDendrite",
								soma.getBasalDendriteTree().getTreeHead(),
								header.toString(),
								traversal, order,
								encoding,
								bunchFileData,
								minLength, maxLength, spurString, lineLength);
						if (outputs[0] != null || outputs[1] != null){
							addToGroupString("BasalDendrite", outputs);
							addToGroupString("BasalDendrite/"+metadata.getClass1(), outputs);
							addToGroupString("BasalDendrite/"+class2, outputs);
							addToGroupString("BasalDendrite/"+brainRegion, outputs);
							addToGroupString("BasalDendrite/"+class2Region, outputs);
						}
					}
					
					if (soma.hasApicalDendriteTree() && 
							(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-ApicalDendrite"))){
						// write apical dendrite file
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(metadata,TREE_TYPE_APICAL_DENDRITE));
						header.append("\n");
		
						outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-ApicalDendrite",
							soma.getApicalDendriteTree().getTreeHead(),
							header.toString(),
							traversal, order,
							encoding,
							bunchFileData,
							minLength, maxLength, spurString, lineLength);
					
//						if (!neuron.getCellType().equalsIgnoreCase("Pyramidal cell")){
//							System.out.println(header.toString()+" non-Pyramidal cell with Apical/Basal trees");
//						}
						if (outputs[0] != null || outputs[1] != null){
							addToGroupString("ApicalDendrite", outputs);
							addToGroupString("ApicalDendrite/"+metadata.getClass1(), outputs);
							addToGroupString("ApicalDendrite/"+class2, outputs);
							addToGroupString("ApicalDendrite/"+brainRegion, outputs);
							addToGroupString("ApicalDendrite/"+class2Region, outputs);
						}
					}
				}
			}
		}else if (classification.toLowerCase().equals("tree")){
			// write axon file (axon should have only one tree)
			if (soma.getAxonalTree() != null && 
					(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Axon"))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(metadata,TREE_TYPE_AXON));
				header.append("\n");

				outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-Axon",
						soma.getAxonalTree().getTreeHead(),
						header.toString(),
						traversal, order,
						encoding,
						bunchFileData,
						minLength, maxLength, spurString, lineLength);
			}
			else{
				System.out.println("File "+neuron.getName()+" has no axon tree information");
			}

			// Create file for each basal dendrite tree that comes from the soma
			if (soma.getBasalDendriteTrees() != null && soma.getBasalDendriteTrees().size() > 0 && 
					(!isRemoveFlaggedNeurites() || 
							(!flaggedNeurites.contains(neuronName+"-BasalDendrite") && !flaggedNeurites.contains(neuronName+"-Dendrite")))){
				for (int i = 0; i < soma.getBasalDendriteTrees().size(); i++){
					// create header for fasta files
					header = new StringBuilder();
					header.append(createHeader(metadata,TREE_TYPE_BDTREE,i+1));
					header.append("\n");

					outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-BasalDendrite"+(i+1),
							(BinaryTreeNode)soma.getBasalDendriteTrees().get(i),
							header.toString(),
							traversal, order,
							encoding,
							bunchFileData,
							minLength, maxLength, spurString, lineLength);

				}
			}

			// Create file for each apical dendrite tree that comes from the soma
			if (soma.getApicalDendriteTrees() != null && soma.getApicalDendriteTrees().size() > 0 && 
					(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-ApicalDendrite"))){
				for (int i = 0; i < soma.getApicalDendriteTrees().size(); i++){
					// create header for fasta files
					header = new StringBuilder();
					header.append(createHeader(metadata,TREE_TYPE_ADTREE,i+1));
					header.append("\n");

					outputs = writeTree(baseDirectory+"/"+neuron.getName()+"-ApicalDendrite"+(i+1),
							(BinaryTreeNode)soma.getApicalDendriteTrees().get(i),
							header.toString(),
							traversal, order,
							encoding,
							bunchFileData,
							minLength, maxLength, spurString, lineLength);

				}
			}

		}*/
	}

	/**
	 *
	 * @param fullFileName
	 * @param treeHead
	 * @param header
	 * @param traversal
	 * @param order
	 * @param encoding
	 * @param bunchFileData
	 * @param minLength
	 * @param maxLength
	 * @param spurString
	 * @param lineLength
     * @return
     * @throws IOException
     * @throws Exception
     */
	private static String[] writeTree(
			String fullFileName, 
			BinaryTreeNode treeHead, 
			String header, 
			String traversal,
			String order,
			EncodingMode encoding,
			String[] bunchFileData,
			int minLength, int maxLength, String spurString, int lineLength)
	throws IOException, Exception{
		
	    // Add data to larger data file
		String[] outputs = null;
		outputs = new String[2];
		NodeEncoder nodeEncoder = NodeEncoderFactory.createNewNodeEncoder(encoding);

		int seqLength = 1;
		if (!treeHead.isLeaf()){
			seqLength = encoding.getLengthFromSize(treeHead.getTreeSize());
		}
		
		if (seqLength >= minLength && (maxLength == 0 || seqLength <= maxLength)){
			String sequence;
			
			StringBuilder output;
			if (isSTL(order)){
				sequence = nodeEncoder.encodeTree(treeHead, "STL");
				output = new StringBuilder();
				output.append(header);
				if (lineLength > 0){
					output.append(StringUtils.insertBreaksAtLength(sequence, 80));
				}else{
					output.append(sequence);
				}
				output.append("\n");
				// Output to file
				DataWriter.writeStringToFile(
						fullFileName+spurString+".stl.fasta", output.toString());
				outputs[0] = output.toString();
				bunchFileData[0] = bunchFileData[0].concat(output.toString());
			}
			if (isLTS(order)){
				sequence = nodeEncoder.encodeTree(treeHead, "LTS");
				output = new StringBuilder();
				output.append(header);
				if (lineLength > 0){
					output.append(StringUtils.insertBreaksAtLength(sequence, 80));
				}else{
					output.append(sequence);
				}
				output.append("\n");
				// Output to file
				DataWriter.writeStringToFile(
						fullFileName+spurString+".lts.fasta", output.toString());
				outputs[1] = output.toString();
				bunchFileData[1] = bunchFileData[1].concat(output.toString());
			}
		}
		else if (debug){
			if (seqLength < minLength){
				System.out.println("Tree too small (less than "+minLength+"): \n"+header);
			}
			else if (maxLength != 0 && seqLength > maxLength){
				System.out.println("Tree too big (greater than "+maxLength+"): \n"+header);
			}
		}
		return outputs;
	}

	/**
	 * For the simpler case of writing all trees of a neuron (based on value of 'classification') into a single file
	 * @param neuron
	 * @param classification
	 * @param traversal
	 * @param order
	 * @param encoding
	 * @param minLength
	 * @param maxLength
     * @throws IOException
     * @throws Exception
     */
	public String convertNeuronToFasta(Neuron<BinaryTreeNode> neuron, NeuronTreeClassification classification, String traversal, TraversalOrder order,
										EncodingMode encoding, int minLength, int maxLength)
			throws IOException, Exception{

		StringBuilder sb = new StringBuilder();
		Soma<BinaryTreeNode> soma = neuron.getSoma();
		String neuronName = neuron.getName();
		NodeEncoder nodeEncoder = NodeEncoderFactory.createNewNodeEncoder(encoding);

		//System.out.println(baseDirectory +" " +neuron.getName());

		StringBuilder header;

		switch(classification){
		case NEURON:
			if (soma.getSomaRoot() != null && soma.getSomaRoot().getTreeHead() != null
				&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(neuronName,TreeType.CELL));
				header.append("\n");

				sb.append(
						writeTree(
								soma.getSomaRoot().getTreeHead(),
								header.toString(),
								traversal, order,
								nodeEncoder,
								minLength, maxLength, lineLength));
			}
			break;
		// Putting these three together to process axon, then separating out the classifications with a sub-switch
		case TREE:
		case SUB_ARBORIZATION:
		case ARBORIZATION:
			if (soma.getAxonalTree() != null && soma.getAxonalTree().getTreeHead() != null
			&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Axon"))){
				// create header for both fasta files
				header = new StringBuilder();
				header.append(createHeader(neuronName,TreeType.AXON));
				header.append("\n");

				sb.append(
						writeTree(
							soma.getAxonalTree().getTreeHead(),
							header.toString(),
							traversal, order,
							nodeEncoder,
							minLength, maxLength, lineLength));
			}/*else{
				System.out.println("File "+neuron.getName()+" has no axon tree information");
			}*/

			// Sub-switch for Arborization, Sub-arborization, and Tree
			switch(classification){
			case TREE:
				// Create file for each basal dendrite tree that comes from the soma
				if (soma.getBasalDendriteTrees() != null && soma.getBasalDendriteTrees().size() > 0 && 
						(!isRemoveFlaggedNeurites() || 
								(!flaggedNeurites.contains(neuronName+"-BasalDendrite") && !flaggedNeurites.contains(neuronName+"-Dendrite")))){
					for (int i = 0; i < soma.getBasalDendriteTrees().size(); i++){
						// create header for fasta files
						header = new StringBuilder();
						header.append(createHeader(neuronName,TreeType.BDTREE,i+1));
						header.append("\n");

						sb.append(
								writeTree(
									(BinaryTreeNode)soma.getBasalDendriteTrees().get(i),
									header.toString(),
									traversal, order,
									nodeEncoder,
									minLength, maxLength, lineLength));

					}
				}

				// Create file for each apical dendrite tree that comes from the soma
				if (soma.getApicalDendriteTrees() != null && soma.getApicalDendriteTrees().size() > 0 && 
						(!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-ApicalDendrite"))){
					for (int i = 0; i < soma.getApicalDendriteTrees().size(); i++){
						// create header for fasta files
						header = new StringBuilder();
						header.append(createHeader(neuronName,TreeType.ADTREE,i+1));
						header.append("\n");

						sb.append(
								writeTree(
									(BinaryTreeNode)soma.getApicalDendriteTrees().get(i),
									header.toString(),
									traversal, order,
									nodeEncoder,
									minLength, maxLength, lineLength));

					}
				}	
				break;
				
			case SUB_ARBORIZATION:
				NeuronProcess<BinaryTreeNode> generalDendrite = null;
				if (soma.hasBasalDendriteTree() && !soma.hasApicalDendriteTree()){
					generalDendrite = soma.getBasalDendriteTree();
					if (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Dendrite")){
						header = new StringBuilder();
						header.append(createHeader(neuronName,TreeType.GENERAL_DENDRITE));
						header.append("\n");

						sb.append(
								writeTree(generalDendrite.getTreeHead(),
										header.toString(),
										traversal, order, nodeEncoder,
										minLength, maxLength, lineLength));
					}
				}
				else{
					// write basal dendrite file
					if (soma.hasBasalDendriteTree()){
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(neuronName,TreeType.BASAL_DENDRITE));
						header.append("\n");

						sb.append(
								writeTree(soma.getBasalDendriteTree().getTreeHead(),
										header.toString(),
										traversal, order,
										nodeEncoder,
										minLength, maxLength, lineLength));
					}

					if (soma.hasApicalDendriteTree()){
						// write apical dendrite file
						// create header for both fasta files
						header = new StringBuilder();
						header.append(createHeader(neuronName,TreeType.APICAL_DENDRITE));
						header.append("\n");

						sb.append(
								writeTree(soma.getApicalDendriteTree().getTreeHead(),
										header.toString(),
										traversal, order,
										nodeEncoder,
										minLength, maxLength, lineLength));
					}
				}
				break;
				
			case ARBORIZATION:
				if (soma.getBasalDendriteTree() != null && soma.getBasalDendriteTree().getTreeHead() != null
				&& (!isRemoveFlaggedNeurites() || !flaggedNeurites.contains(neuronName+"-Dendrite"))){
					header = new StringBuilder();
					header.append(createHeader(neuronName,TreeType.GENERAL_DENDRITE));
					header.append("\n");
	
					sb.append(
							writeTree(
								soma.getBasalDendriteTree().getTreeHead(),
								header.toString(),
								traversal, order,
								nodeEncoder,
								minLength, maxLength, lineLength));
				}
				break;
			default:
				// Can't get here	
			}
			
			break;
		}

		return sb.toString();
	}

	/**
	 *
	 * @param treeHead
	 * @param header
	 * @param traversal
	 * @param order
	 * @param nodeEncoder
	 * @param minLength
	 * @param maxLength
	 * @param lineLength
	 * @return
     * @throws IOException
     * @throws Exception
     */
	private static String writeTree(
			BinaryTreeNode treeHead,
			String header,
			String traversal,
			TraversalOrder order,
			NodeEncoder nodeEncoder,
			int minLength, int maxLength, int lineLength)
			throws IOException, Exception{

		int seqLength = 1;
		if (!treeHead.isLeaf()){
			seqLength = nodeEncoder.getLengthFromSize(treeHead.getTreeSize());
		}

		if (seqLength >= minLength && (maxLength == 0 || seqLength <= maxLength)){
			String sequence;

			sequence = nodeEncoder.encodeTree(treeHead, order);
			StringBuilder output = new StringBuilder();
			output.append(header);
			if (lineLength > 0){
				output.append(StringUtils.insertBreaksAtLength(sequence, 80));
			}else{
				output.append(sequence);
			}
			output.append("\n");
			return output.toString();
		}
		return "";
	}

	private static String createHeader(NeuronMetadata neuron, TreeType type){
		return createHeader(neuron,type,0);
	}
	private static String createHeader(NeuronMetadata neuron, TreeType type, int treeNumber){
		StringBuilder sb = new StringBuilder();
		sb.append(">"+neuron.getName());
		switch (type){
		case AXON:
			sb.append("-Axon");
			break;
		case BASAL_DENDRITE:
			sb.append("-BasalDendrite");
			break;
		case APICAL_DENDRITE:
			sb.append("-ApicalDendrite");
			break;
		case GENERAL_DENDRITE:
			sb.append("-Dendrite");
			break;
		case BDTREE:
			sb.append("-BasalDendrite"+treeNumber);
			break;
		case ADTREE:
			sb.append("-ApicalDendrite"+treeNumber);
			break;
		default:
			break;
		}
		sb.append(" | ");
		sb.append(neuron.getCellType());
		sb.append(" | ");
		sb.append(neuron.getBrainRegion());
		sb.append(" | ");
		sb.append(neuron.getAnimal());
		sb.append(" | ");
		sb.append(neuron.getLabName());
		return sb.toString();
	}

	private static String createHeader(String neuronName, TreeType type) {
		return createHeader(neuronName, type, 0);
	}
	private static String createHeader(String neuronName, TreeType type, int treeNumber) {
		StringBuilder sb = new StringBuilder();
		sb.append(">" + neuronName);
		switch (type) {
			case AXON:
				sb.append("-Axon");
				break;
			case BASAL_DENDRITE:
				sb.append("-BasalDendrite");
				break;
			case APICAL_DENDRITE:
				sb.append("-ApicalDendrite");
				break;
			case GENERAL_DENDRITE:
				sb.append("-Dendrite");
				break;
			case BDTREE:
				sb.append("-BasalDendrite" + treeNumber);
				break;
			case ADTREE:
				sb.append("-ApicalDendrite" + treeNumber);
				break;
			default:
				break;
		}
		return sb.toString();
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        Switch sw;
        
        opt = new FlaggedOption(NSConstants.PARAM_TRAVERSAL)
            .setStringParser(JSAP.STRING_PARSER)
            .setRequired(false)
            .setDefault(NSConstants.DEFAULT_TRAVERSAL)
            .setShortFlag('t') 
            .setLongFlag(NSConstants.PARAM_TRAVERSAL);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
	        .setStringParser(JSAP.STRING_PARSER)
	        .setRequired(false)
	        .setDefault("Both")
	        .setShortFlag('o') 
	        .setLongFlag(NSConstants.PARAM_ORDER);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
    	.setStringParser(JSAP.STRING_PARSER)
    	.setRequired(false)
    	.setDefault(""+NSConstants.DEFAULT_MAPPING)
    	.setShortFlag('M') 
    	.setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_CLASSIFICATION)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(false)
        	.setDefault(NSConstants.DEFAULT_CLASSIFICATION.toString())
        	.setShortFlag('c') 
        	.setLongFlag(NSConstants.PARAM_CLASSIFICATION);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MIN_LENGTH)
        	.setStringParser(JSAP.INTEGER_PARSER)
        	.setRequired(false)
        	.setDefault(""+DEFAULT_MIN_LENGTH)
        	.setShortFlag('m') 
        	.setLongFlag(PARAM_MIN_LENGTH);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(PARAM_SWC_DIR)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(true) 
    		.setShortFlag('s') 
    		.setLongFlag(PARAM_SWC_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FASTA_DIR)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(true)
			.setShortFlag('f') 
			.setLongFlag(PARAM_FASTA_DIR);
        jsap.registerParameter(opt);
        
        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(false)
			.setDefault(NSConstants.DEFAULT_DATASET)
			.setShortFlag('d') 
			.setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SPUR_LENGTH)
		.setStringParser(JSAP.FLOAT_PARSER)
		.setRequired(false)
		.setDefault("0")
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_SPUR_LENGTH);
        jsap.registerParameter(opt);

        sw = new Switch(PARAM_REMOVE_DENDRITE_SPURS)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_DENDRITE_SPURS);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_APICAL_SPURS)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_APICAL_SPURS);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_AXON_SPURS)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_AXON_SPURS);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_FLAGGED_NEURITES)
        .setDefault("true")
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_FLAGGED_NEURITES);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_NON_CONTROL)
        .setDefault("true")
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_NON_CONTROL);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_ACUTE_CULTURE)
        .setDefault("true")
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(PARAM_REMOVE_ACUTE_CULTURE);
        jsap.registerParameter(sw);

        sw = new Switch(NSConstants.PARAM_DEBUG)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(NSConstants.PARAM_DEBUG);
        jsap.registerParameter(sw);
        
        opt = new FlaggedOption(PARAM_WRITE_ARCHIVES)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_WRITE_ARCHIVES)
			.setShortFlag(JSAP.NO_SHORTFLAG) 
			.setLongFlag(PARAM_WRITE_ARCHIVES);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_WRITE_CELLTYPES)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_WRITE_CELLTYPES)
			.setShortFlag(JSAP.NO_SHORTFLAG) 
			.setLongFlag(PARAM_WRITE_CELLTYPES);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_GENERATE_DATABASE)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_GENERATE_DATABASE)
			.setShortFlag(JSAP.NO_SHORTFLAG) 
			.setLongFlag(PARAM_GENERATE_DATABASE);
	    jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SwcToFastaConverter.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}
	
	private static boolean isSTL(TraversalOrder order){
		return order != TraversalOrder.LTS;
	}

	private static boolean isLTS(TraversalOrder order){
		return order != TraversalOrder.STL;
	}

	private static boolean isSTL(String order){
		return !order.equalsIgnoreCase("LTS");
	}

	private static boolean isLTS(String order){
		return !order.equalsIgnoreCase("STL");
	}

}
