package org.cn3.ns.generation;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappingUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Generates a FASTA file containing surrogates of an original "real" FASTA file (of neurite-derived sequences).
 * Allows surrogates to be based on selected properties of the real sequences, depending on the chosen
 * baseline (i.e. tree generation) method. Current options (using -b/--baseline) include:
 * 2 constrained by real sequence lengths, generated using terminal branch growth 
 * 3 constrained by real sequence lengths, generated using Harding method (uniform distribution of tree shapes)
 * 5 constrained by real sequence lengths and monomer frequencies
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBOfSource -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 3 -s 1
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBOfSource -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 5 -s 1
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBOfSource -i /homes/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 5 -s 1
 */
public class CreateRandomDBOfSource {

	public static final String PARAM_INPUT = "input";
	public static final String PARAM_DATABASE_NAME = "database-name";
	public static final String PARAM_SET_NUM = "set-num";

	private static String order;
	private static int baseline;
	private static File originalFasta;
	private static int setNum;
	private static File outfile;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			
			order = config.getString(NSConstants.PARAM_ORDER);
			baseline = config.getInt(NSConstants.PARAM_BASELINE);
			originalFasta = new File(config.getString(PARAM_INPUT));
			setNum = config.getInt(PARAM_SET_NUM);

			Pattern p = Pattern.compile("(.*)\\.fasta");
			Matcher m = p.matcher(originalFasta.getName());
			if (m.matches()){
				outfile = new File(originalFasta.getParent()+"/"+m.group(1)
						+"-randomsB"+baseline+"_"+setNum+".fasta");
			}
			else{
				throw new Exception("First parameter not a Fasta file");
			}

			FastaReader fr = new FastaReader(originalFasta);

			int length, cCount;
			String randomSequence;
			StringBuffer sb = new StringBuffer();
			while (fr.next()) {
				length = new Integer(fr.getSequenceLength());
				switch (baseline){
				case NSConstants.DISTRIBUTION_RANDOM_SEQUENCE:
					randomSequence = generateRandomACTSequence(length);
					break;
				case NSConstants.DISTRIBUTION_TERMINAL_GROWTH:
				case NSConstants.DISTRIBUTION_UNIFORM:
					randomSequence = generateRandomTreeSequence(length, baseline, order);
					break;
				case NSConstants.DISTRIBUTION_RESIDUE_FREQ_CONSTRAINED:
					cCount = MotifFinder.findMotifs(fr.getSequence(), 1).get("C").intValue();
					randomSequence = generateBackgroundFrequencyRandomSequence(length, cCount, order);
					break;
				default:
					throw new Exception("Unsupported random generation method: "+baseline);
				}
				sb.append(fr.getHeader()+"\n");
				sb.append(randomSequence+"\n");
			}

			String randomDatabase = sb.toString();

			DataWriter.writeStringToFile(outfile, randomDatabase);
			System.out.println("Wrote file "+outfile);
		
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static String generateRandomACTSequence(int length){
		String randomSequence = "";
		for (int i = 0; i < length; i++){
			switch((int)(3*Math.random())){
			case 0:
				randomSequence += "A";
				break;
			case 1:
				randomSequence += "C";
				break;
			case 2:
				randomSequence += "T";
				break;
			}
		}
		return randomSequence;
	}

	private static String generateRandomTreeSequence(int length, int baseline, String order) 
	throws Exception{
		int degree = length+1;
		BinaryTreeNode headNode;
		switch(baseline){
		case NSConstants.DISTRIBUTION_TERMINAL_GROWTH:
			headNode = NodeSequenceBinaryTreeGenerator.generateTerminalGrowthBinaryTreeByDegree(degree);
			break;
		case NSConstants.DISTRIBUTION_UNIFORM:
			headNode = NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeByDegree(degree);
			break;
		default:
			return null;
		}
		return(MappingUtils.traverseTreeToString(
				headNode, NSConstants.MAPPING_MODE_TOPO_ACT, order));
	}

	private static String generateBackgroundFrequencyRandomSequence(int length, int cCount, String order) 
	throws Exception{
		int degree = length+1;
		BinaryTreeNode headNode;
		headNode = NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedNonuniformTree(degree, cCount);
		return(MappingUtils.traverseTreeToString(
				headNode, NSConstants.MAPPING_MODE_TOPO_ACT, order));
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_INPUT)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(true) 
        	.setShortFlag('i') 
        	.setLongFlag(PARAM_INPUT);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_BASELINE)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(true)
			.setDefault(""+NSConstants.DEFAULT_BASELINE)
			.setShortFlag('b') 
			.setLongFlag(NSConstants.PARAM_BASELINE);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false)
    		.setDefault(NSConstants.DEFAULT_ORDER)
    		.setShortFlag('o') 
    		.setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_DATABASE_NAME)
        	.setStringParser(JSAP.STRING_PARSER)
			.setRequired(false)
			.setDefault("0")
			.setShortFlag('n') 
			.setLongFlag(PARAM_DATABASE_NAME);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(PARAM_SET_NUM)
	    	.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault("0")
			.setShortFlag('s') 
			.setLongFlag(PARAM_SET_NUM);
	    jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CreateRandomDBOfSource.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
