package org.cn3.ns.generation;

import java.io.File;
import java.util.Random;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappingUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Generates two FASTA files per length, each containing a given number of random tree sequences.
 * Input parameters determine the sequence lengths included in the run.
 * Used for producing length-pair baselines in turn used for alignment score normalization.
 * 
 * java org.cn3.ns.generation.CreateRandomTreeSequencesByLength -o STL
 * 
 * On Krasnow Cluster
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.CreateRandomTreeSequencesByLength -o STL --output-dir /home/tgillett/fastaFromSwc/random/Mapping2/Method3/ --startLength 20 --stopLength 220 --lengthAdd 5
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.CreateRandomTreeSequencesByLength -o STL --output-dir /home/tgillett/fastaFromSwc/random/Mapping2/Method3/ --startLength 20 --stopLength 500 --lengthMult 1.1 -n 1000 --lengthAdd 0
 *
 * For a random, non-tree-based sequence (but still with the given node-type frequencies)
 * java org.cn3.ns.scripts.CreateRandomTreeSequencesByLength -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS --tree false
 */
public class CreateRandomTreeSequencesByLength {

	public static final String PARAM_OUTPUT_DIR = "output-dir";
	public static final String PARAM_NUM_SEQUENCES = "numSequences";
	public static final String PARAM_TREE_SEQUENCE = "tree";
	public static final String PARAM_SEQUENCE_LENGTH = "length";
	public static final String PARAM_START_LENGTH = "startLength";
	public static final String PARAM_STOP_LENGTH = "stopLength";
	public static final String PARAM_LENGTH_ADD = "lengthAdd";
	public static final String PARAM_LENGTH_MULT = "lengthMult";
	
	private static final int DEFAULT_NUM_SEQUENCES = 100;
	private static final int DEFAULT_START_LENGTH = 150;
	private static final int DEFAULT_STOP_LENGTH = 180;
	private static final int DEFAULT_LENGTH_ADD = 1;
	private static final double DEFAULT_LENGTH_MULT = 1;
	private static final long RANDOM_SEED = 1200;

	private static String order;
	private static Random rand = new Random(RANDOM_SEED);

	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			
			order = config.getString(NSConstants.PARAM_ORDER);
			int seqLen = config.getInt(PARAM_SEQUENCE_LENGTH);
			int numSequences = config.getInt(PARAM_NUM_SEQUENCES);
			int startLength = config.getInt(PARAM_START_LENGTH);
			int stopLength = config.getInt(PARAM_STOP_LENGTH);
			int lengthAdd = config.getInt(PARAM_LENGTH_ADD);
			int baseline = config.getInt(NSConstants.PARAM_BASELINE);
			double lengthMult = config.getDouble(PARAM_LENGTH_MULT);
			boolean treeSequence = config.getBoolean(PARAM_TREE_SEQUENCE);
			File outputDir = new File(config.getString(PARAM_OUTPUT_DIR));
			if (!outputDir.isDirectory()){
				throw new Exception("Output directory required. "+outputDir+" is not a directory.");
			}
			
			StringBuffer sb;
			
			if (seqLen > 0){
				startLength = stopLength = seqLen;
				lengthAdd = 1;
				lengthMult = 1;
			}
			String seqNameBase, filename, sequence;
			for (int sequenceLength = startLength; sequenceLength <= stopLength; sequenceLength=(int)Math.round(sequenceLength*lengthMult)+lengthAdd){
				System.out.println("Working on length "+sequenceLength);
				for (int i = 1; i <= 2; i++){
					if (i == 1){
						seqNameBase = "RandB"+baseline+"_SetA_Len"+sequenceLength;
						filename = outputDir.getAbsolutePath()+"/RandA";
					}
					else{
						seqNameBase = "RandB"+baseline+"_SetB_Len"+sequenceLength;
						filename = outputDir.getAbsolutePath()+"/RandB";
					}
					sb = new StringBuffer();
					for (int seqNum = 1; seqNum <= numSequences; seqNum++){
						sb.append(">"+seqNameBase+"_"+seqNum+"\n");
						switch(baseline){
						case NSConstants.DISTRIBUTION_UNIFORM:
							sequence = MappingUtils.traverseTreeToString(
									NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeByDegree(sequenceLength+1),
									NSConstants.MAPPING_MODE_TOPO_ACT,order);
							break;
						case NSConstants.DISTRIBUTION_TERMINAL_GROWTH:
							sequence = MappingUtils.traverseTreeToString(
									NodeSequenceBinaryTreeGenerator.generateTerminalGrowthBinaryTreeByDegree(sequenceLength+1),
									NSConstants.MAPPING_MODE_TOPO_ACT,order);
							break;
						case NSConstants.DISTRIBUTION_INTERSTITIAL_GROWTH:
							sequence = MappingUtils.traverseTreeToString(
									NodeSequenceBinaryTreeGenerator.generateSegmentalGrowthBinaryTreeByDegree(sequenceLength+1),
									NSConstants.MAPPING_MODE_TOPO_ACT,order);
							break;
						case NSConstants.DISTRIBUTION_RANDOM_SEQUENCE:
							sequence = generateRandomSequence(sequenceLength);
							break;
						default:
							throw new Exception("Baseline "+baseline+" not accepted.");
						}
						sb.append(sequence+"\n");
					}
					// Write sequences to file
					if (treeSequence){
						filename += order;
					}
					else{
						filename += "NonTree";
					}
					filename += "-Length"+sequenceLength+"_"+numSequences+"Seqs.fasta";
					
					DataWriter.writeStringToFile(
							filename,
							sb.toString());
				}
			}
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static String generateRandomSequence(int sequenceLength){
		int cCount;
		if (sequenceLength%2 == 1){
			cCount = sequenceLength/3-1;
		}
		else{
			cCount = (sequenceLength-1)/3;
		}
		int aCount = (sequenceLength - cCount - 1)/2;
		int tCount = aCount+1;
		
		String seq = "";
		int r;
		while (cCount+aCount+tCount > 0){
			r = (int)(Math.random()*(cCount+aCount+tCount));
			if (r < cCount){
				seq += "C";
				cCount--;
			}
			else if (r < cCount+aCount){
				seq += "A";
				aCount--;
			}
			else{
				seq += "T";
				tCount--;
			}
		}
		return seq;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false)
    		.setDefault(NSConstants.DEFAULT_ORDER)
    		.setShortFlag('o') 
    		.setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_BASELINE)
		.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+NSConstants.DISTRIBUTION_UNIFORM)
		.setShortFlag('b') 
		.setLongFlag(NSConstants.PARAM_BASELINE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SEQUENCE_LENGTH)
		.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault("0")
		.setShortFlag('l') 
		.setLongFlag(PARAM_SEQUENCE_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT_DIR)
        .setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setDefault("G:/RandomSequenceFiles/")
		.setShortFlag('O') 
		.setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_NUM_SEQUENCES)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_NUM_SEQUENCES)
		.setShortFlag('n') 
		.setLongFlag(PARAM_NUM_SEQUENCES);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_TREE_SEQUENCE)
    	.setStringParser(JSAP.BOOLEAN_PARSER)
		.setRequired(false)
		.setDefault("true")
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_TREE_SEQUENCE);
        jsap.registerParameter(opt);
        
        opt = new FlaggedOption(PARAM_START_LENGTH)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_START_LENGTH)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_START_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_STOP_LENGTH)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_STOP_LENGTH)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_STOP_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_LENGTH_ADD)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_LENGTH_ADD)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_LENGTH_ADD);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_LENGTH_MULT)
    	.setStringParser(JSAP.DOUBLE_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_LENGTH_MULT)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_LENGTH_MULT);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CreateRandomTreeSequencesByLength.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
