package org.cn3.ns.motif.analysis;

import java.text.NumberFormat;
import java.util.List;
import java.util.Map;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.generation.ConstraintsDataReader;
import org.cn3.ns.motif.generation.MotifConstants;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Produces the C frequencies of some number of surrogates for each real sequence.
 * This data will ultimately be used to determine what proportion of the distribution
 * lies between the 2.5th and 97.5th percentiles. Normally one would expect it to be
 * 95%, but if the values at the 2.5th and/or 97.5th percentiles exist in the 
 * distribution multiple times, they may take up more than 95% of the distribution.
 * 
 * New surrogate set storage
 * java org.krasnow.cng.motif.analyze.CDistributionAnalyzer -o LTS -s 100 -T 0 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/
 * 
 */
public class CDistributionAnalyzer {
	
	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_SURROGATE_SET = "surrogateSet";
	
	private static int method;
	private static String order;
	private static int sets;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
//	private static int debug = 0;
	
	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		method = config.getInt(NSConstants.PARAM_RANDOM_METHOD);
		order = config.getString(NSConstants.PARAM_ORDER);
		sets = config.getInt(MotifConstants.PARAM_SET_NUM);
		String outputDir = config.getString(PARAM_OUTPUT_DIR);
		int surrogateSet = config.getInt(PARAM_SURROGATE_SET);
		boolean compressed = false;
		
		System.out.println("Initializing...");
		
		// Get sequence names and map them to unique ids
		Map<String,Integer> lengthMap = NSDataReader.loadLengthDatabase(mapping);
		
		// Initializing string buffers for output
		StringBuffer sb = new StringBuffer();
		sb.append("Sequence,Length");
		for (int i = 1; i <= sets; i++){
			sb.append(",S"+i);
		}
		sb.append("\n");

		// Load up virtual sequences (key by sequenceId) from the N virtual sequence sets (standard N=20)
		System.out.println("Loading virtual sequences");
		Map<String,List<String>> virtualSequenceMap;
		System.out.println(MotifConstants.getSurrogatesFilename(surrogateSet, order, method, sets));
		virtualSequenceMap = ConstraintsDataReader.loadSurrogates(
				surrogateSet, order, method, sets, compressed).get(0);

		// Determine whether nmers are constrained
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(3);
		
		String virtualSequence;
		List<String> sequenceList;
		Map<String,Number> motifCounts;
		for (String sequenceName : virtualSequenceMap.keySet()){
			sb.append(sequenceName);
			sequenceList = virtualSequenceMap.get(sequenceName);
			sb.append(","+lengthMap.get(sequenceName));
			for (int i = 0; i < sequenceList.size(); i++){
				virtualSequence = sequenceList.get(i);
				//System.out.println(virtualSequence);
				motifCounts = MotifFinder.findMotifs(virtualSequence,1);
				sb.append(","+motifCounts.get("C"));
			}
			sb.append("\n");
		}
		
 		// Write results to a csv file
		//String suffix = "-"+sets+"Sets-"+surrogateSet;
		//suffix += "-Size"+order+"-B"+method+"-T"+threshold+thresholdTypeString;
		String filename = outputDir+"/CDistributions-set"+surrogateSet+".csv";
		DataWriter.writeStringToFile(filename, sb.toString());

	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters();
        
        FlaggedOption opt;
        
	    opt = new FlaggedOption(NSConstants.PARAM_DEBUG)
	    	.setStringParser(JSAP.INTEGER_PARSER)
	        .setRequired(false)
	        .setDefault("0")
	        .setShortFlag(JSAP.NO_SHORTFLAG) 
	        .setLongFlag(NSConstants.PARAM_DEBUG);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_OUTPUT_DIR)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("C:/Documents and Settings/gillette/My Documents/research/MotifAnalysis/PercentileRanks/")
        .setShortFlag('O') 
        .setLongFlag(PARAM_OUTPUT_DIR);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_SURROGATE_SET)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("-1")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SURROGATE_SET);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator<?> errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CDistributionAnalyzer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
