package org.cn3.ns.motif.analysis;

import java.text.NumberFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.utils.StringUtils;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.motif.generation.ConstraintsDataReader;
import org.cn3.ns.motif.generation.MotifConstants;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * 
 * Generates csv file for R data-frame use which contains a row for every position for every sequence.
 * Each row contains the position and a 0 or 1 depending on whether the position has a C node.
 * Output of this program was used to generate Figure 4e of Gillette and Ascoli, 2015.
 *
 * 4/25/14
 * Biased Terminal Growth (Q0S0.4), relative range
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name BiasedTerminalGrowth --fastaFile ~/results/MotifAnalysis/BiasedTerminalGrowthSequences.fasta -o STL --outputDir ~/results/
 *
 * 2/28/14, Reals STL SpurLen2D
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -D NM_SpurLen2D -o STL --outputDir ~/results/
 * Surrogate set: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -D NM_SpurLen2D -o STL --outputDir ~/results/ --useRandomSet 1 --useRandomMethod 8 --useRandomN 40
 * Growth Models
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name ApicalGrowth --fastaFile ~/results/MotifAnalysis/ApicalModelV2GrowthSequences.fasta -o STL --outputDir ~/results/ 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name TerminalGrowth --fastaFile ~/results/MotifAnalysis/TerminalGrowthSequencesExt.fasta -o STL --outputDir ~/results/
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name SegmentalGrowth --fastaFile ~/results/MotifAnalysis/SegmentalGrowthSequencesExt.fasta -o STL --outputDir ~/results/
 * Growth Models relative range
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name TerminalGrowth --fastaFile ~/results/MotifAnalysis/TerminalGrowthSequencesExt.fasta -o STL --outputDir ~/results/
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer --name SegmentalGrowth --fastaFile ~/results/MotifAnalysis/SegmentalGrowthSequencesExt.fasta -o STL --outputDir ~/results/
 *
 * 1/10/14, Reals STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -o STL --outputDir ~/results/
 * Surrogate set: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -o STL --outputDir ~/results/ --useRandomSet 1 --useRandomMethod 8 --useRandomN 40
 *
 * Relative position range
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -o STL --outputDir ~/results/ -r 2
 * Surrogate set: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CsPerPositionAnalyzer -o STL --outputDir ~/results/ -r 2 --useRandomSet 1 --useRandomMethod 8 --useRandomN 40
 */
public class CsPerPositionAnalyzer {
	
	private static final String PARAM_NAME = "name";
	private static final String PARAM_FASTA_FILE = "fastaFile";
	private static final String PARAM_OUTPUT_DIR = "outputDir";
	// If given, this determines which surrogate set to use in place of real neurites
	private static final String PARAM_USE_RANDOM_SET = "useRandomSet";
	// If using surrogate set in place of real neurites, this will determine which k from which to take (what constraint level)
	private static final String PARAM_USE_RANDOM_METHOD = "useRandomMethod";
	 // Number of random sequences per real sequence; required for name of surrogate filename 
	private static final String PARAM_USE_RANDOM_N = "useRandomN";
	
	private static String dataset = "NM";
	private static String name = "NM";
	private static String fastaFile;
	private static String order;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
	private static String sourceTraversal;
	private static boolean retraverse = false;
	
	private static int useRandomSet = 0;
	private static int useRandomMethod = 0;
	private static int useRandomN = 0;

	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		dataset = config.getString(NSConstants.PARAM_DATASET);
		name = config.getString(PARAM_NAME);
		fastaFile = config.getString(PARAM_FASTA_FILE);
		// Traversal order
		order = config.getString(NSConstants.PARAM_ORDER);
		
		String outputDir = config.getString(PARAM_OUTPUT_DIR);

		sourceTraversal = order;
		if (config.userSpecified(NSConstants.PARAM_SOURCE_TRAVERSAL)){
			sourceTraversal = config.getString(NSConstants.PARAM_SOURCE_TRAVERSAL);
			if (!sourceTraversal.equals(order)){
				retraverse = true;
			}
		}
		
		// If provided, an additional surrogate set will take the place of real neurites
		useRandomSet = config.getInt(PARAM_USE_RANDOM_SET);
		// Provides the opportunity to use a different surrogate generation method than used by the baseline
		useRandomMethod = config.getInt(PARAM_USE_RANDOM_METHOD);
		useRandomN = config.getInt(PARAM_USE_RANDOM_N);

		
		Map<String,String> sequences;
		if (useRandomSet > 0){
			sequences = loadSurrogateSequences();
		}else if (!StringUtils.isEmpty(name)&&!StringUtils.isEmpty(fastaFile)){
			sequences = loadSequencesFromFile();
		}else{
			sequences = loadRealSequences();
		}

		// Initializing string buffer for output
		StringBuffer sb = new StringBuffer();
		sb.append("Sequence,Position,IsC,Length\n");

		// Loop through sequences to calculate %C within a given sequence position range
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(3);
		String sequence;
		for (String sequenceName : sequences.keySet()){
			System.out.println(sequenceName+" ");
			sequence = sequences.get(sequenceName);
			// Get kmer counts
			for (int pos = 0; pos < sequence.length(); pos++){
				sb.append(sequenceName+","+(pos+1)+","+(sequence.charAt(pos)=='C'?1:0)+","+sequence.length()+"\n");
			}
		}	
		
 		// Write results to a csv file
		String filename;
		if (useRandomSet > 0){
			filename = outputDir + "/CPercentPerPos-"+dataset+"-B"+useRandomMethod+"_Set"+useRandomSet+"_"+useRandomN+"-Size"+order+(retraverse?"-From_"+sourceTraversal:"")+".csv";
		} else if (!StringUtils.isEmpty(name)){
			filename = outputDir + "/CPercentPerPos-"+name+"-Size"+order+(retraverse?"-From_"+sourceTraversal:"")+".csv";
		}else{
			filename = outputDir + "/CPercentPerPos-"+dataset+"-Size"+order+(retraverse?"-From_"+sourceTraversal:"")+".csv";
		}
		
		DataWriter.writeStringToFile(filename, sb.toString());
		
	}

	private static Map<String,String> loadRealSequences() throws Exception{
		FastaReader fr = new FastaReader();
		Map<String,String> sequences = new HashMap<String,String>();
		fr.setFile(DataProperties.getInstance().getFastaArborDir(dataset)+"Mapping"+mapping+"/Size/"+dataset+"ArborSize"+order+"Map"+mapping+".fasta");
		while (fr.next()){
			sequences.put(fr.getSequenceName(),fr.getSequence());
		}
		return sequences;
	}

	private static Map<String,String> loadSequencesFromFile() throws Exception{
		FastaReader fr = new FastaReader();
		Map<String,String> sequences = new HashMap<String,String>();
		fr.setFile(fastaFile);
		while (fr.next()){
			sequences.put(fr.getSequenceName(),fr.getSequence());
		}
		return sequences;
	}

	private static Map<String,String> loadSurrogateSequences() throws Exception{
		Map<String,String> sequences = new HashMap<String,String>();
		
		// New method of loading extra surrogate set
		Map<Integer,Map<String,List<String>>> virtualSequencesPerK = 
			ConstraintsDataReader.loadSurrogates(
				useRandomSet, dataset, sourceTraversal, useRandomMethod, useRandomN, false);
		
		Map<String,List<String>> surrogateLists = virtualSequencesPerK.get(0);
		for (String sequenceName : surrogateLists.keySet()){
			sequences.put(sequenceName, surrogateLists.get(sequenceName).get(0));
		}
		
		return sequences;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters(false);
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(NSConstants.DEFAULT_DATASET)
        .setShortFlag('D') 
        .setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_NAME)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_NAME);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FASTA_FILE)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_FASTA_FILE);
        jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_SET)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_SET);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_METHOD)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_METHOD);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_N)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_N);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_DEBUG)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_DEBUG);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_OUTPUT_DIR)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("C:/Documents and Settings/gillette/My Documents/research/MotifAnalysis/PercentileRanks/")
        .setShortFlag('O') 
        .setLongFlag(PARAM_OUTPUT_DIR);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_SOURCE_TRAVERSAL)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DEFAULT_ORDER)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_SOURCE_TRAVERSAL);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CsPerPositionAnalyzer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
