package org.cn3.ns.motif.analysis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.utils.ClassificationUtils;
import org.cn3.ns.data.DataConverter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Find matching segments (motifs) of a given length within a sequence or between two sequences.
 * Keeps count of how many are found of each motif and stores the frequencies in a HashMap
 *
 */
public class MotifMonitor {
	
	private final static String PARAM_MOTIF_SIZE = "motif-size";
	private final static String PARAM_PRENORMALIZE = "prenormalize";
	private final static String PARAM_FROMTOP = "fromtop";
	private final static String PARAM_ARBORS = "arbors";
	private final static String PARAM_RANDOMS = "randoms";
	private final static String PARAM_SUMMARY = "summary";
	private final static String PARAM_MIN_LENGTH = "minLength";
	private final static String PARAM_MAX_LENGTH = "maxLength";
	private final static int DEFAULT_MOTIF_SIZE = 6;
	
	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			int mapping = config.getInt(NSConstants.PARAM_MAPPING_MODE);
			boolean arbors = config.getBoolean(PARAM_ARBORS);
			boolean randoms = config.getBoolean(PARAM_RANDOMS);
			boolean summary = config.getBoolean(PARAM_SUMMARY);
			
			String traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
			String order = config.getString(NSConstants.PARAM_ORDER);
			int motifSize = config.getInt(PARAM_MOTIF_SIZE);
			
			int minLength = config.getInt(PARAM_MIN_LENGTH);
			int maxLength = config.getInt(PARAM_MAX_LENGTH);

			DataProperties props = DataProperties.getInstance();
			
			Map cellData = MetaDataManager.getCellMetaData();
			List arborizations = NSDataReader.loadSequenceList(mapping);
			List lengths = NSDataReader.loadUniqueSequenceLengths(mapping);
			String arborizationFileBase = props.getFastaDir()+"Arborization/Mapping"+mapping+"/"+traversal+"/";
			String randomFileBase = props.getFastaRandomsDir()+"Method3/Mapping"+mapping+"/"+traversal+order+"/";
			String random4FileBase = props.getFastaRandomsDir()+"Method4/Mapping"+mapping+"/"+traversal+order+"/";
			String fastaFilename;

			Map currMotifMap, mapOfMotifMaps = new HashMap(), countMap = new HashMap();
			// Initialize mapOfMotifMaps - keys are motifs
			//initializeMotifs(mapOfMotifMaps, motifSize);
			
			NeuronMetadata neuron;
			String arborization, neuronName, arborizationType, cellTypeArbType;
			FastaReader fr;
			Set motifs = NeuriteSequenceStringUtils.generatePossibleACTMotifsSet(motifSize),
				cellTypes = new TreeSet(), arbTypes = new TreeSet(), 
				cellTypeArbTypes = new TreeSet(), arbTypeLens = new TreeSet(), 
				randLenBins = new TreeSet();
			Set lengthTypes = new LinkedHashSet(), all = new LinkedHashSet();

			if (arbors){
				all.add("Arborizations");
				if (randoms){
					all.add("DBDistRandoms"); all.add("DistRandoms");
				}
			}
			if (randoms){
				all.add("DBRandoms"); all.add("UniDist");
				all.add("MonomerGenUniDist");
			}
			
			int length, current; Integer lenObj;
			Map currentPerLength = new HashMap();
			
			// Loop arborizations
			for (int i = 0; arbors && i < arborizations.size(); i++){
				arborization = (String)arborizations.get(i);
				neuronName = NeuriteSequenceStringUtils.getNeuronNameFromSequence(arborization);
				if (cellData.containsKey(neuronName)){
					System.out.println(arborization);
					// Get neuron data for arborization's lab - used to find file; + classification
					neuron = (NeuronMetadata)cellData.get(neuronName);
					fastaFilename = arborizationFileBase + neuron.getLabName() 
						+ "/" + arborization + "." + order.toLowerCase() + ".fasta";
					// Get sequence
					fr = new FastaReader(fastaFilename);
					fr.next();
					length = fr.getSequenceLength();
	
					if (length >= minLength && (maxLength == 0 || length <= maxLength)){
						
						// Generate motif map for individual arborization sequence
						currMotifMap = getMotifs(fr.getSequence(), config);
						
						// ClassNames & reclassification
						//arborizationType = ClassificationUtils.reclassifyArborizationTypes(
						//		StringUtils.getArborizationType(arborization));
						arborizationType = ClassificationUtils.reclassifyArborizationTypes(
								NeuriteSequenceStringUtils.getArborizationType(arborization));
						neuron.setCellType(ClassificationUtils.reclassifyCellTypes(neuron.getCellType()));
						cellTypeArbType = neuron.getCellType()+"."+arborizationType;
						
						// Add to class type lists
						cellTypes.add(neuron.getCellType());
						arbTypes.add(arborizationType);
						cellTypeArbTypes.add(cellTypeArbType);
						//lengthTypes.add("ArbLen"+fr.getSequenceLength());
						
						// Add motif map to respective class specific maps
						mergeMaps(mapOfMotifMaps,countMap,"Arborizations",motifs,currMotifMap);
						mergeMaps(mapOfMotifMaps,countMap,neuron.getCellType(),motifs,currMotifMap);
						mergeMaps(mapOfMotifMaps,countMap,arborizationType,motifs,currMotifMap);
						mergeMaps(mapOfMotifMaps,countMap,cellTypeArbType,motifs,currMotifMap);
						mergeMaps(mapOfMotifMaps,countMap,"ArbLen"+fr.getSequenceLength(),motifs,currMotifMap);
						
						// Add map of arborType crossed with length bin
						int len = fr.getSequenceLength();
						String arbTypeLenBin = getLenBinName(arborizationType, len);
						arbTypeLens.add(arbTypeLenBin);
						mergeMaps(mapOfMotifMaps,countMap,arbTypeLenBin,motifs,currMotifMap);
						
						if (i%10==0){
							System.gc();
						}
						
						// Run random of the same length as arborization
						lenObj = new Integer(length);
						if (!currentPerLength.containsKey(lenObj)){
							currentPerLength.put(lenObj, new Integer(0));
						}
						current = ((Integer)currentPerLength.get(lenObj)).intValue()+1;
						if (current == 101){
							current = 1;
						}
						currentPerLength.put(lenObj, new Integer(current));
						
						if (randoms){
							// Uniform dist randoms, length distribution same as database
							fastaFilename = randomFileBase + "LEN"+length+"NUM"+current+"LTS.fasta";
							// Get sequence
							fr = new FastaReader(fastaFilename);
							fr.next();
							
							currMotifMap = getMotifs(fr.getSequence(), config);
							motifs.addAll(currMotifMap.keySet());
							
							mergeMaps(mapOfMotifMaps,countMap,"DistRandoms",motifs,currMotifMap);
			
							fastaFilename = random4FileBase + "LEN"+length+"NUM"+current+"LTS.fasta";
							// Get sequence
							fr = new FastaReader(fastaFilename);
							fr.next();
							
							// DB randoms, length distribution same as database 
							currMotifMap = getMotifs(fr.getSequence(), config);
							motifs.addAll(currMotifMap.keySet());
							
							mergeMaps(mapOfMotifMaps,countMap,"DBDistRandoms",motifs,currMotifMap);
						}
					}
				}
			}
			
			// Loop randoms (every other, only 40 per each length)
			for (int i = 0; randoms && i < lengths.size(); i+=2){
				length = ((Integer)lengths.get(i)).intValue();
				System.out.println("Rand length "+length);
				//lengthTypes.add("ArbLen"+length);
				lengthTypes.add("RandomLen"+length);
				String lenBin = getLenBinName("Rand_", length);
				randLenBins.add(lenBin);
				for (int j = 1; j <= 40; j++){
					// Randoms method 3
					fastaFilename = randomFileBase + "LEN"+length+"NUM"+j+order+".fasta";
					// Get sequence
					fr = new FastaReader(fastaFilename);
					fr.next();
					
					// Generate motif map for individual arborization sequence
					currMotifMap = getMotifs(fr.getSequence(), config);
					motifs.addAll(currMotifMap.keySet());
					
					mergeMaps(mapOfMotifMaps,countMap,"UniDist",motifs,currMotifMap);
					//mergeMaps(mapOfMotifMaps,countMap,"RandomLen"+fr.getSequenceLength(),currMotifMap);
					mergeMaps(mapOfMotifMaps,countMap,lenBin,motifs,currMotifMap);

					
					// Randoms method 4
					fastaFilename = random4FileBase + "LEN"+length+"NUM"+j+order+".fasta";
					// Get sequence
					fr = new FastaReader(fastaFilename);
					fr.next();
					
					// Generate motif map for individual arborization sequence
					currMotifMap = getMotifs(fr.getSequence(), config);
					motifs.addAll(currMotifMap.keySet());
					
					mergeMaps(mapOfMotifMaps,countMap,"DBRandoms",motifs,currMotifMap);
					//mergeMaps(mapOfMotifMaps,countMap,"RandomLen"+fr.getSequenceLength(),currMotifMap);
					randLenBins.add("DB"+lenBin);
					mergeMaps(mapOfMotifMaps,countMap,"DB"+lenBin,motifs,currMotifMap);

				}
			}
			/*
			arbTypes.add("MonomerGen-Axon");
			runMonomerDist("Axon", mapOfMotifMaps, countMap, motifs, config);
			arbTypes.add("MonomerGen-ApicalDendrite");
			runMonomerDist("ApicalDendrite", mapOfMotifMaps, countMap, motifs, config);
			arbTypes.add("MonomerGen-Dendrite");
			runMonomerDist("Dendrite", mapOfMotifMaps, countMap, motifs, config);*/
			
			
			String outputBase = props.getResultsDir()+"Moftifs"+motifSize+"-"
				+traversal+order+"-Mapping"+mapping+"-";
			if (!config.getBoolean(PARAM_FROMTOP)){
				outputBase = outputBase+"FromLeaf-";
			}
			if (config.getBoolean(PARAM_PRENORMALIZE)){
				outputBase = outputBase+"prenorm-";
			}
			
			if (minLength > 0 && maxLength > 0){
				outputBase = outputBase+minLength+"To"+maxLength;
			}
			else if (minLength > 0){
				outputBase = outputBase+minLength+"Up";
			}
			else if (maxLength > 0){
				outputBase = outputBase+"UpTo"+maxLength;
			}
			if (!summary){
				outputBase = outputBase+"-fullDistr";
			}
			
			if (arbors){
				writeDistributions(outputBase+"All.csv",
						all, mapOfMotifMaps, countMap, motifs, summary);
				writeDistributions(outputBase+"ArbType.csv",
						arbTypes, mapOfMotifMaps, countMap, motifs, summary);/*
				writeDistributions(outputBase+"CellType.csv",
						cellTypes, mapOfMotifMaps, countMap, motifs, summary);
				writeDistributions(outputBase+"CellTypeArbType.csv",
						cellTypeArbTypes, mapOfMotifMaps, countMap, motifs, summary);
				writeDistributions(outputBase+"Length.csv",
						lengthTypes, mapOfMotifMaps, countMap, motifs, summary);
				writeDistributions(outputBase+"ArbTypeLength.csv",
						arbTypeLens, mapOfMotifMaps, countMap, motifs, summary);*/
			}
			if (randoms){
				writeDistributions(outputBase+"Randoms.csv",
						all, mapOfMotifMaps, countMap, motifs, summary);
				writeDistributions(outputBase+"RandLengthBins.csv",
						randLenBins, mapOfMotifMaps, countMap, motifs, summary);
			}
			
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static String getLenBinName(String typeName, int len){
		if (len < 45){
			return typeName + "025-043";
		}
		else if (len < 65){
			return typeName + "045-063";
		}
		else if (len < 85){
			return typeName + "065-083";
		}
		else if (len < 125){
			return typeName + "085-123";
		}
		else if (len < 165){
			return typeName + "125-163";
		}
		else if (len < 235){
			return typeName + "165-233";
		}
		else if (len < 345){
			return typeName + "235-343";
		}
		else if (len < 645){
			return typeName + "345-643";
		}
		else{
			return typeName + "645+";
		}
	}
	
	private static void runMonomerDist(
			String type, Map mapOfMotifMaps, Map countMap, Set motifs, JSAPResult config)
	throws Exception{
		String monomerDistBase = DataProperties.getInstance().getFastaRandomsDir()
			+"/Method3/Mapping"+config.getInt(NSConstants.PARAM_MAPPING_MODE)+"/1-merDist-";
		FastaReader fr = new FastaReader(monomerDistBase+type+".fasta");
		Map currMotifMap;
		while(fr.next()){
			// Generate motif map for individual sequence
			currMotifMap = getMotifs(fr.getSequence(), config);
			motifs.addAll(currMotifMap.keySet());

			// Add motif map to respective class specific maps
			mergeMaps(mapOfMotifMaps,countMap,"MonomerGenUniDist",motifs,currMotifMap);
			mergeMaps(mapOfMotifMaps,countMap,"MonomerGen-"+type,motifs,currMotifMap);
		}
	}
	
	// Frequencies NOT in terms of percentages
	private static void writeDistributions(
			String filename, Set classNamesSet, 
			Map mapOfMotifMaps, Map countMap, Set motifs, boolean summary)
	throws Exception{
		if (!summary){
			writeDistributionsComplete(filename, classNamesSet, 
					mapOfMotifMaps, countMap, motifs);
			return;
		}
		// Normalize
		//indirectNormalizeDistribution(mapOfMotifMaps, classNamesSet);
		
		DistributionBean dist;
		boolean usedMotif;
		Map currMotifMap;
		String motif, className;
		StringBuffer sb = new StringBuffer(), line;
		Iterator it1, it2;
		sb.append("Motif");
		for (it1 = classNamesSet.iterator(); it1.hasNext();){
			className = (String)it1.next();
			sb.append(","+className+"("+countMap.get(className+"Count")+"),");
		}
		sb.append("\r\n");
		for (it1 = classNamesSet.iterator(); it1.hasNext();){
			it1.next();
			sb.append(",Mean Freq.,StdDev");
		}
		sb.append("\r\n");
		for (it1 = motifs.iterator(); it1.hasNext();){
			motif = (String)it1.next();
			currMotifMap = (Map)mapOfMotifMaps.get(motif);
			usedMotif = false;

			line = new StringBuffer();
			line.append(motif);
			
			for (it2 = classNamesSet.iterator(); it2.hasNext();){
				className = (String)it2.next();
				//count = ((Integer)countMap.get(className+"Count")).intValue();
				if (currMotifMap.containsKey(className)){
					//freq = ((Double)currMotifMap.get(className)).doubleValue()/count;
					dist = (DistributionBean)currMotifMap.get(className);
					line.append(","+dist.getMean()+","+dist.getStandardDeviation());
					if (dist.getTotal() > 0){
						usedMotif = true;
					}
				}
				else{
					line.append(",0,0");
				}
			}
			if (usedMotif){
				sb.append(line+"\r\n");
			}
		}
		DataWriter.writeStringToFile(filename, sb.toString());
	}

	private static void writeDistributionsComplete(
			String outputBase, Set classNamesSet, Map mapOfMotifDistMaps, Map countMap, Set motifs)
	throws Exception{
		List dist;
		Map arborTypeMotifDists;
		String type, motif;
		StringBuffer sb = new StringBuffer();
		int maxCount = 0, arborCount;

		List columnList = new ArrayList();

		// Setup for writing out full distributions in columnar format
		for (Iterator motifIt = motifs.iterator(); motifIt.hasNext();){
			motif = (String)motifIt.next();
			arborTypeMotifDists = (Map)mapOfMotifDistMaps.get(motif);
				// Find size of largest data set (will be size of largest arbor type)
			for (Iterator it = classNamesSet.iterator(); it.hasNext();){
				type = (String)it.next();
				arborCount = 
					((List)arborTypeMotifDists.get(type+"Full")).size();
				if (arborCount > maxCount){
					maxCount = arborCount;
				}
				// Put all distributions into a list
				columnList.add((List)arborTypeMotifDists.get(type+"Full"));
				// Write header
				sb.append(motif+"_"+type+",");
			}
		}
		sb.append("\r\n");

		for (int i = 0; i < maxCount; i++){
			for (int j = 0; j < columnList.size(); j++){
				dist = (List)columnList.get(j);
				if (j > 0){
					sb.append(",");
				}
				if (i < dist.size() && dist.get(i) != null){
					sb.append(dist.get(i));
				}
			}
			sb.append("\r\n");
		}
		
		DataWriter.writeStringToFile(outputBase, sb.toString());
	}
	
	/*private static void writeDistributions(String filename, Set classNamesSet, Map mapOfMotifMaps)
	throws Exception{
		int count;
		double freq;
		Map currMotifMap;
		String motif, className;
		StringBuffer sb = new StringBuffer();
		Iterator it1, it2;
		sb.append("ClassName,N,Motif,Frequency\r\n");
		for (it1 = classNamesSet.iterator(); it1.hasNext();){
			className = (String)it1.next();
			currMotifMap = (Map)mapOfMotifMaps.get(className);
			count = ((Integer)mapOfMotifMaps.get(className+"Count")).intValue();
			for (it2 = currMotifMap.keySet().iterator(); it2.hasNext();){
				motif = (String)it2.next();
				freq = ((Integer)currMotifMap.get(motif)).doubleValue()/count;
				sb.append(className+","+count+","+motif+","+freq+"\r\n");
			}
		}
		DataWriter.writeStringToFile(filename, sb.toString());
	}*/

	
	private static void mergeMaps(Map masterMap, Map countMap, String className, 
			Set motifs, Map motifCountMap){
		String motif;
		Map classDistMap;
		double val;
		DistributionBean distBean;
		List dataList;
		for (Iterator motifIt = motifs.iterator(); motifIt.hasNext();){
			motif = (String)motifIt.next();
			if (!masterMap.containsKey(motif)){
				masterMap.put(motif, new HashMap());
			}
			classDistMap = (Map)masterMap.get(motif);
			if (!classDistMap.containsKey(className)){
				classDistMap.put(className,new DistributionBean());
				classDistMap.put(className+"Full",new ArrayList());
			}
			distBean = (DistributionBean)classDistMap.get(className);
			dataList = (List)classDistMap.get(className+"Full");
			if (motifCountMap.containsKey(motif)){
				val = ((Number)motifCountMap.get(motif)).doubleValue();
				distBean.addData(val);
				dataList.add(new Double(val));
			}
			else{
				distBean.addData((double)0);
				dataList.add(new Double(0));
			}
		}
		if (countMap.containsKey(className+"Count")){
			countMap.put(className+"Count", 
				new Integer(((Integer)countMap.get(className+"Count")).intValue()+1));
		}
		else{	
			countMap.put(className+"Count", new Integer(1));
		}
	}
	/*
	private static void mergeMaps(Map masterMap, String className, Map map2){
		if (masterMap.containsKey(className)){
			masterMap.put(className, mergeMaps((Map)masterMap.get(className), map2));
			masterMap.put(className+"Count", 
					new Integer(((Integer)masterMap.get(className+"Count")).intValue()+1));
		}
		else{
			masterMap.put(className, map2);
			masterMap.put(className+"Count", new Integer(1));
		}
	}

	private static Map mergeMaps(Map map1, Map map2){
		Map output = new HashMap();
		Object key;
		int value1, value2;
		output.putAll(map1);
		for (Iterator it = map2.keySet().iterator(); it.hasNext();){
			key = it.next();
			if (output.containsKey(key)){
				value1 = ((Integer)map1.get(key)).intValue();
				value2 = ((Integer)map2.get(key)).intValue();
				output.put(key, new Integer(value1 + value2));
			}
			else{
				output.put(key, map2.get(key));
			}
		}
		return output;
	}*/

	private static Map getMotifs(String sequence, JSAPResult config) throws Exception{
		Map motifMap;
		if (config.getBoolean(PARAM_FROMTOP)){
			if (config.getBoolean(PARAM_PRENORMALIZE)){
				motifMap = MotifFinder.findMotifFrequencies(
						sequence,config.getInt(PARAM_MOTIF_SIZE));
			}
			else{
				motifMap = MotifFinder.findMotifs(sequence,config.getInt(PARAM_MOTIF_SIZE));	
			}
		}
		else{
			BinaryTreeNode tree;
			switch (config.getInt(NSConstants.PARAM_MAPPING_MODE)){
			case NSConstants.MAPPING_MODE_TOPO_CA:
				tree = DataConverter.fastaCAStringToTree(sequence);
				break;
			case NSConstants.MAPPING_MODE_TOPO_ACT:
			case NSConstants.MAPPING_MODE_TOPO_GACT:
				tree = DataConverter.fastaACTStringToTree(sequence);
				break;
			case NSConstants.MAPPING_MODE_TOPO_LEN_ANG:
			case NSConstants.MAPPING_MODE_TOPO_LOCAL_LEN_ANG:
				tree = DataConverter.fastaACTStringToTree(sequence);
				break;
			default:
					throw new Exception(
							"Unknown node mapping "+config.getInt(NSConstants.PARAM_MAPPING_MODE));
			}
			motifMap = MotifFinder.findMotifs(
					tree,
					config.getInt(PARAM_MOTIF_SIZE),
					config.getBoolean(PARAM_PRENORMALIZE));
		}
		return motifMap;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
    		.setStringParser(JSAP.INTEGER_PARSER)
    		.setRequired(false)
    		.setDefault(""+NSConstants.DEFAULT_MAPPING_MODE)
    		.setShortFlag('M') 
    		.setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(NSConstants.PARAM_TRAVERSAL)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(false)
        	.setDefault("Size")
        	.setShortFlag('T') 
        	.setLongFlag(NSConstants.PARAM_TRAVERSAL);
        jsap.registerParameter(opt);
        
        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false)
    		.setDefault("LTS")
    		.setShortFlag('O') 
    		.setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(PARAM_SUMMARY)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("true")
			.setShortFlag('y') 
			.setLongFlag(PARAM_SUMMARY);
	    jsap.registerParameter(opt);
        
        opt = new FlaggedOption(PARAM_MOTIF_SIZE)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_MOTIF_SIZE)
			.setShortFlag('S') 
			.setLongFlag(PARAM_MOTIF_SIZE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_PRENORMALIZE)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("TRUE")
			.setShortFlag('N') 
			.setLongFlag(PARAM_PRENORMALIZE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FROMTOP)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("TRUE")
			.setShortFlag('t') 
			.setLongFlag(PARAM_FROMTOP);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_RANDOMS)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("TRUE")
			.setShortFlag('r') 
			.setLongFlag(PARAM_RANDOMS);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_ARBORS)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("TRUE")
			.setShortFlag('a') 
			.setLongFlag(PARAM_ARBORS);
        jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_MIN_LENGTH)
	    	.setStringParser(JSAP.INTEGER_PARSER)
	        .setRequired(false)
	        .setDefault("0")
	        .setShortFlag(JSAP.NO_SHORTFLAG) 
	        .setLongFlag(PARAM_MIN_LENGTH);
	    jsap.registerParameter(opt);
	
	    opt = new FlaggedOption(PARAM_MAX_LENGTH)
			.setStringParser(JSAP.INTEGER_PARSER)
		    .setRequired(false)
		    .setDefault("0")
		    .setShortFlag(JSAP.NO_SHORTFLAG) 
		    .setLongFlag(PARAM_MAX_LENGTH);
		jsap.registerParameter(opt);
    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + MotifMonitor.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
}
