package org.cn3.ns.motif.generation;

import java.io.File;
import java.math.BigInteger;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.utils.ClassificationUtils;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.generation.NodeSequenceBinaryTreeGenerator;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappedBinaryTreeUtils;
import org.cn3.ns.utils.MappingUtils;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 *
 * Produces surrogates strictly sampled uniformly from the space of tree shapes,
 * and constrained by dimer proportions. This requires sampling 
 *
 */
public class LengthDimerDistGenerator {

	private static final String PARAM_SET_NUM = "set-num";
	private static final String PARAM_METHOD = "method";
	private static final int MAX_COUNT_STORAGE_DEGREE = 18;
	private static final int MOTIF_SIZE = 2;
	
	private static Map<Integer,Map<BigInteger,Map<String,Object>>>
		treeMotifCountMap = new HashMap<Integer,Map<BigInteger,Map<String,Object>>>();
	private static List<String> motifs = NeuriteSequenceStringUtils.generatePossibleACTMotifs(MOTIF_SIZE);
	private static int storageDegree = MAX_COUNT_STORAGE_DEGREE;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			
			List<String> arborizations = NSDataReader.loadSequenceList();

			//Map arborTypeBufferMap = generatePopulations(arborizations, config);
			generatePopulations(arborizations, config);
			
			/*
			String arborType;
			for (Iterator it = arborTypeBufferMap.keySet().iterator(); it.hasNext();){
				arborType = (String)it.next();
				DataWriter.writeStringToFile(
						ProjectProperties.getInstance().getFastaRandomsDir()
							+"/Method3/Mapping2/1-merDistSet"+setNum+"-"+arborType+".fasta", 
						(StringBuffer)arborTypeBufferMap.get(arborType));
			}*/
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static void generatePopulations(List arborizations, JSAPResult config) throws Exception{
		int setNum = config.getInt(PARAM_SET_NUM);
		int method = config.getInt(PARAM_METHOD);
		Set<String> outliers = NSDataReader.loadDimerTooUnlikely(NSConstants.MAPPING_MODE_TOPO_ACT, method);
		DataProperties props = DataProperties.getInstance();
		FastaReader randFr, 
			fr = new FastaReader(props.getFastaArborDir()+"Mapping2/Size/NMArborSizeLTSMap2.fasta");

		// Get files of existing n-mer generated sets to avoid duplicating existing arbor sequences
		Map<String,String> existingRandomFilenames = setupFilenames(props, setNum, method);
		Map<String,FastaReader> existingRandomFRs = setupFRs(existingRandomFilenames, setNum);

		String arborization, arborType;
		Map<String,Number> arborMotifFreqs, randomMotifFreqs;
		int degree, tryCount, tryCountSample;
		BigInteger shapeNum = null;
		String generatedSequence;
		long startTime, timeTaken, tooLongTime = 8 * 60 * 60 * 1000; // 8 hours in milliseconds 
		boolean tooLong;
		String sequence = null;
		
		while (fr.next()){
			// Pull sequence info
			arborType = ClassificationUtils.reclassifyArborizationTypes(fr.getArborType());
			arborization = fr.getSequenceName();

			// Make sure this sequence isn't an outlier or would take too long to run (before checking the randFr)
			if (!outliers.contains(arborization)){
			
				randFr = existingRandomFRs.get(arborType);
				// Only make random of this sequence if it hasn't been generated already (no previously existing file or exhausted list)
				if (randFr == null || !randFr.next()){
					// If it wasn't null before, it should be now
					existingRandomFRs.put(arborType,null);
					
					// Get motif frequencies
					arborMotifFreqs = MotifFinder.findMotifFrequencies(fr.getSequence(), MOTIF_SIZE);

					System.out.println(arborization+" length: "+fr.getSequenceLength()+"; frequencies: "+arborMotifFreqs);
					
					// Find a tree shape that has a %C within +/- 2.5% of the sequence's within a reasonable timeframe
					tooLong = false;
					tryCount = 0;
					startTime = new Date().getTime();
					degree = fr.getSequenceLength()+1;
					tryCountSample = 24000000 / fr.getSequenceLength();
					do {
						switch (method){
						case NSConstants.RANDOM_TERMINAL_GROWTH:
							sequence = MappingUtils.traverseTreeToStringWithMappingMode2(
								NodeSequenceBinaryTreeGenerator.generateTerminalGrowthBinaryTreeByDegree(degree));
							randomMotifFreqs = MotifFinder.findMotifFrequencies(sequence);
							break;
						case NSConstants.RANDOM_UNIFORM_DIST:
						default:
							// Generate random tree number and check its %C
							BigInteger possibleShapes = MappedBinaryTreeUtils.numShapes(degree);
							shapeNum = NodeSequenceBinaryTreeGenerator.randomBigInteger(possibleShapes);
							randomMotifFreqs = getMotifFrequencies(degree, shapeNum);
							//System.out.println("trying; attempt is "+BinaryTreeGenerator.generateHardingTreeMap2Sequence(degree, shapeNum));
							//System.out.println("Motif Freqs: "+randomMotifFreqs);
							// Make sure this isn't taking too long
						}
						tryCount++;
						if (tryCount % tryCountSample == 0){
							timeTaken = new Date().getTime() - startTime;
							tooLong = timeTaken > tooLongTime;
							System.out.println("TryCount: " + tryCount+"; Time taken (min): " + timeTaken/(60000));
							System.out.println(randomMotifFreqs);
							System.gc();
						}
					} while (!isWithinRange(arborMotifFreqs, randomMotifFreqs) && !tooLong);
					
					if (tooLong){
						DataWriter.appendStringToFile(
								props.getDimerTooUnlikelyFile(NSConstants.MAPPING_MODE_TOPO_ACT, method),
								arborization+"\r\n");
					}
					else{
						System.out.println("Tree found ("+tryCount+" tries). freqs%: "+randomMotifFreqs);
						switch (method){
						case NSConstants.RANDOM_TERMINAL_GROWTH:
							generatedSequence = sequence;
							break;
						case NSConstants.RANDOM_UNIFORM_DIST:
						default:
							generatedSequence = NodeSequenceBinaryTreeGenerator.generateHardingTreeMap2Sequence(degree, shapeNum);
						}
						DataWriter.appendStringToFile((String)existingRandomFilenames.get(arborType), 
								"> Equiv-"+arborization+"\r\n"+NeuriteSequenceStringUtils.insertBreaksAtLength(generatedSequence, 80)+"\r\n");
					}
				}
			}
		}
		
		//return map;
	}

	private static Map<String,String> setupFilenames(DataProperties props, int setNum, int method) throws Exception{
		Map<String,String> existingRandomFilenames = new HashMap<String,String>();
		existingRandomFilenames.put("Axon", props.getFastaRandomsDir()+"Method"+method+"/Mapping2/2-merDistSet"+setNum+"-Axon.fasta");
		existingRandomFilenames.put("ApicalDendrite", props.getFastaRandomsDir()+"Method"+method+"/Mapping2/2-merDistSet"+setNum+"-ApicalDendrite.fasta");
		existingRandomFilenames.put("Dendrite", props.getFastaRandomsDir()+"Method"+method+"/Mapping2/2-merDistSet"+setNum+"-Dendrite.fasta");
		return existingRandomFilenames;
	}

	private static Map<String,FastaReader> setupFRs(Map<String,String> existingRandomFilenames, int setNum) throws Exception{
		Map<String,FastaReader> existingRandomFRs = new HashMap<String,FastaReader>();
		File existingRandomFile = new File(existingRandomFilenames.get("Axon"));
		if (existingRandomFile.exists()){
			existingRandomFRs.put("Axon", new FastaReader(existingRandomFile));
		}
		existingRandomFile = new File(existingRandomFilenames.get("ApicalDendrite"));
		if (existingRandomFile.exists()){
			existingRandomFRs.put("ApicalDendrite", new FastaReader(existingRandomFile));
		}
		existingRandomFile = new File(existingRandomFilenames.get("Dendrite"));
		if (existingRandomFile.exists()){
			existingRandomFRs.put("Dendrite", new FastaReader(existingRandomFile));
		}
		return existingRandomFRs;
	}

	private static boolean isWithinRange(Map<String,Number> arborMotifFreqs, Map<String,Number> randomMotifFreqs){
		double arborFreq;
		double randomFreq;
		for (String motif : motifs){
			arborFreq = arborMotifFreqs.containsKey(motif) ?
				arborMotifFreqs.get(motif).doubleValue() : 0;
			randomFreq = randomMotifFreqs.containsKey(motif) ? 
				randomMotifFreqs.get(motif).doubleValue() : 0;
			if (Math.abs(arborFreq-randomFreq) > .025) return false;
		}
		return true;
	}
	
	public static Map<String,Number> getMotifFrequencies(int degree, BigInteger shapeNum)
	throws Exception{
		return getMotifFrequencies(degree, shapeNum, true);
	}
	public static Map<String,Number> getMotifFrequencies(int degree, BigInteger shapeNum, boolean orderLTS)
	throws Exception{
		int effectiveLength = degree-MOTIF_SIZE;
		Map<String,Object> countMap = getMotifCounts(degree,shapeNum,orderLTS);
		Map<String,Number> frequencyMap = new HashMap<String,Number>();
		int count;
		for (String motif : motifs){
			count = ((Integer)countMap.get(motif)).intValue();
			//System.out.println(motif+" "+count);
			frequencyMap.put(motif, (double)count/effectiveLength);
		}
		
		return frequencyMap;
	}
	
	public static Map<String,Number> getMotifCounts(int degree, BigInteger shapeNum)
	throws Exception{
		return getMotifCounts(degree,shapeNum,"LTS");
	}
	public static Map<String,Number> getMotifCounts(int degree, BigInteger shapeNum, String order)
	throws Exception{
		Map<String,Object> countsHolder = getMotifCounts(degree, shapeNum, order.equals("LTS"));
		Map<String,Number> motifCounts = new HashMap<String,Number>();
		for (String s : countsHolder.keySet()){
			if (countsHolder.get(s).getClass().equals(Integer.class)){
				motifCounts.put(s, (Integer)countsHolder.get(s));
			}
		}
		return motifCounts;
	}
	/**
	 * DON'T RUN THIS TWICE WITH DIFFERENT ORDERS WITHOUT CLEARING MOTIF COUNTS MAP
	 */
	public static Map<String,Object> getMotifCounts(int degree, BigInteger shapeNum, boolean orderLTS)
	throws Exception{
		//System.out.println("degree:" +degree);
		Map<String,Object> countMap = new HashMap<String,Object>(), largerCountMap, smallerCountMap;
		// Resolve simple cases
		if (degree == 2){
			for (int i = 0; i < motifs.size(); i++){
				countMap.put(motifs.get(i), new Integer(0));
			}
			countMap.put("1", "T");
		}
		else if (degree == 3){
			for (int i = 0; i < motifs.size(); i++){
				countMap.put(motifs.get(i), new Integer(0));
			}
			countMap.put("CT", new Integer(1));
			countMap.put("1", "C");
		}
		else{
			// Check to see if it's been calculated before
			if (degree <= storageDegree){ // Limits possible storage
				if (treeMotifCountMap.containsKey(degree)){
					Map<BigInteger,Map<String,Object>> shapeMotifCountMap = treeMotifCountMap.get(degree);
					if (shapeMotifCountMap.containsKey(shapeNum)){
						return shapeMotifCountMap.get(shapeNum);
					}
				}
				else{
					synchronized(treeMotifCountMap){
						treeMotifCountMap.put(degree, 
							new HashMap<BigInteger,Map<String,Object>>(MappedBinaryTreeUtils.numShapes(degree).intValue()/4));
					}
				}
			}
			
			// Find whether the top node of this tree is a C node
			int leftDegree = 0;
			String first, motif;
			BigInteger partialSum;
			do{
				leftDegree++;
				partialSum = MappedBinaryTreeUtils.partialSumOfNumShapes(degree,leftDegree);
			} while(partialSum.compareTo(shapeNum) < 0);
			if (leftDegree == 1){
				first = "C";
				largerCountMap = getMotifCounts(degree-1, shapeNum, orderLTS);
				for (int i = 0; i < motifs.size(); i++){
					motif = (String)motifs.get(i);
					countMap.put(motif, largerCountMap.get(motif));
				}
				incrementCount("C" + largerCountMap.get("1"), countMap);
			}
			else{
				first = "A";
				int rightDegree = degree - leftDegree;
				BigInteger remainder, numShapesLeft, rightShape, leftShape;
				remainder =
					shapeNum.subtract(MappedBinaryTreeUtils.partialSumOfNumShapes(degree,degree-rightDegree-1));
				if (rightDegree == leftDegree){
					// Left and Right degrees are the same
					// Swapping right & left for proper traversal output
					numShapesLeft = MappedBinaryTreeUtils.numShapes(leftDegree);
					leftShape = numShapesLeft;
					rightShape = leftShape.add(BigInteger.ONE);
					BigInteger two = BigInteger.ONE.add(BigInteger.ONE);
					// Find the shape out of n*(n+1)/2 
					leftShape = NodeSequenceBinaryTreeGenerator.findRightShape(remainder,BigInteger.ONE,numShapesLeft);
					rightShape = remainder.subtract(
							leftShape.multiply(leftShape.subtract(BigInteger.ONE))
								.divide(two));
					/*
					numShapesLeft = BinaryTreeUtils.numShapes(leftDegree);
					rightShape = numShapesLeft;
					leftShape = rightShape.add(BigInteger.ONE);
					BigInteger two = BigInteger.ONE.add(BigInteger.ONE);
					// Find the shape out of n*(n+1)/2
					rightShape = BinaryTreeGenerator.findRightShape(remainder,BigInteger.ONE,numShapesLeft);
					leftShape = remainder.subtract(
							rightShape.multiply(rightShape.subtract(BigInteger.ONE))
								.divide(two));*/
				}
				else{
					numShapesLeft = MappedBinaryTreeUtils.numShapes(leftDegree);
					rightShape = remainder.subtract(BigInteger.ONE).divide(numShapesLeft).add(BigInteger.ONE);
					leftShape = remainder.subtract(numShapesLeft.multiply(rightShape.subtract(BigInteger.ONE)));
				}
				//System.out.println("LeftDeg:" +leftDegree+"; rightDeg: " +rightDegree);
				largerCountMap = getMotifCounts(rightDegree, rightShape, orderLTS);
				smallerCountMap = getMotifCounts(leftDegree, leftShape, orderLTS);
				for (int i = 0; i < motifs.size(); i++){
					motif = (String)motifs.get(i);
					countMap.put(motif, 
						new Integer(
								((Integer)largerCountMap.get(motif)).intValue()+
								((Integer)smallerCountMap.get(motif)).intValue()));
				}
				// Increment first & middle
				if (orderLTS){
					incrementCount("A" + largerCountMap.get("1"), countMap);
					incrementCount("T" + smallerCountMap.get("1"), countMap);
				}
				else{
					incrementCount("A" + smallerCountMap.get("1"), countMap);
					incrementCount("T" + largerCountMap.get("1"), countMap);
				}
			}

			countMap.put("1", first);
	
			// Store new result (if appropriate)
			if (degree <= storageDegree){
				synchronized(treeMotifCountMap){
					Map<BigInteger,Map<String,Object>> shapeMotifCountMap = treeMotifCountMap.get(degree);
					shapeMotifCountMap.put(shapeNum, countMap);
				}
			}
		}
		return countMap;
	}

	private static void incrementCount(String motif, Map<String,Object> countMap){
		int count = countMap.containsKey(motif) 
			? ((Integer)countMap.remove(motif)).intValue() : 0;
			//System.out.println("incrementing "+motif+" count");
		countMap.put(motif, new Integer(count+1));
	}

	@SuppressWarnings("rawtypes")
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_SET_NUM)
            .setStringParser(JSAP.INTEGER_PARSER)
            .setRequired(true)
            .setShortFlag('s') 
            .setLongFlag(PARAM_SET_NUM);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_METHOD)
	        .setStringParser(JSAP.INTEGER_PARSER)
	        .setRequired(false)
	        .setDefault(""+NSConstants.DISTRIBUTION_UNIFORM)
	        .setShortFlag('M') 
	        .setLongFlag(PARAM_METHOD);
	    jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + LengthDimerDistGenerator.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	public static void setStorageDegree(int storageDeg){
		storageDegree = storageDeg;
	}
	public static int getStorageDegree(){
		return storageDegree;
	}
	
	public static void clearMap(){
		treeMotifCountMap.clear();
		System.gc();
	}

}
