package org.cn3.ns.scripts;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.persistence.NeuronDAO;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.utils.StringUtils;
import org.cn3.ns.data.FastaReader;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Takes metadata description and makes a fasta file out of neurite sequences
 * that fit the description.
 * 
 * java org.cn3.ns.scripts.FastaGroupGenerator -f C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeSTLMap2.fasta -C "G:/AlignmentData/ClassFiles/DomainClassDefinitions.txt" -o G:/AlignmentData/ClassFiles/FastaFiles
 * java org.cn3.ns.scripts.FastaGroupGenerator -f C:/fastaFromSwc/NM_SpurLen5/Arborization/Mapping2/Size/NMArborSizeSTLMap2_SpurLen5.fasta -C "G:/AlignmentData/ClassFiles/DomainClassDefinitions.txt" -o G:/AlignmentData/ClassFiles/SpurRemovedFastaFiles
 * java org.cn3.ns.scripts.FastaGroupGenerator -f C:/fastaFromSwc/NM_SpurLen5/Arborization/Mapping2/Size/NMArborSizeSTLMap2_SpurLen5.fasta -C "G:/AlignmentData/ClassFiles/DomainClassDefinitions.txt" -o G:/AlignmentData/ClassFiles/ExtraTestCase
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.FastaGroupGenerator -f /homes/tgillett/fastaFromSwc/NM_SpurLen5/Arborization/Mapping2/Size/NMArborSizeSTLMap2_SpurLen5.fasta -C "/homes/tgillett/ClassCompareFiles/DomainClassDefinitions.txt" -o /homes/tgillett/ClassCompareFiles/SpurRemovedFastaFiles
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.FastaGroupGenerator -f /homes/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeSTLMap2.fasta -C "/homes/tgillett/ClassCompareFiles/DomainClassDefinitions.txt" -o /homes/tgillett/ClassCompareFiles/FastaFiles
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.FastaGroupGenerator -f /homes/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeSTLMap2.fasta -C "/homes/tgillett/ClassCompareFiles/DomainClassDefinitions.txt" -o /homes/tgillett/fastaFromSwc/SelectedClasses/
 *
 */
public class FastaGroupGenerator {

	private static final String PARAM_FASTA_FILE = "fasta";
	private static final String PARAM_CLASS_FILE = "classFile";
	private static final String PARAM_OUTPUT_DIRECTORY = "output";
	
	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			String fastaFile = config.getString(PARAM_FASTA_FILE);
			
			String outputDir = config.getString(PARAM_OUTPUT_DIRECTORY);
			// Load file containing descriptions
			Map<String,Map<String,List<String>>> classDescriptions = readComparisonListFile(config.getString(PARAM_CLASS_FILE));

			// Load metadata
			//Map<String, NeuronInfo> metadata = MetaDataManager.getExtendedCellMetaData();
			
			// Variables
			List<NeuronMetadata> neurons;
			Set<String> neuronNames;
			NeuronDAO neuronDAO = NeuronDAO.getInstance();
			List<String> arborTypes;
			
			// Loop through descriptions
			Map<String,List<String>> classDescription;
			for (String className : classDescriptions.keySet()){
				classDescription = classDescriptions.get(className);
				System.out.println(className);
				
				// Generate list of sequences fitting each description
				neurons = neuronDAO.getNeuronsByMetadata(classDescription, true);
				neuronNames = createNeuronClassSet(neurons);
				arborTypes = classDescription.get("ArborType");
				
				/*
				StringBuffer sb = new StringBuffer();
				for (String neuronName : neuronNames){
					sb.append(neuronName+", ");
				}
				System.out.println(sb.toString()+"\n");
				*/
				
				// Loop through sequences in source fasta file and put in SB
				writeNewFasta(fastaFile, outputDir, className, neuronNames, arborTypes);
			}			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static void writeNewFasta(
			String fastaFile, String outputDir, String className, Set<String> neuronNames, List<String> arborTypes) throws Exception{
		StringBuffer sb = new StringBuffer();
		
		FastaReader fr = new FastaReader(fastaFile);
		String neuronName, arborType;
		while (fr.next()){
			neuronName = StringUtils.getNeuronNameFromSequence(fr.getSequenceName());
			arborType = StringUtils.getArborizationType(fr.getSequenceName());
			if (neuronNames.contains(neuronName) && (arborTypes == null || arborTypes.contains(arborType))){
				sb.append(fr.getHeader()+"\n");
				sb.append(fr.getSequence()+"\n");
			}
		}
		
		// Write new fasta file
		DataWriter.writeStringToFile(outputDir+"/"+className+".fasta", sb.toString());
	}
	
	public static Set<String> createNeuronClassSet(List<NeuronMetadata> neurons){
		Set<String> neuronNames = new HashSet<String>(neurons.size());
		for (NeuronMetadata currNeuron : neurons){
			neuronNames.add(currNeuron.getName());
		}
		return neuronNames;
	}
	
	/**
	 * 
	 * @param filename
	 * @return Map<class name, Map<MetaType,List<MetaValues>> >
	 * For each class, its defining metatypes and associated values
	 * @throws Exception
	 */
	public static Map<String,Map<String,List<String>>> readComparisonListFile(String filename) throws Exception{
		FileReader freader = null;
	    BufferedReader reader = null;
	    Map<String,Map<String,List<String>>> classMap = new HashMap<String,Map<String,List<String>>>();
	    Map<String,List<String>> metadataMap = null;
	    try{
		    freader = new FileReader(filename);
		    reader = new BufferedReader(freader);
		    String[] split, keyValueSplit, valuesSplit;
		    List<String> values;
			String dataString = "";
			String className, metaDefs;
			
			while ( (dataString=reader.readLine()) != null ) {
				metadataMap = new HashMap<String,List<String>>();
		        split = dataString.split(":");
		        //System.out.println(dataString);
		        className = split[0];
		        metaDefs = split[1];
		        split = metaDefs.split(",");
		        for (int i = 0; i < split.length; i++){
		        	values = new ArrayList<String>();
		        	//System.out.println(split[i]);
		        	keyValueSplit = split[i].split("=");
		        	valuesSplit = keyValueSplit[1].split("\\|");
		        	for (int j = 0; j < valuesSplit.length; j++){
		        		values.add(valuesSplit[j]);
		        		//System.out.println(valuesSplit[j]);
		        	}
		        	metadataMap.put(keyValueSplit[0],values);
		        }
		        classMap.put(className,metadataMap);
			}
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
		return classMap;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(PARAM_FASTA_FILE)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(true)
    		.setShortFlag('f') 
    		.setLongFlag(PARAM_FASTA_FILE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_CLASS_FILE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true)
		.setShortFlag('C') 
		.setLongFlag(PARAM_CLASS_FILE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT_DIRECTORY)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true)
		.setShortFlag('o') 
		.setLongFlag(PARAM_OUTPUT_DIRECTORY);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + FastaGroupGenerator.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}

}