package org.cn3.ns.scripts;

import java.io.File;

import org.cn3.morpho.data.NeuronTreeClassification;
import org.cn3.ns.data.SwcToFastaConverter;
import org.cn3.ns.nodeencoding.EncodingMode;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

public class SimpleSwcToFastaConverter {

    private static final String PARAM_SWC_DIR = "swc-dir";
    private static final String PARAM_FASTA_DIR = "fasta-dir";
    private static final String PARAM_FASTA_FILE = "fasta-file";
    private static final String PARAM_MIN_LENGTH = "min-length";
    private static final String PARAM_SPUR_LENGTH = "spurLength";
    private static final String PARAM_REMOVE_DENDRITE_SPURS = "removeDendriteSpurs";
    private static final String PARAM_REMOVE_APICAL_SPURS = "removeApicalSpurs";
    private static final String PARAM_REMOVE_AXON_SPURS = "removeAxonSpurs";

    public static void main(String args[]) {
        try {
            System.out.println("Running");

            JSAPResult config = readParameters(args);
            String swcDir = config.getString(PARAM_SWC_DIR);
            String fastaDir = config.getString(PARAM_FASTA_DIR);
            String classification = config.getString(NSConstants.PARAM_CLASSIFICATION);
            String traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
            String order = config.getString(NSConstants.PARAM_ORDER);
            EncodingMode encoding = EncodingMode.find(config.getString(NSConstants.PARAM_MAPPING_MODE));

            SwcToFastaConverter swcToFastaConverter = new SwcToFastaConverter(fastaDir);
            // Classification is "Cell", "Arborization", or "Tree"
            swcToFastaConverter.setClassification(classification);
            // Traversal is "Size", "Height", or "AvgHeight"
            swcToFastaConverter.setTraversal(traversal);
            swcToFastaConverter.setOrder(order);
            swcToFastaConverter.setEncoding(encoding);
            swcToFastaConverter.setMinLength(config.getInt(PARAM_MIN_LENGTH));

            swcToFastaConverter.setSpurLength(config.getFloat(PARAM_SPUR_LENGTH));
            swcToFastaConverter.setRemoveDendriteSpurs(config.getBoolean(PARAM_REMOVE_DENDRITE_SPURS));
            swcToFastaConverter.setRemoveApicalSpurs(config.getBoolean(PARAM_REMOVE_APICAL_SPURS));
            swcToFastaConverter.setRemoveAxonSpurs(config.getBoolean(PARAM_REMOVE_AXON_SPURS));

            if (config.userSpecified(PARAM_FASTA_DIR) && config.userSpecified(PARAM_FASTA_FILE)){
                File initFastaFile = new File(config.getString(PARAM_FASTA_FILE));
                File fastaFile = new File(fastaDir+"/"+initFastaFile.getName());
                swcToFastaConverter.convertSwcFilesToFasta(swcDir,fastaFile);
            }else if (config.userSpecified(PARAM_FASTA_DIR)){
                swcToFastaConverter.convertSwcFilesToFasta(swcDir,fastaDir);
            }else if (config.userSpecified(PARAM_FASTA_FILE)){
                File fastaFile = new File(config.getString(PARAM_FASTA_FILE));
                swcToFastaConverter.convertSwcFilesToFasta(swcDir,fastaFile);
            }else{
                System.out.println("At least one of --fasta-dir (-d) or --fasta-file (-f) is required");
                System.exit(0);
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static JSAPResult readParameters(String[] args) throws Exception{
        JSAP jsap = new JSAP();

        FlaggedOption opt;
        Switch sw;

        opt = new FlaggedOption(PARAM_SWC_DIR)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(true)
                .setShortFlag('s')
                .setLongFlag(PARAM_SWC_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FASTA_DIR)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setShortFlag('d')
                .setLongFlag(PARAM_FASTA_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FASTA_FILE)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setShortFlag('f')
                .setLongFlag(PARAM_FASTA_FILE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_TRAVERSAL)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setDefault(NSConstants.DEFAULT_TRAVERSAL)
                .setShortFlag('t')
                .setLongFlag(NSConstants.PARAM_TRAVERSAL);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setDefault("STL")
                .setShortFlag('o')
                .setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setDefault(""+NSConstants.MAPPING_MODE_TOPO_ACT)
                .setShortFlag('M')
                .setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_CLASSIFICATION)
                .setStringParser(JSAP.STRING_PARSER)
                .setRequired(false)
                .setDefault(NeuronTreeClassification.SUB_ARBORIZATION.toString())
                .setShortFlag('c')
                .setLongFlag(NSConstants.PARAM_CLASSIFICATION);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MIN_LENGTH)
                .setStringParser(JSAP.INTEGER_PARSER)
                .setRequired(false)
                .setDefault("10")
                .setShortFlag('m')
                .setLongFlag(PARAM_MIN_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SPUR_LENGTH)
                .setStringParser(JSAP.FLOAT_PARSER)
                .setRequired(false)
                .setDefault("0")
                .setShortFlag(JSAP.NO_SHORTFLAG)
                .setLongFlag(PARAM_SPUR_LENGTH);
        jsap.registerParameter(opt);

        sw = new Switch(PARAM_REMOVE_DENDRITE_SPURS)
                .setShortFlag(JSAP.NO_SHORTFLAG)
                .setLongFlag(PARAM_REMOVE_DENDRITE_SPURS);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_APICAL_SPURS)
                .setShortFlag(JSAP.NO_SHORTFLAG)
                .setLongFlag(PARAM_REMOVE_APICAL_SPURS);
        jsap.registerParameter(sw);

        sw = new Switch(PARAM_REMOVE_AXON_SPURS)
                .setShortFlag(JSAP.NO_SHORTFLAG)
                .setLongFlag(PARAM_REMOVE_AXON_SPURS);
        jsap.registerParameter(sw);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
            for (@SuppressWarnings("rawtypes")
                 java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
                System.err.println("Error: " + errs.next());
            }
            System.err.println();
            System.err.println("Usage: java "
                    + SimpleSwcToFastaConverter.class.getName());
            System.err.println("                "
                    + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }

        return config;
    }
}